/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib;

import ikr.simlib.events.calendar.Calendar;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.logging.ConsoleHandler;
import java.util.logging.Formatter;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

/**
 * Log formatter, which prints SimLib time, logger name, log level and message
 * to a single line. Loglevels of SimLib messages are:
 * <ul>
 * <li>INFO for simulation phases
 * <li>CONFIG for parameter parsing
 * <li>FINE for connecting ports
 * </ul>
 */
public class SimLibFormatter extends Formatter {

   private final String newline = System.getProperty("line.separator");
   private final DecimalFormat timeFormatter = new DecimalFormat("0000.000000000");

   /**
    * Same as {@link #setupAsFormatterForANewRootConsoleHandler(Level level)}
    * with the default log level "FINE"
    */
   public static void setupAsFormatterForANewRootConsoleHandler() {
      setupAsFormatterForANewRootConsoleHandler(Level.FINE);
   }

   /**
    * Setups a new ConsoleHandler and sets the log level accordingly. If a own
    * logger configuration is passed to the JVM via the command line, then this
    * method does nothing.
    */
   public static void setupAsFormatterForANewRootConsoleHandler(Level level) {
      if (System.getProperty("java.util.logging.config.file") == null) {
         final ConsoleHandler handler = new ConsoleHandler();
         handler.setFormatter(new SimLibFormatter());
         Logger.getLogger("").addHandler(handler);

         StringBuilder configuration = new StringBuilder();
         configuration.append("handlers = java.util.logging.ConsoleHandler\n");
         configuration.append(".level = " + level.toString() + "\n");
         configuration.append("java.util.logging.ConsoleHandler.level=ALL\n");
         configuration.append("java.util.logging.ConsoleHandler.formatter=ikr.simlib.SimLibFormatter\n");
         byte[] buffer = configuration.toString().getBytes();

         try {
            LogManager.getLogManager().readConfiguration(new ByteArrayInputStream(buffer));

         } catch (SecurityException e) {
            e.printStackTrace();

         } catch (IOException e) {
            e.printStackTrace();
         }
      }
   }

   @Override
   public synchronized String format(LogRecord record) {
      final StringBuffer sb = new StringBuffer();
      sb.append(timeFormatter.format(Calendar.getInstance().getSystemTime().toSeconds()));
      sb.append(": ");
      sb.append(record.getLoggerName());
      sb.append(" ");
      sb.append(formatMessage(record));
      sb.append(newline);
      return sb.toString();
   }
}
