/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.algorithms.MessageClassificator;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;

import java.util.LinkedList;
import java.util.TreeMap;

/**
 * This class is the basis for fair queuing mechanisms, i.e. hierarchical
 * Queuing Disciplines that achieve instantaneous and work saving fair sharing
 * of a bottlenecks bandwidth for the classes / aggregates defined by a
 * {@link MessageClassificator}.
 * 
 * Fair queuing involves two tasks:
 * <ul>
 * <li>maintaining a map of the sizes of the sub-QDiscs
 * <li>maintaining a data structure needed to dequeue the "right" sub-QDisc.
 * This differs depending on the fairness metric, e.g. bytes or messages.
 * </ul>
 * While the first can mostly be implemented generally, the second cannot and
 * therefore this class is abstract.
 * 
 * @author dwagner
 * 
 */
public abstract class FairQueuing extends StackableQueuingDiscipline {
   public abstract class SortableQDiscContainer implements Comparable {
      protected StackableQueuingDiscipline qdisc;
      
      public StackableQueuingDiscipline getQDisc() {
         return qdisc; 
      }

      public abstract double getsortIndex();

      public abstract long getLengthInMetric();
      
      @Override
      public int compareTo(Object o) {
         SortableQDiscContainer other = (SortableQDiscContainer) o;
         if (getsortIndex() < other.getsortIndex())
            return -1;
         else if (getsortIndex() == other.getsortIndex())
            return 0;
         else
            return 1;
      }
   }

   protected MessageClassificator<?, SortableQDiscContainer> queues;
   protected TreeMap<Long, LinkedList<SortableQDiscContainer>> longestQList = new TreeMap<Long, LinkedList<SortableQDiscContainer>>();
   protected SortableQDiscContainer activeQ;
   protected int maxSizeInUnits;
   private int numQ = 0;
   private Parameters pars;

   @SuppressWarnings("unchecked")
   public FairQueuing(SimNode ownNode, Parameters pars) {
      super(ownNode, pars);
      this.pars = pars;
      maxSizeInUnits = pars.get(simNode, "MaxSizeInUnits").asInteger();
      queues = pars.get(simNode, "MessageClassificator").asInstance(MessageClassificator.class, simNode.createChildNode("MessageClassificator"), pars);
   }

   @Override
   abstract public boolean wouldStore(Message msg);

   abstract public SortableQDiscContainer createSortableQDisc(StackableQueuingDiscipline q);

   protected abstract boolean activeQListIsEmpty();

   protected abstract void addToActiveQList(SortableQDiscContainer sq);

   protected abstract void updateActiveQList(SortableQDiscContainer sq, double oldPos);

   protected abstract void advanceActiveQ();

   protected void updateLongestQList(SortableQDiscContainer sq, long oldLength) {
      long newLength = sq.getLengthInMetric();
      // remove
      // if not newly created, remove from oldPos
      if (oldLength != 0) {
         assert longestQList.containsKey(oldLength);
         // remove from old place in sortedQList
         longestQList.get(oldLength).remove(sq);
         if (longestQList.get(oldLength).size() == 0)
            longestQList.remove(oldLength);
      }
      // insert into new place
      if (newLength != 0) {
         if (longestQList.containsKey(newLength)) {
            longestQList.get(newLength).add(sq);
         } else {
            LinkedList<SortableQDiscContainer> newList = new LinkedList<SortableQDiscContainer>();
            newList.add(sq);
            longestQList.put(newLength, newList);
         }
      }
   }

   @Override
   public int getMaxSizeInUnits() {
      return maxSizeInUnits;
   }

   @Override
   public boolean isEmpty() {
      return activeQListIsEmpty();
   }

   @Override
   protected void push(QueueEntry qe) {
      // enqueue
      SortableQDiscContainer sq = queues.getConsumer(qe.getMessage());
      StackableQueuingDiscipline q;
      if (sq == null) { // new Q needed
         q = pars.get(simNode, "SubQueuingDiscipline" + numQ).asInstance(StackableQueuingDiscipline.class, simNode.createChildNode("SubQueuingDiscipline" + numQ), pars);
         sq = createSortableQDisc(q);
         numQ++;
         queues.insertConsumer(qe.getMessage(), sq);
      } else {
         q = sq.getQDisc();
      }
      if (q.isEmpty()) {
         addToActiveQList(sq);
      }
      double oldPos = sq.getsortIndex();
      long oldLength = sq.getLengthInMetric();
      q.enqueue(qe.getMessage());
      updateActiveQList(sq, oldPos);
      updateLongestQList(sq, oldLength);

      bufferStealing();
   }

   /**
    * Discard if necessary to obey the maximum size. This implements buffer
    * stealing as introduced in McKenney - "Stochastic Fair Queuing"
    * http://ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=91316
    */
   protected void bufferStealing() {
      while (getCurrentNumberOfUnits() > getMaxSizeInUnits()) {
         SortableQDiscContainer longestQ = longestQList.lastEntry().getValue().getFirst();
         double oldIndexLongestQ = longestQ.getsortIndex();
         long oldLengthLongestQ = longestQ.getLengthInMetric();
         QueueEntry dropQE = longestQ.getQDisc().dropScheduledNext();
         handleLossMessage(dropQE);
         updateActiveQList(longestQ, oldIndexLongestQ);
         updateLongestQList(longestQ, oldLengthLongestQ);
      }
   }

   @Override
   protected QueueEntry peekInternal() {
      if (activeQ == null || activeQ.getQDisc().isEmpty()) {
         throw new SimulationException("Get Message from empty queue");
      }
      return activeQ.getQDisc().peek();
   }

   @Override
   protected QueueEntry pop() {
      if (activeQ == null)
         throw new SimulationException("Trying to get message from activeQ that is null");
      if (activeQ.getQDisc().isEmpty())
         throw new SimulationException("Get Message from empty queue");
      // actually remove the qe
      double oldPos = activeQ.getsortIndex();
      long oldLength = activeQ.getLengthInMetric();
      QueueEntry qe = activeQ.getQDisc().dequeue();
      // ...otherwise delete it from the active list
      // Do not remove from queues to allow continuous statistics
      updateActiveQList(activeQ, oldPos);
      updateLongestQList(activeQ, oldLength);
      // advance the round robin pointer
      advanceActiveQ();
      if (getCurrentNumberOfUnits() < 0) {
         throw new SimulationException("Actual number of units in the queue is less than 0");
      }
      return qe;
   }
}
