/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import java.util.LinkedList;

import ikr.simlib.algorithms.MessageClassificator;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * This class implements fair queuing by messages. This means
 * <ul>
 * <li>
 * it needs a {@link MessageClassificator} to aggregate incoming {@link Message}
 * s</li>
 * <li>
 * it maintains one {@link StackableQueuingDiscipline} per class of
 * {@link Message}s, as identified by the {@link MessageClassificator}</li>
 * <li>
 * the fairness is per-message: if one class consists of 60 Byte packets while
 * the other if of 1500 Byte messages, both will get same packet rate but very
 * different bitrates (1:25 in average)</li>
 * <li>
 * it does no stochastic aggregating of messages that belong to different
 * aggregates according to the {@link MessageClassificator}. This can be
 * achieved by using a {@link MessageClassificator}, that uses a limited number
 * of classes as for the usual Stochastic Fair Queuing (perturbation would have
 * to be implemented in the {@link MessageClassificator}, too).</li>
 * </ul>
 * 
 * @author dwagner
 * 
 */
public class MessageFairQueuing extends FairQueuing {
   class NumberOfMessagesStackableQueuingDiscipline extends SortableQDiscContainer {
      public NumberOfMessagesStackableQueuingDiscipline(StackableQueuingDiscipline q) {
         qdisc = q;
      }

      @Override
      public double getsortIndex() {
         return qdisc.getCurrentNumberOfMessages();
      }

      @Override
      public long getLengthInMetric() {
         return qdisc.getCurrentNumberOfMessages();
      }
   }

   protected LinkedList<SortableQDiscContainer> activeQs = new LinkedList<SortableQDiscContainer>();
   protected int activeQIndex = 0;

   /** as required by {@link ReflectionConstructable} */
   public static MessageFairQueuing createInstance(SimNode ownNode, Parameters pars) {
      return new MessageFairQueuing(ownNode, pars);
   }

   public MessageFairQueuing(SimNode ownNode, Parameters pars) {
      super(ownNode, pars);
   }

   @Override
   public boolean wouldStore(Message msg) {
      // simplified, but holds if all messages are of the same size
      if (getCurrentNumberOfUnits() + msg.getLength() > getMaxSizeInUnits() && longestQList.lastEntry().getValue().contains(queues.getConsumer(msg))) {
         return false;
      } else {
         return true;
      }
   }

   @Override
   public SortableQDiscContainer createSortableQDisc(StackableQueuingDiscipline q) {
      return new NumberOfMessagesStackableQueuingDiscipline(q);
   }

   @Override
   protected boolean activeQListIsEmpty() {
      return activeQs.isEmpty();
   }

   @Override
   protected void addToActiveQList(SortableQDiscContainer sq) {
      activeQs.add(sq);
      if (activeQ == null) {
         activeQ = sq;
         assert (activeQIndex == 0); // shouldn't be neccessary
      }
   }

   @Override
   protected void updateActiveQList(SortableQDiscContainer sq, double oldPos) {
      // MessageFairness is just round robin, so the list itself is normally not
      // touched on pushs and pops.
      // It just needs to remove sq from activeQs if sq is empty
      if (sq.getQDisc().getCurrentNumberOfMessages() == 0) {
         if (!activeQs.get(activeQIndex).equals(activeQ))
            throw new RuntimeException("activeQ does not point to activeQIndex!");
         activeQs.remove(activeQIndex);
      }
   }

   @Override
   protected void advanceActiveQ() {
      if (activeQs.size() > 0) {
         activeQIndex = (activeQIndex + 1) % activeQs.size();
         activeQ = activeQs.get(activeQIndex);
      } else {
         activeQIndex = 0;
         activeQ = null;
      }
   }
}
