/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;

public interface QueuingDiscipline extends ReflectionConstructable2<SimNode, Parameters> {
   public static class QueueEntry {

      private Message message;
      private PointInTime arrivalTime;

      public QueueEntry(Message m) {
         message = m;
         this.arrivalTime = Calendar.getInstance().getSystemTime();
      }

      public Message getMessage() {
         return message;
      }

      public PointInTime getArrivalTime() {
         return arrivalTime;
      }

      @Override
      public boolean equals(Object obj) {
         if (obj == null) {
            return false;
         }

         if (obj.getClass() != getClass()) {
            return false;
         }

         QueueEntry rhs = (QueueEntry) obj;

         return (message == rhs.message && arrivalTime.equals(rhs.arrivalTime));
      }

      @Override
      public int hashCode() {
         return arrivalTime.hashCode();
      }

   }

   // get configuration information
   /**
    * @return maximum size of queue or <em>Integer.MAX_VALUE</em> if queue is
    *         unbounded.
    */
   public abstract int getMaxSizeInUnits();

   public String getName();

   public SimNode getNode();

   public SimNode getParent();

   public abstract boolean isEmpty();

   /**
    * @return true if the queue cannot is not willing to store the given message
    *         due to its algorithm or constraints (e.g. bounded buffer size).
    */
   public abstract boolean wouldStore(Message msg);

   public int getCurrentNumberOfMessages();

   public long getCurrentNumberOfUnits();

   /**
    * Dequeue the next {@link QueueEntry} according to the
    * {@link QueuingDiscipline}. This will be accounted for as successful
    * forwarding.
    * 
    * @return The {@link QueueEntry}
    */
   public QueueEntry dequeue();

   /**
    * Request the {@link QueuingDiscipline} to handle the {@link Message}. This
    * can either store it, then it returns true, or drop it, then it return
    * false. If it drops it, this drop will be accounted for (as drop with 0
    * waiting time).
    * 
    * @param msg
    * @return if it has accepted the message.
    */
   public boolean handleMessage(Message msg);

   /**
    * This method returns the {@link QueueEntry} that would be delivered next,
    * but accounts it as drop.
    * 
    * @return The dropped {@link QueueEntry}
    */
   public QueueEntry dropScheduledNext();
}
