/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.control.simulationcontrol;

import ikr.simlib.meters.count.MessageCountMeter;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * ControlCounter can be used to count any sort of event by calling the method
 * <em>Count()</em>. Once a certain number of events has been reached, the
 * simulation control is notified of the end of the current batch or the end of
 * the transient phase. Using this object, a control of the simulation's batch
 * duration based on the number of certain events can be achieved.
 * Alternatively, the batch duration can be timer controlled using the class
 * ControlTimer.
 * <p>
 * This class implements the {@link UnweightedSampleCollector} interface, so that it
 * can be attached to a meter like a statistic would be attached. It then counts
 * how often the {@link #update(Void)} method is called. This class would
 * typically be used together with a {@link MessageCountMeter}.
 * <p>
 * You can use asInstance() enforced by {@link ReflectionConstructable} to parse
 * and create objects of type ControlCounter and ControlCountMeter.
 */
public class ControlCounter extends SimNotifier implements UnweightedSampleCollector<Void> {

   public static final String TRANSIENT_CALLS = "TransientCalls";
   public static final String BATCH_CALLS = "BatchCalls";

   /** as required by {@link ReflectionConstructable} */
   public static ControlCounter createInstance(SimNode ownNode, Parameters pars) {
      final int transientCalls = pars.get(ownNode, TRANSIENT_CALLS).asInteger();
      final int batchCalls = pars.get(ownNode, BATCH_CALLS).asInteger();
      return new ControlCounter(transientCalls, batchCalls);
   }

   protected static enum State {
      TRANSIENT, BATCH
   }

   protected State state;
   protected int currentBatch;
   protected long transientCount;
   protected long batchCount;
   protected long limit;
   protected long count;

   protected void check() {
      if (count >= limit) {
         if (state == State.TRANSIENT) {
            endOfTransientPhase();
         } else if (state == State.BATCH) {
            endOfBatch(currentBatch);
         }
      }
   }

   public ControlCounter(int transientCount, int batchCount) {
      state = State.TRANSIENT;
      currentBatch = 0;
      this.transientCount = transientCount;
      this.batchCount = batchCount;
      limit = 0;
      count = 0;
   }

   public void count() {
      count++;

      check();
   }

   @Override
   public void update(Void value) {
      count();
   }

   @Override
   public void handleStartBatch(int batchNumber) {
      state = State.BATCH;
      limit = batchCount;
      currentBatch = batchNumber;
      count = 0;

      check();
   }

   @Override
   public void handleStartTransientPhase() {
      state = State.TRANSIENT;
      currentBatch = 0;
      limit = transientCount;
      count = 0;

      check();
   }
}
