/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.control.simulationcontrol;

import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * <code>ControlTimer</code> is an class which is able to look upon the system
 * time and calls the
 * 
 * <ul>
 * <li><code>EndOfTransientPhase</code></li>
 * <li><code>EndOfBatch</code></li>
 * </ul>
 * 
 * <p>
 * member functions of the <code>SimNotifier</code> to indicate that the desired
 * times of the transient and batch phase are reached.
 * </p>
 * 
 * <p>
 * The Constructor needs
 * </p>
 * 
 * <ul>
 * <li>duration of the transient phase</li>
 * <li>duration of the batch</li>
 * </ul>
 * 
 * You can use asInstance() enforced by {@link ReflectionConstructable} to parse
 * and create objects of type <code>ControlTimer</code>.
 */
public class ControlTimer extends SimNotifier {

   private static final String TRANSIENT_TIME = "TransientTime";
   private static final String BATCH_TIME = "BatchTime";

   /** as required by {@link ReflectionConstructable} */
   public static ControlTimer createInstance(SimNode ownNode, Parameters pars) {
      final Duration transientTime = pars.get(ownNode, TRANSIENT_TIME).asInstance(Duration.class, ownNode.createChildNode(TRANSIENT_TIME), pars);
      final Duration batchTime = pars.get(ownNode, BATCH_TIME).asInstance(Duration.class, ownNode.createChildNode(TRANSIENT_TIME), pars);
      return new ControlTimer(transientTime, batchTime);
   }

   private final Duration transientPhaseDuration;
   private final Duration batchPhaseDuration;

   public ControlTimer(Duration transientPhaseDuration, Duration batchPhaseDuration) {
      super();
      this.transientPhaseDuration = transientPhaseDuration;
      this.batchPhaseDuration = batchPhaseDuration;
   }

   @Override
   public void handleStartTransientPhase() {
      Calendar.getInstance().postEvent(new CalendarCallback() {
         @Override
         public void execute() {
            ControlTimer.this.endOfTransientPhase();
         }
      }, null, Calendar.getInstance().getSystemTime().plus(transientPhaseDuration));
   }

   @Override
   public void handleStartBatch(final int batchNumber) {
      Calendar.getInstance().postEvent(new CalendarCallback() {
         @Override
         public void execute() {
            ControlTimer.this.endOfBatch(batchNumber);
         }
      }, null, Calendar.getInstance().getSystemTime().plus(batchPhaseDuration));
   }
}
