/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.control.simulationcontrol;

import ikr.simlib.control.Simulation;
import ikr.simlib.control.notification.Notifier;
import ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager;
import ikr.simlib.control.simulationcontrol.callbacks.StartBatchCallbackI;
import ikr.simlib.control.simulationcontrol.callbacks.StartSimulationCallbackI;
import ikr.simlib.control.simulationcontrol.callbacks.StartTransientPhaseCallbackI;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;

/**
 * SimNotifier is a special notifier support class that has notifiers for all
 * phases of a simulation. The Run method of Simulation calls several methods
 * for the different phases of a simulation program. Derived classes can
 * override these methods to do additional work.
 */
public abstract class SimNotifier implements ReflectionConstructable2<SimNode, Parameters> {

   protected Notifier endOfTransientPhaseNotifier;

   protected Notifier endOfBatchNotifier;

   public SimNotifier() {
      endOfTransientPhaseNotifier = new Notifier();
      endOfBatchNotifier = new Notifier();

      SimulationControlManager.getInstance().registerStartSimulationCallback(new StartSimulationCallbackI() {
         @Override
         public void handleStartSimulation(Simulation sim) {
            SimNotifier.this.handleStartSimulation(sim);
         }
      });

      SimulationControlManager.getInstance().registerStartTransientPhaseCallback(new StartTransientPhaseCallbackI() {
         @Override
         public void handleStartTransientPhase() {
            SimNotifier.this.handleStartTransientPhase();
         }
      });

      SimulationControlManager.getInstance().registerStartBatchCallback(new StartBatchCallbackI() {
         @Override
         public void handleStartBatch(int batchNumber) {
            SimNotifier.this.handleStartBatch(batchNumber);
         }
      });
   }

   public void handleStartSimulation(Simulation sim) {
      registerSimNotifier(sim);
   }

   public abstract void handleStartTransientPhase();

   public abstract void handleStartBatch(int batchNumber);

   public void endOfTransientPhase() {
      endOfTransientPhaseNotifier.notifyHandler();
   }

   public void endOfBatch(int batchNr) {
      endOfBatchNotifier.notifyHandler();
   }

   public void registerSimNotifier(Simulation s) {
      // only registers notifier if not yet registered
      if (!endOfTransientPhaseNotifier.isRegistered())
         s.getTransientPhaseNotifier().addNotifier(endOfTransientPhaseNotifier);

      if (!endOfBatchNotifier.isRegistered())
         s.getBatchNotifier().addNotifier(endOfBatchNotifier);
   }

   public void deregisterSimNotifier() {
      if (endOfTransientPhaseNotifier.getOwner() != null)
         endOfTransientPhaseNotifier.getOwner().removeNotifier(endOfTransientPhaseNotifier);

      if (endOfBatchNotifier.getOwner() != null)
         endOfBatchNotifier.getOwner().removeNotifier(endOfBatchNotifier);
   }
}
