/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.control.simulationcontrol.callbacks;

import ikr.simlib.control.Simulation;

import java.util.ArrayList;
import java.util.List;

/**
 * All objects of class SimulationControl must register at the global instance
 * of SimulationControlManager.
 * 
 * The latter one is informed of any phase changes by {@link Simulation} (or
 * derived classes).
 */
public class SimulationControlManager {

   protected static SimulationControlManager INSTANCE;

   private List<InitSimulationCallbackI> initSimulationCallbacks;
   private List<StartTransientPhaseCallbackI> startTransientPhaseCallbacks;
   private List<StopTransientPhaseCallbackI> stopTransientPhaseCallbacks;
   private List<StartBatchCallbackI> startBatchCallbacks;
   private List<StopBatchCallbackI> stopBatchCallbacks;
   private List<StartSimulationCallbackI> startSimulationCallbacks;
   private List<StopSimulationCallbackI> stopSimulationCallbacks;

   static {
      INSTANCE = null;
   }

   public static SimulationControlManager getInstance() {
      if (INSTANCE == null)
         INSTANCE = new SimulationControlManager();
      return INSTANCE;
   }

   public SimulationControlManager() {
      initSimulationCallbacks = new ArrayList<InitSimulationCallbackI>();
      startTransientPhaseCallbacks = new ArrayList<StartTransientPhaseCallbackI>();
      stopTransientPhaseCallbacks = new ArrayList<StopTransientPhaseCallbackI>();
      startBatchCallbacks = new ArrayList<StartBatchCallbackI>();
      stopBatchCallbacks = new ArrayList<StopBatchCallbackI>();
      startSimulationCallbacks = new ArrayList<StartSimulationCallbackI>();
      stopSimulationCallbacks = new ArrayList<StopSimulationCallbackI>();
   }

   public void registerInitSimulationCallback(InitSimulationCallbackI handler) {
      initSimulationCallbacks.add(handler);
   }

   public void registerStartTransientPhaseCallback(StartTransientPhaseCallbackI handler) {
      startTransientPhaseCallbacks.add(handler);
   }

   public void registerStopTransientPhaseCallback(StopTransientPhaseCallbackI handler) {
      stopTransientPhaseCallbacks.add(handler);
   }

   public void registerStartBatchCallback(StartBatchCallbackI handler) {
      startBatchCallbacks.add(handler);
   }

   public void registerStopBatchCallback(StopBatchCallbackI handler) {
      stopBatchCallbacks.add(handler);
   }

   public void registerStartSimulationCallback(StartSimulationCallbackI handler) {
      startSimulationCallbacks.add(handler);
   }

   public void registerStopSimulationCallback(StopSimulationCallbackI handler) {
      stopSimulationCallbacks.add(handler);
   }

   public void initSimulation() {
      // use index iteration to allow to append new callbacks
      for (int i = 0; i < initSimulationCallbacks.size(); i++) {
         initSimulationCallbacks.get(i).handleInitSimulation();
      }
   }

   public void startSimulation(Simulation sim) {
      // use index iteration to allow to append new callbacks
      for (int i = 0; i < startSimulationCallbacks.size(); i++) {
         startSimulationCallbacks.get(i).handleStartSimulation(sim);
      }
   }

   public void stopSimulation() {
      // use index iteration to allow to append new callbacks
      for (int i = 0; i < stopSimulationCallbacks.size(); i++) {
         stopSimulationCallbacks.get(i).handleStopSimulation();
      }
   }

   public void startTransientPhase() {
      // use index iteration to allow to append new callbacks
      for (int i = 0; i < startTransientPhaseCallbacks.size(); i++) {
         startTransientPhaseCallbacks.get(i).handleStartTransientPhase();
      }
   }

   public void stopTransientPhase() {
      // use index iteration to allow to append new callbacks
      for (int i = 0; i < stopTransientPhaseCallbacks.size(); i++) {
         stopTransientPhaseCallbacks.get(i).handleStopTransientPhase();
      }
   }

   public void startBatch(int batchNumber) {
      // use index iteration to allow to append new callbacks
      for (int i = 0; i < startBatchCallbacks.size(); i++) {
         startBatchCallbacks.get(i).handleStartBatch(batchNumber);
      }
   }

   public void stopBatch(int batchNumber) {
      // use index iteration to allow to append new callbacks
      for (int i = 0; i < stopBatchCallbacks.size(); i++) {
         stopBatchCallbacks.get(i).handleStopBatch(batchNumber);
      }
   }
}
