/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionVector;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

import java.util.List;

/**
 * <code>EmpiricalDistribution</code> allows to read empirical distributions
 * obtained e.g.from measurements into a simulation program.
 * 
 * <p>
 * This is useful if no mathematical fit can be found to describe a distribution
 * or if a measured distribution has to be integrated into a simulation very
 * fast.
 * </p>
 * 
 * <p>
 * <code>EmpiricalDistribution</code> is a continuous distribution and requires
 * 2-tuples as input data: value and cdf-probability. This allows to have
 * arbitrarily spaced data.
 * </p>
 * 
 * <p>
 * Similar to <code>GeneralDistribution</code> it relies on the
 * <code>DistributionVector</code>. This reads the 2-tuple-vector and generates
 * 2 separate vectors for values and probabilities. Using the same search
 * algorithm as <code>ProbabilityVector</code> it determines the upper bound of
 * the determined bin and then does a linear interpolation between that bin and
 * the next lower bin (if any) to determine the fraction of the bin the search
 * value points to. The return value is determined by calculating that same
 * fraction between the corresponding vector elements of the value vector.
 * </p>
 * 
 * <pre>
 *      |                                           x
 *      |                               x       
 *  RND&gt;| -  -  -  -  -  -  -  -  -  . `                                  
 *      |                         . `| 
 *      |               x   x   x    |               
 *      |       x                    |            
 *      ----x---+---+---+---+---+---+---+---+---+---+--&gt;
 *                                   v
 *                                  RV
 * </pre>
 * 
 * <p>
 * The simulation par-file should look like the following:
 * </p>
 * 
 * <pre>
 *  EmpiricalDist {
 *    CDF = {
 *      0.000 ,   0.00000000 ,
 *      1.000 ,   0.05000000 ,
 *      2.000 ,   0.10000000 ,
 *      10.000 ,  0.25000000 ,
 *      20.000 ,  0.30000000 ,
 *      30.000 ,  0.50000000 ,
 *      40.000 ,  0.60000000 ,
 *      50.000 ,  0.80000000 ,
 *      100.000 , 0.95000000 ,
 *      110.000 , 1.00000000
 *    };
 *  }
 * </pre>
 * 
 * <p>
 * Notes:
 * </p>
 * 
 * <ul>
 * <li>It is best always give the final value of 1.0.</li>
 * <li>The first entry <em>must</em> be <code>(x|0.0)</code>, where
 * <code>x</code> is an arbitrary double.</li>
 * <li>Avoid identical values in the last 2 entries.</li>
 * </ul>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>distribution function is passed to constructor as vector or parsed from
 * file <br>
 * c.f. detailed comments in header file description</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>distribution function</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>passed to constructor as vector or parsed from file</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>depends on DF</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>depends on DF</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>depends on DF</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.continuous.EmpiricalDistribution
         <br>
         # F(x0) must be 0.0, x0 arbitrary
         <br>
         # F(xn) must be 1.000, xn > x0
         <br>
         [...].Distribution.CDF = [0.000 0.01000000 1.000 0.05000000 70.000 0.95000000 110.000 1.00000000 ]</code>
 * </td>
 * </tr>
 * </table>
 * 
 * @see GeneralDistribution
 */
public class EmpiricalDistribution extends ContinuousDistribution {

   private static final String CDF = "CDF";

   /** as required by {@link ReflectionConstructable} */
   public static EmpiricalDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      double[] vector = pars.get(ownNode, CDF).asDoubleVector();
      return new EmpiricalDistribution(vector, rng);
   }

   private final DistributionVector vector;

   public EmpiricalDistribution(double[] vector, RandomNumberGenerator rng) {
      super(rng);
      this.vector = new DistributionVector(vector);
   }

   public EmpiricalDistribution(double[] vector) {
      this(vector, getDefaultRNG());
   }

   public EmpiricalDistribution(List<Double> vector, RandomNumberGenerator rng) {
      super(rng);
      this.vector = new DistributionVector(vector);
   }

   public EmpiricalDistribution(List<Double> vector) {
      this(vector, getDefaultRNG());
   }

   @Override
   public double next() {
      return vector.next(rng.next());
   }
}
