/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Gamma Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>The gamma distribution may, e.g., be applied to characterize video
 * traffic. Special cases contained are the negative-exponential and Erlang k
 * distribution \((\alpha = 1 \mbox{ or } \alpha = k)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * shape parameter \(\alpha > 0\)
 * <li>
 * scale parameter \(\beta > 0\)
 * </ul>
 * The distribution can be alternatively described with the parameters mean
 * value and coefficient of variation:
 * <ul>
 * <li>
 * \(\alpha = \frac{1}{c_T^2}\)
 * <li>
 * \(\beta = E[T] \cdot c_T^2\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \frac{\beta^{-\alpha} \cdot t^{\alpha-1} \cdot
 * exp(-\frac{t}{\beta})}{ \Gamma(\alpha)} \mbox{ for } t>0\) <br>
 * <br>
 * whereby \(\Gamma(x)\) is the gamma function</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>exists only if \(\alpha\) is an integer number and positive (see Erlang k
 * distribution)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \alpha\beta\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \alpha\beta^2\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \frac{1}{\sqrt{\alpha}}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.GammaDistribution
         <br>
         [...].Distribution.Alpha = 0.44
         <br>
         [...].Distribution.Beta = 10.13
      </code> <br>
 * <br>
 * or with the mean value and variation coefficient: <br>
 * <br>
 * <code>
         [...].Distribution.Mean = 4.5
         <br>
         [...].Distribution.CoefficientOfVariation = 1.5
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>
 * Law/Kelton fourth edition, pp. 284-286
 * </td>
 * </tr>
 * </table>
 * 
 * 
 */
public class GammaDistribution extends ContinuousDistribution {

   private static final String ALPHA = "Alpha";
   private static final String BETA = "Beta";
   private static final String COEFF_OF_VARIATION = "CoefficientOfVariation";
   private static final String MEAN = "Mean";

   /** as required by {@link ReflectionConstructable} */
   public static GammaDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {

      double alpha;
      if (pars.hasParameter(ownNode, ALPHA)) {
         alpha = pars.get(ownNode, ALPHA).asDouble();
      } else {
         double coefficientOfVariation = pars.get(ownNode, COEFF_OF_VARIATION).asDouble();
         alpha = 1 / (coefficientOfVariation * coefficientOfVariation);
      }

      double beta;
      if (pars.hasParameter(ownNode, BETA)) {
         beta = pars.get(ownNode, BETA).asDouble();
      } else {
         double mean = pars.get(ownNode, MEAN).asDouble();
         beta = mean / alpha;
      }

      return new GammaDistribution(alpha, beta, rng);
   }

   public final double alpha;
   public final double beta;

   public final double A;
   public final double B;
   public final double D;
   public final double Q;
   public final double theta;

   public GammaDistribution(double alpha, double beta, RandomNumberGenerator rng) {
      super(rng);

      if (alpha <= 0) {
         throw new DistributionException("alpha must be > 0");
      }

      if (beta <= 0) {
         throw new DistributionException("beta must be > 0");
      }

      this.alpha = alpha;
      this.beta = beta;

      if (alpha < 1) {
         B = 1 + alpha * Math.exp(-1.0);
         A = 0;
         D = 0;
         Q = 0;
         theta = 0;
      } else if (alpha > 1) {
         A = 1 / Math.sqrt(2 * alpha - 1);
         B = alpha - Math.log(4.0);
         Q = alpha + 1 / A;
         theta = 4.5;
         D = 1 + Math.log(theta);
      } else {
         A = 0;
         B = 0;
         D = 0;
         Q = 0;
         theta = 0;
      }
   }

   public GammaDistribution(double alpha, double beta) {
      this(alpha, beta, getDefaultRNG());
   }

   @Override
   public double next() {
      double y;

      if (alpha < 1) {
         double p;
         boolean valueFound = false;
         do {
            p = B * rng.next();
            if (p > 1) {
               y = -Math.log((B - p) / alpha);
               if (rng.next() <= Math.pow(y, alpha - 1)) {
                  valueFound = true;
               }
            } else {
               y = Math.pow(p, 1 / alpha);
               if (rng.next() <= Math.exp(-y)) {
                  valueFound = true;
               }
            }
         } while (!valueFound);
      } else if (alpha == 1) {
         y = -Math.log(rng.next());
      } else {
         double u1;
         double u2;
         double v;
         double w;
         double z;
         do {
            u1 = rng.next();
            u2 = rng.next();
            v = A * Math.log(u1 / (1 - u1));
            y = alpha * Math.exp(v);
            z = u1 * u1 * u2;
            w = B + Q * v - y;
         } while ((w + D - theta * z < 0) && (w < Math.log(z)));
      }

      return beta * y;
   }
}
