/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Negative Exponential Distributione<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Time distance between two events (e.g., arrival, process end,...) in a
 * Markovian process with the average rate \(\lambda\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>mean value \(m\) or rate \(\lambda = \frac{1}{m}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \lambda \cdot exp(-\lambda t) = \frac{1}{m} \cdot
 * exp(-\frac{t}{m}) \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = 1-exp(-\lambda t) = 1-exp(-\frac{1}{m}) \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \frac{1}{\lambda} = m \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \frac{1}{\lambda^2} = m^2 \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= 1\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = \frac{\lambda}{\lambda + s} = \frac{1}{1 + ms}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.continuous.NegExpDistribution
         <br>
         [...].Distribution.Mean = 3.6</code></td>
 * </tr>
 * </table>
 * 
 */
public class NegExpDistribution extends ContinuousDistribution {

   private static final String MEAN = "Mean";

   /** as required by {@link ReflectionConstructable} */
   public static NegExpDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      return new NegExpDistribution(mean, rng);
   }

   public final double mean;

   public NegExpDistribution(double mean, RandomNumberGenerator rng) {
      super(rng);

      if (mean < 0) {
         throw new DistributionException("Mean < 0!");
      }

      this.mean = mean;
   }

   public NegExpDistribution(double mean) {
      this(mean, getDefaultRNG());
   }

   @Override
   public double next() {
      return -mean * Math.log(rng.next());
   }
}
