/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Rayleigh Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>If the values of a two dimensional probability vector are independently normal distributed, then the absolute value of this vector is rayleigh distributed</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>\(\sigma\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \begin{cases} \frac{x}{\sigma^2}\exp{-\frac{x^2}{2\sigma^2}} &\mbox{for } x \ge 0 \\ 0 &\mbox{else } \end{cases}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = \begin{cases} 1 - \exp{-\frac{x^2}{2\sigma^2}} &\mbox{for } x \ge 0 \\ 0 &\mbox{else } \end{cases}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \sigma \sqrt{\frac{\pi}{2}}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \frac{(4-\pi}{2}\sigma^2\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \sqrt{\frac{4-\pi}{2}}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.RayleighDistribution
         <br>
         [...].Distribution.Sigma = 1.5
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class RayleighDistribution extends ContinuousDistribution {
   private final UniformDistribution baseDist;
   private final double sigma;

   private static final String SIGMA = "Sigma";

   /** as required by {@link ReflectionConstructable} */
   public static RayleighDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double sigma = pars.get(ownNode, SIGMA).asDouble();
      return new RayleighDistribution(sigma, rng);
   }

   public RayleighDistribution(double sigma, RandomNumberGenerator rng) {
      if (sigma <= 0) {
         throw new DistributionException("sigma is <= 0!");
      }
      this.sigma = sigma;
      this.baseDist = new UniformDistribution(1, 0, rng);
   }

   @Override
   public double next() {
      return this.sigma * Math.sqrt(-2.0 * Math.log(this.baseDist.next()));
   }
}