/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.distributions.discrete.ShiftedGeometricDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * MMDP (Markov Modulated Deterministic Process)<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Most frequent special case of the GMDP (
 * {@link ikr.simlib.distributions.continuous.statedep.GMDPDistribution}) for
 * modeling sources with multiple states (generally 2 - 5) in ATM networks</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>Like the GMDP(
 * {@link ikr.simlib.distributions.continuous.statedep.GMDPDistribution}), the
 * MMDP can be described by a state machine with \(m\) states. Number of
 * arrivals \(X_i\) in the state \(i\) is distributed according to a shifted
 * geometric distribution.</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>Moments and distribution of the arrival intervals</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * number of states \(m\)
 * <li>
 * transition probabilities \(p_{ii}\)
 * <li>
 * average number of arrivals \(a_i\) in the individual states
 * <li>
 * arrival intervals \(d_i\) in the individual states
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Dist = ikr.simlib.distributions.continuous.statedep.MMDPDistribution
         <br>
         [...].Dist.States = 3
         <br>
         [...].Dist.PMAP = [[0 0.5 0.5] [0.5 0 0.5] [0.5 0.5 0] ]
         <br>
         [...].Dist.MeanPhaseLengths = [145.5 11.11 55.55]
         <br>
         # negative value denotes silence state
         <br>
         [...].Dist.InterCellTimes = [-1 1274 255]
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>T. THEIMER: How to compute the moments of a GMDP, RACE 1022, Document
 * UST_123_0023_CD_CC, 1989.</td>
 * </tr>
 * </table>
 * 
 */
public class MMDPDistribution extends GMDPDistribution {

   private static final String STATES = "States";
   private static final String INTER_CELL_TIMES = "InterCellTimes";
   private static final String MEAN_PHASE_LENGTHS = "MeanPhaseLengths";
   private static final String PMAP = "PMAP";

   /** as required by {@link ReflectionConstructable} */
   public static MMDPDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final int noOfStates = pars.get(ownNode, STATES).asInteger();
      final double[] interCellTimes = pars.get(ownNode, INTER_CELL_TIMES).asDoubleVector();
      final double[] meanPhaseLengths = pars.get(ownNode, MEAN_PHASE_LENGTHS).asDoubleVector();
      final double[][] transitionProbs = pars.get(ownNode, PMAP).asDoubleMatrix();
      return new MMDPDistribution(noOfStates, interCellTimes, meanPhaseLengths, transitionProbs, rng);
   }

   public MMDPDistribution(int noOfStates, double[] interCellTimes, double[] meanPhaseLengths, double[][] transitionProbs, RandomNumberGenerator rng) {
      super(noOfStates, rng);

      for (int i = 0; i < noOfStates; i++) {
         stateParameters[i] = new GMDPStateParameters(interCellTimes[i], new ShiftedGeometricDistribution(meanPhaseLengths[i]), transitionProbs[i]);
      }

      reset();
   }

   public MMDPDistribution(int noOfStates, double[] interCellTimes, double[] meanPhaseLengths, double[][] transitionProbs) {
      this(noOfStates, interCellTimes, meanPhaseLengths, transitionProbs, getDefaultRNG());
   }
}
