/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;
import ikr.simlib.support.Matrices;

/**
 * MMPP (Markov Modulated Poisson Process)<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Modeling of sources with multiple states e.g., at the call level</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>State machine with \(m\) states.
 * <ul>
 * <li>
 * Markovian process with the rate \(\lambda_i\) in the state \(i\)
 * <li>
 * Markovian process for modulation with the transition rate \(q_{ij}\) from
 * state \(i\) to state \(j\)
 * <li>
 * According to definition: \(q_{ii} = -\sum\limits_{i \neq i} = q_{ij}\).
 * <li>
 * Special case of the MAP with \(d_{ii} = \lambda_i\), \(d_{ij} = 0 \forall (i
 * \neq j)\) and \(i_{ij} = q_{ij} - d_{ij}\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * number of states \(m\)
 * <li>
 * arrival rates \(\lambda_i\) in the states
 * <li>
 * transition rates \(q_{ij}\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>
 * <ul>
 * <li>
 * sojourn time \(S_i\) in the state \(i\) is negative exponentially distributed
 * with the mean value \(E[S_i] = \frac{1}{\sum\limits_{i \neq i} q_{ij}} =
 * \frac{-1}{q_{ii}}\)
 * <li>
 * sojourn probabilities \(P_i\) from linear equation system \(\sum\limits_i
 * q_{ji} \cdot P_j = 0 \forall i \mbox{ , } \sum\limits_i P_i = 1\)
 * <li>
 * average arrival rate: \(\lambda = \sum\limits_i P_i \cdot \lambda_i \)
 * </ul>
 * -</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].distribution = ikr.simlib.distributions.continuous.statedep.MMPPDistribution
         <br>
         [...].distribution.States = 2
         <br>
         [...].distribution.EventRates = [0.1 0.9]
         <br>
         [...].distribution.RMAP = [[-0.001 0.001] [0.001 -0.001]]
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>
 * <ul>
 * <li>
 * P. J. KHN: Reminder on queueing theory for ATM networks. Telecommunication
 * Systems, No. 5, 1996, pp. 1-24.
 * <li>
 * G. D. STAMOULIS, M. E. ANAGNOSTOU, A. D. GEORGANTAS: Traffic source models
 * for ATM networks: a survey. Computer Communications, Vol. 17, No. 6, Juni,
 * 1994.
 * <li>
 * H. HEFFES, D. M. LUCANTONI: A Markov modulated charcterization of packetized
 * voice and data traffic and related statistical multiplexer performance. IEEE
 * Journal on Selected Areas in Communications, Vol. SAC-4, No. 6, 1986, pp.
 * 856-868.
 * </ul>
 * </td>
 * </tr>
 * </table>
 * 
 */
public class MMPPDistribution extends MAPDistribution {

   private static final String STATES = "States";
   private static final String EVENT_RATES = "EventRates";
   private static final String RMAP = "RMAP";

   /** as required by {@link ReflectionConstructable} */
   public static MMPPDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final int noOfStates = pars.get(ownNode, STATES).asInteger();
      final double[] eventRates = pars.get(ownNode, EVENT_RATES).asDoubleVector();
      final double[][] transitionRates = pars.get(ownNode, RMAP).asDoubleMatrix();

      return new MMPPDistribution(noOfStates, eventRates, transitionRates, rng);
   }

   public MMPPDistribution(int noOfStates, double[] eventRates, double[][] transitionRates, RandomNumberGenerator rng) {
      super(noOfStates, Matrices.sub(transitionRates, Matrices.diagonal(eventRates)), Matrices.diagonal(eventRates), rng);
   }

   public MMPPDistribution(int noOfStates, double[] eventRates, double[][] transitionRates) {
      this(noOfStates, eventRates, transitionRates, getDefaultRNG());
   }
}
