/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Talkspurt Silence Source<br>
 * 
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>
 * <ul>
 * <li>
 * Modeling of on-off sources in packet networks
 * <li>
 * Modeling of voice sources
 * </ul>
 * <br>
 * <br>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>
 * <ul>
 * <li>
 * State machine with 2 states (on/off, talkspurt/silence or burst/silence)
 * <li>
 * In the talkspurt state \(X\) cells/packets arrive at intervals of \(d\).
 * After that a silence phase of the duration \(S\) takes place.
 * <li>
 * Special case of the GMDP, if \(S\) is described by a compound distribution
 * with a constant distribution as inner distribution.
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * (discrete) distribution of the number of cells \(X\) in the talkspurt state
 * <li>
 * (continuous) distribution of silence duration \(S\)
 * <li>
 * arrival interval \(d\) ("inter-cell time") in the talkspurt state
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>
 * <ul>
 * <li>
 * Peak rate: \(h = \frac{1}{d}\)
 * <li>
 * Mean rate: \(m = \frac{E[X]}{E[X] \cdot d + E[S]}\)
 * <li>
 * Burstiness: \(b = \frac{h}{m} = 1 +\frac{E[S]}{E[X] \cdot d}\)
 * <li>
 * For given \(h\) and \(m\) (or \(h\) and \(b\) as well as \(E[X]\)): <br>
 * \(h = \frac{1}{h}\), \(E[S] = E[X] \cdot (\frac{1}{m} - \frac{1}{h}) =
 * \frac{E[X]}{h} \cdot (b-1)\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Dist = ikr.simlib.distributions.continuous.statedep.TalkspurtSilenceDistribution
         <br>
         [...].Dist.TalkspurtDistribution = ShiftedGeometric
         <br>
         [...].Dist.TalkspurtDistribution.Mean = 20
         <br>
         [...].Dist.SilenceDistribution = NegExp
         <br>
         [...].Dist.SilenceDistribution.Mean = 800
         <br>
         [...].Dist.SilenceDistribution.InterCellTime = 10
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>
 * <ul>
 * <li>
 * P. J. KHN: Reminder on queueing theory for ATM networks. Telecommunication
 * Systems, No. 5, 1996, pp. 1-24.
 * <li>
 * G. D. STAMOULIS, M. E. ANAGNOSTOU, A. D. GEORGANTAS: Traffic source models
 * for ATM networks: a survey." Computer Communications, Vol. 17, No. 6, Juni,
 * 1994.
 * </ul>
 * </td>
 * </tr>
 * </table>
 * 
 */
public class TalkspurtSilenceDistribution extends ContinuousDistribution {

   public final DiscreteDistribution talkspurtDistribution;
   public final ContinuousDistribution silenceDistribution;
   public final double interCellTime;

   private int talkspurtCounter;

   private static final String TALKSPURT_DISTRIBUTION = "TalkspurtDistribution";
   private static final String SILENCE_DISTRIBUTION = "SilenceDistribution";
   private static final String INTER_CELL_TIME = "InterCellTime";
   private static final String INTER_PACKET_TIME = "InterPacketTime";

   /** as required by {@link ReflectionConstructable} */
   public static TalkspurtSilenceDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final DiscreteDistribution talkspurtDistribution = pars.get(ownNode, TALKSPURT_DISTRIBUTION).asInstance(DiscreteDistribution.class, ownNode.createChildNode(TALKSPURT_DISTRIBUTION), pars, rng);
      final ContinuousDistribution silenceDistribution = pars.get(ownNode, SILENCE_DISTRIBUTION).asInstance(ContinuousDistribution.class, ownNode.createChildNode(SILENCE_DISTRIBUTION), pars, rng);

      double interCellTime = Double.NaN;
      if (pars.hasParameter(ownNode, INTER_CELL_TIME)) {
         interCellTime = pars.get(ownNode, INTER_CELL_TIME).asDouble();
      } else if (pars.hasParameter(ownNode, INTER_PACKET_TIME)) {
         interCellTime = pars.get(ownNode, INTER_PACKET_TIME).asDouble();
      } else {
         throw new SimulationException("inter-cell time or inter-packet time must be specified");
      }

      return new TalkspurtSilenceDistribution(talkspurtDistribution, silenceDistribution, interCellTime, rng);
   }

   public TalkspurtSilenceDistribution(DiscreteDistribution talkspurtDist, ContinuousDistribution silenceDist, double interCellTime, RandomNumberGenerator rng) {
      super(rng);

      talkspurtDistribution = talkspurtDist;
      silenceDistribution = silenceDist;
      this.interCellTime = interCellTime;
      talkspurtCounter = 0;
   }

   public TalkspurtSilenceDistribution(DiscreteDistribution talkspurtDist, ContinuousDistribution silenceDist, double interCellTime) {
      this(talkspurtDist, silenceDist, interCellTime, getDefaultRNG());
   }

   @Override
   public double next() {
      double silenceTime = 0.0;

      while (talkspurtCounter == 0) {
         talkspurtCounter = talkspurtDistribution.next();
         silenceTime += silenceDistribution.next();
      }

      talkspurtCounter--;
      return interCellTime + silenceTime;
   }

   @Override
   public void reset() {
      talkspurtCounter = 0;
   }

   public int getTalkspurtCounter() {
      return talkspurtCounter;
   }
}
