/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.transform;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Continuous Bounded Distribution <br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Continuous distribution of a random varaible \(T\) that is bounded
 * between a lower bound and an upper bound. For the bounding of \(T\), two
 * different modes are available:
 * <ol>
 * <li>
 * <i>with resampling</i>: If the random number generator returns a value that
 * is smaller than the lower boundary or greater than the upper boundary, a new
 * random number is drawn (this is done as long as a value between lower and
 * upper boundary is obtained).
 * <li>
 * <i>without resampling</i>: If the random number generator returns a value
 * that is smaller than the lower boundary or greater than the upper boundary,
 * then the lower respectively upper boundary is return.
 * </ol>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * base distribution with PDF \(g(t)\) and DF \(G(t)\)
 * <li>
 * lower limit \(b_l\)
 * <li>
 * upper limit \(b_u\)
 * <li>
 * boolean value <code>
               resampling
            </code>
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.transform.BoundedDistribution
         <br>
         [...].Distribution.BaseDist = ikr.simlib.distributions.continuous.NormalDistribution
         <br>
         [...].Distribution.BaseDist.Mean = 5.0
         <br>
         [...].Distribution.BaseDist.StandardDeviation = 4.0
         <br>
         [...].Distribution.LowerBound = 0.0 # optional, default 0
         <br>
         [...].Distribution.UpperBound = 10.0 # optional, def. inf.
         <br>
         [...].Distribution.Resampling = false # optional, def. tr.
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class BoundedDistribution extends ContinuousDistribution {

   private static final String BASE_DISTRIBUTION = "BaseDist";
   private static final String LOWER_BOUND = "LowerBound";
   private static final String LOWER_BOUND_DEFAULT = "0";
   private static final String UPPER_BOUND = "UpperBound";
   private static final String UPPER_BOUND_DEFAULT = new Double(Double.MAX_VALUE).toString();
   private static final String RESAMPLING = "Resampling";
   private static final String RESAMPLING_DEFAULT = "true"; 

   /** as required by {@link ReflectionConstructable} */
   public static BoundedDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final ContinuousDistribution baseDist = pars.get(ownNode, BASE_DISTRIBUTION).asInstance(ContinuousDistribution.class, ownNode.createChildNode(BASE_DISTRIBUTION), pars, rng);
      final double lowerBound = pars.getOrUseDefault(ownNode, LOWER_BOUND, LOWER_BOUND_DEFAULT).asDouble();
      final double upperBound = pars.getOrUseDefault(ownNode, UPPER_BOUND, UPPER_BOUND_DEFAULT).asDouble();
      final boolean resampling = pars.getOrUseDefault(ownNode, RESAMPLING, RESAMPLING_DEFAULT).asBoolean();
      return new BoundedDistribution(baseDist, lowerBound, upperBound, resampling);
   }

   public final ContinuousDistribution baseDist;
   public final double lowerBound;
   public final double upperBound;
   public final boolean resampling;

   public BoundedDistribution(ContinuousDistribution baseDist, double lowerBound, double upperBound, boolean resampling) {
      super();

      this.baseDist = baseDist;
      this.lowerBound = lowerBound;
      this.upperBound = upperBound;
      this.resampling = resampling;
   }

   public BoundedDistribution(ContinuousDistribution baseDist, double lowerBound, double upperBound) {
      this(baseDist, lowerBound, upperBound, Boolean.parseBoolean(RESAMPLING_DEFAULT));
   }

   public BoundedDistribution(ContinuousDistribution baseDist, double lowerBound) {
      this(baseDist, lowerBound, Double.parseDouble(UPPER_BOUND_DEFAULT));
   }

   public BoundedDistribution(ContinuousDistribution baseDist) {
      this(baseDist, Double.parseDouble(LOWER_BOUND_DEFAULT));
   }

   @Override
   public double next() {
      double result;

      if (resampling) {
         do {
            result = baseDist.next();
         } while (result < lowerBound || result > upperBound);
      } else {
         result = baseDist.next();
         if (result < lowerBound) {
            result = lowerBound;
         } else if (result > upperBound) {
            result = upperBound;
         }
      }

      return result;
   }
}
