/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.transform;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Continuous Distribution with a Discrete Value Range<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>
 * <ul>
 * <li>
 * Distribution of a real random variable \(T = d \cdot N\), whereby \(N\)
 * represents an arbitrarily distributed discrete random variable and \(d\) the
 * scale factor.
 * <li>
 * Main application is the description of a cell distance in a time slot system
 * e.g., on an ATM link. In this case \(d\) is the time slot duration.
 * <li>
 * A special case of the compound distribution with a constant "inner
 * distribution".
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * scale factor/time slot duration \(d\)
 * <li>
 * discrete distribution \(p_i\) of \(N\) / the number of time slots
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \sum\limits_{n=0}^{\infty} p_n \cdot \delta (t-nd) \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T] = E[N] \cdot d\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= VAR[N] \cdot d^2\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T = c_N\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = H(exp(-sd))\) , if \(H(z)\) represents the generating
 * function of \(N\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.transform.SlottedDistribution
         <br>
         [...].Distribution.SlotDuration = 1.5
         <br>
         [...].Distribution.NoOfSlotsDist = ikr.simlib.distributions.discrete.GeometricDistribution
         <br>
         [...].Distribution.NoOfSlotsDist.Mean = 9
         <br>
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class SlottedDistribution extends ContinuousDistribution {

   private static final String SLOT_DURATION = "SlotDuration";
   private static final String NO_OF_SLOTS_DISTRIBUTION = "NoOfSlotsDist";

   /** as required by {@link ReflectionConstructable} */
   public static SlottedDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double slotDuration = pars.get(ownNode, SLOT_DURATION).asDouble();
      final DiscreteDistribution noOfSlotsDist = pars.get(ownNode, NO_OF_SLOTS_DISTRIBUTION).asInstance(DiscreteDistribution.class, ownNode.createChildNode(NO_OF_SLOTS_DISTRIBUTION), pars, rng);
      return new SlottedDistribution(noOfSlotsDist, slotDuration);
   }

   public final DiscreteDistribution noOfSlotsDist;
   public final double slotDuration;

   public SlottedDistribution(DiscreteDistribution noOfSlotsDist, double slotDuration) {
      super();
      this.noOfSlotsDist = noOfSlotsDist;
      this.slotDuration = slotDuration;
   }

   @Override
   public double next() {
      return slotDuration * noOfSlotsDist.next();
   }
}
