/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Binomial Distribution <br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Probability for \(i\) successes in \(n\) Bernoulli trials with the
 * parameter \(q(0 \le q \le 1)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * success probability \(q(0 < q \le 1\)
 * <li>
 * number of trials \(n > 0\)
 * </ul>
 * Alternative: mean \(E[X] \ge 0\) and variance \(VAR[X] (0 \le VAR[X] \le
 * E[X]\) parameters:
 * <ul>
 * <li>
 * \(q = 1- \frac{VAR[X]}{E[X]}\)
 * <li>
 * \(n = \frac{(E[X])^2}{E[X] - VAR[X]}\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = \binom{n}{i} \cdot q^i \cdot (1-q)^{n-i} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X]= nq \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]= nq(1-q)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \sqrt{ \frac{ 1-q }{nq} }\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Generating func.:</th>
 * <td>\(G(z)= (1-q+qz)^n \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.discrete.BinomialDistribution
         <br>
         [...].Distribution.Mean = 15.0
         <br>
         [...].Distribution.Variance = 10.5
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class BinomialDistribution extends DiscreteDistribution {

   final int n;
   final BernoulliDistribution bernoulli;

   private static final String MEAN = "Mean";
   private static final String VARIANCE = "Variance";

   /** as required by {@link ReflectionConstructable} */
   public static BinomialDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      final double variance = pars.get(ownNode, VARIANCE).asDouble();
      return new BinomialDistribution(mean, variance, rng);
   }

   public BinomialDistribution(double mean, double variance, RandomNumberGenerator rng) {
      super(rng);

      if (mean <= 0) {
         throw new DistributionException("Mean <= 0");
      }

      if (mean <= variance) {
         throw new DistributionException("Mean <= variance");
      }

      // mean=n*p
      // variance=n*p(1-p)
      // therefore p=1-(variance/mean)
      // n=mean/p

      this.n = (int) Math.floor(mean / (1 - variance / mean));
      this.bernoulli = new BernoulliDistribution(1 - variance / mean, rng);
   }

   public BinomialDistribution(double mean, double variance) {
      this(mean, variance, getDefaultRNG());
   }

   public BinomialDistribution(double p, int upperBound, RandomNumberGenerator rng) {
      super(rng);
      this.n = (int) Math.floor(upperBound);
      this.bernoulli = new BernoulliDistribution(p, rng);
   }

   public BinomialDistribution(double p, int upperBound) {
      this(p, upperBound, getDefaultRNG());
   }

   @Override
   public int next() {
      int result = 0;
      for (int i = 0; i < this.n; i++) {
         result += this.bernoulli.next();
      }
      return result;
   }
}
