/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.distributions.ProbabilityVector;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

import java.util.List;

/**
 * General Discrete Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>The first \(n\) values \(p_i (i = 0, ... , n-1 )\) of the distribution
 * are defined separately</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>single probabilities \(p_i (0 \le p_i \le 1 )\)</td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = \begin{cases} p_i &\mbox{for } 0 \le i < n \\ 0 & \mbox{else }
 * \end{cases} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X]=\sum\limits_{ i=0 }^{ n-1 } p_i \cdot i \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]=\sum\limits_{ i=0 }^{ n-1 } p_i \cdot i^2-(E[X])^2\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T=\frac{ \sqrt{ VAR[X] } }{ E[X] }\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Generating func.:</th>
 * <td>\(G(z)=\sum\limits_{ i=0 }^{ n-1 } p_i \cdot z^i \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.discrete.DiscreteGeneralDistribution
         <br>
         [...].Distribution.Data = [ 0.1 0.2 0.1 0.4 0.2 ]</code></td>
 * </tr>
 * </table>
 * 
 */
public class DiscreteGeneralDistribution extends DiscreteDistribution {

   public final ProbabilityVector vector;

   private static final String DATA = "data";

   /** as required by {@link ReflectionConstructable} */
   public static DiscreteGeneralDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double[] vector = pars.get(ownNode, DATA).asDoubleVector();
      return new DiscreteGeneralDistribution(vector, rng);
   }

   public DiscreteGeneralDistribution(double[] probVector, RandomNumberGenerator rng) {
      super(rng);
      vector = new ProbabilityVector(probVector);
   }

   public DiscreteGeneralDistribution(double[] probVector) {
      this(probVector, getDefaultRNG());
   }

   public DiscreteGeneralDistribution(List<Double> probVector, RandomNumberGenerator rng) {
      super(rng);
      vector = new ProbabilityVector(probVector);
   }

   public DiscreteGeneralDistribution(List<Double> probVector) {
      this(probVector, getDefaultRNG());
   }

   @Override
   public int next() {
      return vector.next(rng.next());
   }
}
