/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Discrete Uniform Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>All integer values \(i\) in the interval \(b_l \le i < b_u \) (\(b_l\)
 * and \(b_u\) being also integer values) have the same probability
 * \(\frac{1}{b_u-b_l}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * lower limit \(b_l\)
 * <li>
 * upper limit \(b_u > b_l\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = \begin{cases} \frac{1}{b_u-b_l} &\mbox{for } b_l \le i < b_u
 * \\ 0 & \mbox{else } \end{cases} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X]=\frac{b_l+b_u-1}{2} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]=\frac{(b_u-b_l-1) \cdot (b_u-b_l+1)}{12}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_X=\frac{ \sqrt{ (b_u-b_l-1) \cdot (b_u-b_l+1) } }{ \sqrt{3} \cdot
 * (b_l+b_u-1)}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Generating func.:</th>
 * <td>\(G(z)=z^{b_l}+...+z^{b_u-1} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.discrete.DiscreteUniformDistribution
         <br>
         [...].Distribution.LowerBound = 3
         <br>
         [...].Distribution.UpperBound = 7
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class DiscreteUniformDistribution extends DiscreteDistribution {

   public final int lowerBound;
   public final int interval;

   private static final String LOWER_BOUND = "LowerBound";
   private static final String UPPER_BOUND = "UpperBound";

   /** as required by {@link ReflectionConstructable} */
   public static DiscreteUniformDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final int lowerBound = pars.get(ownNode, LOWER_BOUND).asInteger();
      final int upperBound = pars.get(ownNode, UPPER_BOUND).asInteger();
      return new DiscreteUniformDistribution(lowerBound, upperBound, rng);
   }

   public DiscreteUniformDistribution(int lowerBound, int upperBound, RandomNumberGenerator rng) {
      super(rng);

      if (lowerBound >= upperBound) {
         throw new DistributionException("Lower bound >= Upper bound!");
      }

      this.lowerBound = lowerBound;
      interval = upperBound - lowerBound;
   }

   public DiscreteUniformDistribution(int lowerBound, int upperBound) {
      this(lowerBound, upperBound, getDefaultRNG());
   }

   @Override
   public int next() {
      return lowerBound + (int) Math.floor(interval * rng.next());
   }
}
