/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Poisson Distribution <br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Probability of the number of arrivals in a time interval with the
 * duration \(t\) for a Markovian arrival process (limit distribution of a
 * binomial distribution for \(n \to \infty , q \to 0, nq \to \lambda t\))</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>mean value \(m = \lambda t > 0\)</td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = \frac{(\lambda t)^i}{i!} \cdot exp(-\lambda t) =
 * \frac{m^i}{i!} \cdot exp(-m) \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X]= \lambda t = m \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]= \lambda t = m\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T=\frac{1}{ \sqrt{ \lambda t }} = \frac{1}{ \sqrt{m}}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Generating func.:</th>
 * <td>\(G(z)= exp(-\lambda t \cdot (1-z)) = exp(-m \cdot (1-z))\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.discrete.PoissonDistribution
         <br>
         [...].Distribution.Mean = 2.5</code></td>
 * </tr>
 * </table>
 * 
 */
public class PoissonDistribution extends DiscreteDistribution {

   public final double expNegMean;

   private static final String MEAN = "Mean";

   /** as required by {@link ReflectionConstructable} */
   public static PoissonDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      return new PoissonDistribution(mean, rng);
   }

   public PoissonDistribution(double mean, RandomNumberGenerator rng) {
      super(rng);

      if (mean <= 0) {
         throw new DistributionException("Mean <= 0!");
      }

      expNegMean = Math.exp(-mean);
   }

   public PoissonDistribution(double mean) {
      this(mean, getDefaultRNG());
   }

   @Override
   public int next() {
      double product;
      int no_of_attempts = 0;

      product = rng.next();
      while (product > expNegMean) {
         product *= rng.next();
         no_of_attempts += 1;
      }

      return no_of_attempts;
   }
}
