/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete.transform;

import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Linear Transformed Continuous Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Distribution of a discrete random variable \(X\), that results from a
 * linear transformation \(X = aY + b\) of the random variable \(Y\) with a
 * given discrete distribution ("base distribution").</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * base distribution \(q_i\)
 * <li>
 * integer number factor \(a \neq 0\)
 * <li>
 * integer number offset \(b\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = p_i = \begin{cases} q_j &\mbox{for } i=aj+b, j=0,1,... \\ 0
 * &\mbox{else} \end{cases}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X] = a \cdot E[Y] +b\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]= a^2 \cdot VAR[Y]\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T = \frac{1}{\frac{1}{c_z} + \frac{b}{a \sqrt{VAR[Y]}}}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.discrete.transform.DiscreteTransformedDistribution
         <br>
         [...].Distribution.BaseDist = ikr.simlib.distributions.discrete.Geometric
         <br>
         [...].Distribution.BaseDist.Mean = 2.5
         <br>
         [...].Distribution.Factor = 2 # optional, default = 1
         <br>
         [...].Distribution.Offset = 1 # optional, default = 0
      </code></td>
 * </tr>
 * </table>
 */
public class DiscreteTransformedDistribution extends DiscreteDistribution {

   public final DiscreteDistribution baseDist;
   public final int offset;
   public final int factor;

   private static final String BASE_DISTRIBUTION = "BaseDist";
   private static final String OFFSET = "Offset";
   private static final String FACTOR = "Factor";

   /** as required by {@link ReflectionConstructable} */
   public static DiscreteTransformedDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final DiscreteDistribution baseDist = pars.get(ownNode, BASE_DISTRIBUTION).asInstance(DiscreteDistribution.class, ownNode.createChildNode(BASE_DISTRIBUTION), pars, rng);
      final int offset = pars.get(ownNode, OFFSET).asInteger();
      final int factor = pars.get(ownNode, FACTOR).asInteger();
      return new DiscreteTransformedDistribution(baseDist, offset, factor);
   }

   public DiscreteTransformedDistribution(DiscreteDistribution baseDist, int offset, int factor) {
      super();

      this.baseDist = baseDist;
      this.offset = offset;
      this.factor = factor;
   }

   public DiscreteTransformedDistribution(DiscreteDistribution baseDist, int offset) {
      this(baseDist, offset, 1);
   }

   public DiscreteTransformedDistribution(DiscreteDistribution baseDist) {
      this(baseDist, 0, 1);
   }

   @Override
   public int next() {
      return offset + factor * baseDist.next();
   }
}
