/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.branches;

import ikr.simlib.entities.Entity;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.Port;
import ikr.simlib.ports.input.InputPort;
import ikr.simlib.ports.output.OutputPort;
import ikr.simlib.ports.output.SynchronousOutputPort;

import java.util.ArrayList;
import java.util.List;

/**
 * The template <code>Expander</code> is meant to be a base classe for components with one input port and
several output ports. Each of these output ports can be assigned an attribute (its type is defined
by the template parameter).<br>
<br>
- Upon construction, <code>Expander</code> requires the number of ports to be created. The created
ports are named "output 1", "output 2", ... . However, new ports can be added with the help
of the method {@link #addPort(String, Object)} using the port name and its attribute. With {@link #removePort(String)} an
output port can be removed.<br>
<br>
- The attributes can be changed subsequently using {@link #setPortParameter(int, Object)}, which
requires port identification with its index (1, 2,... , n).<br>
<br>
- The purely virtual method {@link #handleMessageIndication(Message)} must be defined in derived
classes.<br>
- The type used for the output ports in both classes is {@link ikr.simlib.ports.output.SynchronousOutputPort}, i.e. no
blocking may occur at an output port.<br><br>
 * 
 * An expander element generally has one input port and an arbitrary number of
 * (synchronous) output ports.
 * 
 * <p>
 * To each output port in the generic class <code>Expander</code> an attribute
 * can be associated whose type is the generic parameter.
 * </p>
 * 
 * <p>
 * An initial number of output ports can be defined on construction. The names
 * of the ports are then "output 1", "output 2", ... Moreover, it is possible to
 * add and remove ports dynamically using <code>AddPort</code> and
 * <code>RemovePort</code>.
 * </p>
 * 
 * @param <P>
 *           type of parameter
 */
public abstract class Expander<P> extends Entity {

   protected InputPort inputPort;
   protected List<SynchronousOutputPort> outputPorts;
   protected List<P> portParameters;

   public Expander(int nrOfPorts, P initValue, SimNode ownNode) {
      this(ownNode);

      for (int i = 0; i < nrOfPorts; i++) {
         outputPorts.add(new SynchronousOutputPort(this, "output " + (i + 1)));
         portParameters.add(initValue);
      }
   }

   public Expander(SimNode ownNode) {
      super(ownNode);

      inputPort = new InputPort(this) {
         @Override
         protected void handleMessageIndication(Message msg) {
            Expander.this.handleMessageIndication(msg);
         }
      };
      outputPorts = new ArrayList<SynchronousOutputPort>();
      portParameters = new ArrayList<P>();
   }

   protected int portName2Index(String portName) {
      int i = 0;
      for (Port p : outputPorts) {
         if (p.getLocalName().equals(portName)) {
            return i;
         }
         i++;
      }

      throw new SimulationException("port name not found: " + portName);
   }

   protected abstract void handleMessageIndication(Message msg);

   public void addPort(String portName, P para) {
      outputPorts.add(new SynchronousOutputPort(this, portName));
      portParameters.add(para);
   }

   public void addPort(P para) {
      addPort(OutputPort.OUTPUT_PORT_NAME + " " + (outputPorts.size() + 1), para);
   }

   public void removePort(String portName) {
      int index = portName2Index(portName);

      outputPorts.remove(index);
      portParameters.remove(index);
   }

   public void setPortParameter(int portNr, P para) {
      if (portNr < 1 || portNr > outputPorts.size()) {
         throw new SimulationException("Port Id = " + portNr + " (out of range)");
      }

      portParameters.set(portNr - 1, para);
   }
}
