/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.demultiplexers;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.messages.IPPacket;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.Port;

/**
 * <code>IPDemultiplexer</code> is derived from <code>LabelDemultiplexer</code>
 * and therefore provides the same functionality with respect to port creation
 * and management. Instead of using <code>LabelMessage</code>,
 * <code>IPDemultiplexer</code> assumes incoming messages to be of
 * <code>IPPacket</code> type. The demux may be configured to consider either
 * the destination id, source id, type id, or group id to determine the output
 * port. A start id and a modulus can be defined like in
 * <code>LabelDemultiplexer</code>.
 */
public class IPDemultiplexer extends LabelDemultiplexer {

   public static enum IPDemuxMode {
      DESTINATION, SRC, TYPE, GROUP
   }

   protected final IPDemuxMode mode;

   @Override
   protected Port determineOutputPort(Message msg) {
      IPPacket packet = (IPPacket) msg;

      int id = -1;

      if (mode == IPDemuxMode.DESTINATION) {
         id = packet.getDestinationId();
      } else if (mode == IPDemuxMode.SRC) {
         id = packet.getSourceId();
      } else if (mode == IPDemuxMode.GROUP) {
         id = packet.getGroupId();
      } else if (mode == IPDemuxMode.TYPE) {
         id = packet.getTypeId();
      }

      if (id < startId || id >= (startId + outputPorts.size() * modulus)) {
         throw new SimulationException("label out of range:" + id);
      }

      return outputPorts.get((id - startId) / modulus);
   }

   public IPDemultiplexer(int noOfPorts, SimNode ownNode, int startLabel, int labelModulus, IPDemuxMode mode) {
      super(noOfPorts, ownNode, startLabel, labelModulus);

      this.mode = mode;
   }

   public IPDemultiplexer(int noOfPorts, SimNode ownNode, int startLabel, IPDemuxMode mode) {
      this(noOfPorts, ownNode, startLabel, 1, mode);
   }

   public IPDemultiplexer(int noOfPorts, SimNode ownNode, int startLabel) {
      this(noOfPorts, ownNode, startLabel, 1, IPDemuxMode.DESTINATION);
   }

   public IPDemultiplexer(int noOfPorts, SimNode ownNode) {
      this(noOfPorts, ownNode, 0, 1, IPDemuxMode.DESTINATION);
   }

   public IPDemultiplexer(SimNode ownNode, int startLabel, int labelModulus, IPDemuxMode mode) {
      this(0, ownNode, startLabel, labelModulus, mode);
   }

   public IPDemultiplexer(SimNode ownNode, int startLabel) {
      this(0, ownNode, startLabel, 1, IPDemuxMode.DESTINATION);
   }

   public IPDemultiplexer(SimNode ownNode) {
      this(0, ownNode, 0, 1, IPDemuxMode.DESTINATION);
   }
}
