/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.generators;

import ikr.simlib.distributions.Distribution;
import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.factories.Factory;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * <code>BurstGenerator</code> generates bursts of information that are
 * segmented in messages of max. size and sent according to an
 * <em>interPacketTime</em>. <code>MessageFactory</code> and parsing is done as
 * in <code>StdGenerator</code>.
 */
public class BurstGenerator extends Generator {

   private static final String BURST_IAT_DIST = "BurstIATDist";
   private static final String BURST_LENGTH_DIST = "BurstLengthDist";
   private static final String INTER_PACKET_TIME = "InterPacketTime";
   private static final String MAX_PACKET_SIZE = "MaxPacketSize";

   /** as required by {@link ReflectionConstructable} */
   public static BurstGenerator createInstance(SimNode ownNode, Parameters pars) {
      ContinuousDistribution burstIATDist = pars.get(ownNode, BURST_IAT_DIST).asInstance(ContinuousDistribution.class, ownNode.createChildNode(BURST_IAT_DIST), pars, Distribution.getDefaultRNG());
      DiscreteDistribution burstLengthDist = pars.get(ownNode, BURST_LENGTH_DIST).asInstance(DiscreteDistribution.class, ownNode.createChildNode(BURST_LENGTH_DIST), pars, Distribution.getDefaultRNG());
      Duration interPacketTime = pars.get(ownNode, INTER_PACKET_TIME).asInstance(Duration.class, ownNode.createChildNode(INTER_PACKET_TIME), pars);
      if (Duration.ZERO.isGreaterThan(interPacketTime))
         throw new SimulationException("inter cell time < 0");
      int maxPacketSize = pars.get(ownNode, MAX_PACKET_SIZE).asInteger();
      if (maxPacketSize < 0)
         throw new SimulationException("max packet size < 0");

      return new BurstGenerator(burstLengthDist, burstIATDist, interPacketTime, maxPacketSize, ownNode);
   }

   protected static class BurstData implements Comparable<BurstData> {
      public final PointInTime burstArrivalTime;
      public PointInTime nextPacketArrivalTime;
      public int residualLength;

      public BurstData(PointInTime burstArrivalTime, int length) {
         this.burstArrivalTime = burstArrivalTime;
         nextPacketArrivalTime = burstArrivalTime;
         residualLength = length;
      }

      @Override
      public int compareTo(BurstData o) {
         return nextPacketArrivalTime.compareTo(o.nextPacketArrivalTime);
      }
   }

   protected final DiscreteDistribution burstLengthDist;
   protected final ContinuousDistribution burstIATDist;
   protected final Duration interPacketTime;
   protected final int maxPacketSize;

   protected final List<BurstData> activeBursts;
   protected BurstData nextBurst;

   public BurstGenerator(DiscreteDistribution burstLengthDist, ContinuousDistribution burstIATDist, Duration interPacketTime, int maxPacketSize, SimNode ownNode, Factory<Message> factory, boolean activateOnStart) {
      super(ownNode, factory, activateOnStart);

      if (interPacketTime.compareTo(Duration.ZERO) < 0) {
         throw new SimulationException("inter cell time < 0");
      }

      this.burstLengthDist = burstLengthDist;
      this.burstIATDist = burstIATDist;
      this.interPacketTime = interPacketTime;
      this.maxPacketSize = maxPacketSize;

      this.activeBursts = new ArrayList<BurstData>();
   }

   public BurstGenerator(DiscreteDistribution burstLengthDist, ContinuousDistribution burstIATDist, Duration interPacketTime, int maxPacketSize, SimNode ownNode, Factory<Message> factory) {
      this(burstLengthDist, burstIATDist, interPacketTime, maxPacketSize, ownNode, factory, true);
   }

   public BurstGenerator(DiscreteDistribution burstLengthDist, ContinuousDistribution burstIATDist, Duration interPacketTime, int maxPacketSize, SimNode ownNode) {
      this(burstLengthDist, burstIATDist, interPacketTime, maxPacketSize, ownNode, null);
   }

   public BurstGenerator(DiscreteDistribution burstLengthDist, ContinuousDistribution burstIATDist, Duration interPacketTime, int maxPacketSize, SimNode ownNode, boolean activateOnStart) {
      this(burstLengthDist, burstIATDist, interPacketTime, maxPacketSize, ownNode, null, activateOnStart);
   }

   @Override
   public void setActivated(boolean start) {
      if (start != activated) {
         if (start) {
            assert currentEventToken == null || !currentEventToken.isPending();
            assert activeBursts.isEmpty();
            assert nextBurst == null;
            createNewNextBurst(Calendar.getInstance().getSystemTime());
            currentEventToken = Calendar.getInstance().postEvent(new CalendarCallback() {
               @Override
               public void execute() {
                  BurstGenerator.this.handleProcessEvent();
               }
            }, null, getNextEventTime(Calendar.getInstance().getSystemTime()));
         } else {
            assert currentEventToken.isPending();
            nextBurst = null;
            activeBursts.clear();
            currentEventToken.cancel();
         }
         activated = start;
      }
   }

   @Override
   protected PointInTime getNextEventTime(PointInTime currentTime) {
      return getCurentBurst().nextPacketArrivalTime;
   }

   @Override
   protected Message createMessage() {
      final BurstData currentBurst = getCurentBurst();
      final int messageLength = Math.min(currentBurst.residualLength, maxPacketSize);
      currentBurst.residualLength -= messageLength;
      currentBurst.nextPacketArrivalTime = currentBurst.nextPacketArrivalTime.plus(interPacketTime);

      assert currentBurst.residualLength >= 0;
      if (currentBurst.residualLength == 0) {
         activeBursts.remove(currentBurst);
      }

      return createMessageToSend(messageLength);
   }

   protected Message createMessageToSend(int length) {
      final Message tempmsg = super.createMessage();
      tempmsg.setLength(length);
      return tempmsg;
   }

   /**
    * creates a new burst and stores it in nextBurst
    * 
    * @param refTime
    *           PointInTime which serves as reference for the burst IAT
    */
   protected void createNewNextBurst(PointInTime refTime) {
      nextBurst = new BurstData(refTime.plus(Duration.fromSeconds(burstIATDist.next())), burstLengthDist.next());
   }

   /**
    * decides which burst is the one which sends the next packet. If the
    * nextBurst is the one to send the next packet, it is added to the burstList
    * and a new nextBurst is created.
    * 
    * @return the burst which sends the next packet
    */
   protected BurstData getCurentBurst() {
      assert nextBurst != null;
      BurstData nextPacketBurst = activeBursts.isEmpty() ? null : Collections.min(activeBursts);
      if (nextPacketBurst == null || nextPacketBurst.compareTo(nextBurst) > 0) {
         // add nextBurst to the list and create a new nextBurst
         nextPacketBurst = nextBurst;
         activeBursts.add(nextBurst);
         createNewNextBurst(nextBurst.burstArrivalTime);
      }
      assert nextPacketBurst != null;
      assert nextPacketBurst == Collections.min(activeBursts);
      assert nextPacketBurst.compareTo(nextBurst) <= 0;
      return nextPacketBurst;
   }
}
