/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.generators;

import ikr.simlib.control.Simulation;
import ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager;
import ikr.simlib.control.simulationcontrol.callbacks.StartSimulationCallbackI;
import ikr.simlib.entities.Entity;
import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.EventToken;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.factories.Factory;
import ikr.simlib.messages.Message;
import ikr.simlib.messages.factories.MessageFactory;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.ports.output.OutputPort;
import ikr.simlib.ports.output.SynchronousOutputPort;

/**
 * The class <code>Generator</code> is derived from
 * {@link ikr.simlib.entities.Entity} and is the abstract base class for all
 * message generators. Deriving classes must implement
 * {@link #getNextEventTime(PointInTime)} to calculate the time for the next message
 * to be generated.
 * <p>
 * Each Generator has to have exactly one MessageFactory. If no MessageFactory
 * is provided, Generator creates a new MessageFactory with an default Message
 * of type Message and length 1.
 * 
 * Generator provides the basic functionality for all derived generators:<br>
 * <br>
 * - The generation of messages is controlled by events.<br>
 * <br>
 * - Each time an event has been processed, a new event is registered.<br>
 * <br>
 * <br>
 * The creation of a message is performed by calling {@link #createMessage()},
 * which forces an aggregated object derived from the class
 * {@link ikr.simlib.factories.Factory} to create a new message. Almost all
 * details of message generation are therefore hidden within the message
 * factory. The term message is used as a synonym for all possible
 * representations of a messages, which are derived from
 * {@link ikr.simlib.messages.Message}.<br>
 * <br>
 * - It is possible to set and get message factories during any phase of
 * simulation by calling {@link #setMessageFactory(Factory)} and
 * {@link #getMessageFactory()}. Several entities (generators) can share the
 * same message factory.<br>
 * <br>
 * - For each generator class, there is a corresponding parser class.
 * 
 */
public abstract class Generator extends Entity implements ReflectionConstructable2<SimNode, Parameters> {

   protected SynchronousOutputPort outputPort;

   protected Factory<Message> messageFactory;

   protected boolean activated;

   protected EventToken currentEventToken;

   protected Generator(SimNode ownNode, Factory<Message> factory, final boolean activateOnStart) {
      super(ownNode);

      outputPort = new SynchronousOutputPort(this);

      activated = false;

      if (factory == null) {
         messageFactory = new MessageFactory();
      } else {
         messageFactory = factory;
      }

      if (activateOnStart) {
         SimulationControlManager.getInstance().registerStartSimulationCallback(new StartSimulationCallbackI() {
            @Override
            public void handleStartSimulation(Simulation sim) {
               Generator.this.setActivated(true);
            }
         });
      }
   }

   protected Generator(SimNode ownNode) {
      this(ownNode, null, true);
   }

   protected abstract PointInTime getNextEventTime(PointInTime currentTime);

   protected Message createMessage() {
      return messageFactory.create();
   }

   public void setActivated(boolean start) {
      if (start != activated) {
         if (start) {
            assert currentEventToken == null || !currentEventToken.isPending();
            currentEventToken = Calendar.getInstance().postEvent(new CalendarCallback() {
               @Override
               public void execute() {
                  Generator.this.handleProcessEvent();
               }
            }, null, getNextEventTime(Calendar.getInstance().getSystemTime()));
         } else {
            assert currentEventToken.isPending();
            currentEventToken.cancel();
         }
         activated = start;
      }
   }

   public final void activate() {
      setActivated(true);
   }

   public final void deactivate() {
      setActivated(false);
   }

   public void setMessageFactory(Factory<Message> Factory) {
      messageFactory = Factory;
   }

   public Factory getMessageFactory() {
      return messageFactory;
   }

   protected void handleProcessEvent() {
      final Message message = createMessage();
      outputPort.sendMessage(message);

      // Schedule next event
      currentEventToken = Calendar.getInstance().postEvent(new CalendarCallback() {
         @Override
         public void execute() {
            Generator.this.handleProcessEvent();
         }
      }, null, getNextEventTime(Calendar.getInstance().getSystemTime()));
   }

   public OutputPort getOutput() {
      return outputPort;
   }
}
