/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.generators;

import ikr.simlib.control.Simulation;
import ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager;
import ikr.simlib.control.simulationcontrol.callbacks.StartSimulationCallbackI;
import ikr.simlib.distributions.Distribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.entities.Entity;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.factories.Factory;
import ikr.simlib.messages.Message;
import ikr.simlib.messages.factories.MessageFactory;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.ports.Port;
import ikr.simlib.ports.output.OutputPort;

/**
 * <code>GreedyGenerator</code> generates as many messages as the attached
 * component can handle. That is, as soon as a message has been fetched, the
 * next one is generated and offered. The greedy generator only works with an
 * <em>active receiver</em> attached to it. If a distribution for the length of
 * a message is provided/parsed it sets the message length according to the
 * given distribution each time a message is created. A
 * <code>MessageFactory</code> may be provided to create messages other than
 * <code>Message</code>.
 * 
 * If a length distribution is provided, the length of the default
 * <code>Message</code> generated by the <code>MessageFactory</code> is
 * overwritten.
 */
public class GreedyGenerator extends Entity implements ReflectionConstructable2<SimNode, Parameters> {

   private static final String LENGTH_DIST = "LengthDist";

   /** as required by {@link ReflectionConstructable} */
   public static GreedyGenerator createInstance(SimNode ownNode, Parameters pars) {
      DiscreteDistribution messageLengthDist = null;
      if (pars.hasParameter(ownNode, LENGTH_DIST)) {
         messageLengthDist = pars.get(ownNode, LENGTH_DIST).asInstance(DiscreteDistribution.class, ownNode.createChildNode(LENGTH_DIST), pars, Distribution.getDefaultRNG());
      }

      return new GreedyGenerator(messageLengthDist, ownNode);
   }

   private Message message;
   private boolean activated;
   private OutputPort outputPort;
   private Factory<Message> messageFactory;
   private DiscreteDistribution messageLengthDist;

   public GreedyGenerator(DiscreteDistribution lengthdist, SimNode ownNode, Factory<Message> factory, boolean activateOnStart) {
      super(ownNode);

      outputPort = new OutputPort(this) {
         @Override
         protected Message handleGetMessage() {
            return GreedyGenerator.this.handleGetMessage(this);
         }

         @Override
         protected boolean handleIsMessageAvailable() {
            return GreedyGenerator.this.handleIsMessageAvailable(this);
         }
      };

      message = null;
      activated = false;

      if (factory == null) {
         messageFactory = new MessageFactory();
      } else {
         messageFactory = factory;
      }

      if (activateOnStart) {
         SimulationControlManager.getInstance().registerStartSimulationCallback(new StartSimulationCallbackI() {
            @Override
            public void handleStartSimulation(Simulation sim) {
               GreedyGenerator.this.setActivated(true);
            }
         });
      }

      messageLengthDist = lengthdist;
   }

   public GreedyGenerator(DiscreteDistribution lengthdist, SimNode ownNode, Factory<Message> factory) {
      this(lengthdist, ownNode, factory, true);
   }

   public GreedyGenerator(DiscreteDistribution lengthdist, SimNode ownNode) {
      this(lengthdist, ownNode, null, true);
   }

   protected Message createMessage() {
      Message tempMes = messageFactory.create();
      if (messageLengthDist != null) {
         tempMes.setLength(messageLengthDist.next());
      }
      return tempMes;
   }

   public void setActivated(boolean start) {
      if (start != activated) {
         if (start) {
            if (message == null) {
               message = createMessage();
               outputPort.messageIndication(message);
            }
         } else {
            throw new SimulationException("deactivating a greedy generator is not allowed");
         }
         activated = start;
      }
   }

   public void activate() {
      setActivated(true);
   }

   public void deactivate() {
      setActivated(false);
   }

   public DiscreteDistribution getLengthDist() {
      return messageLengthDist;
   }

   protected Message handleGetMessage(Port port) {
      if (message == null) {
         // throw no message
         throw new SimulationException("no message available");
      }

      Message tmp = message;
      message = createMessage();
      return tmp;
   }

   protected boolean handleIsMessageAvailable(Port port) {
      return message != null;
   }
}
