/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.generators;

import ikr.simlib.distributions.Distribution;
import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.factories.Factory;
import ikr.simlib.messages.Message;
import ikr.simlib.messages.factories.MessageFactory;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * <code>StdGenerator</code> generates messages according to a renewal point
 * process for which IAT distribution is provided. If a distribution for the
 * length of a message is provided, it sets the message length in an iid
 * fashion. That is, each time a message is created, the message length is drawn
 * according to the given distribution. A {@link MessageFactory} may be provided
 * to create messages other than {@link Message}.
 * 
 * <p>
 * If a length distribution is provided, the length of the default Message
 * generated by the <code>MessageFactory</code> is overwritten.
 * <p>
 * StdGenerator is a very flexible Generator with the following characteristics:
 * <ul>
 * <li>Messages are generated according to the interarrival time distribution.
 * <li>Message length follows an optional discrete distribution. Each time a new
 * message is generated by a <code>MessageFactory</code>, the length is adapted
 * by calling <code>setLength()</code> of the message and passing the new length
 * according to the length distribution.
 * <li>If no length distribution is provided, the length of the generated
 * message is not changed. Therefore control of the length remains at the
 * <code>MessageFactory</code>.
 * <li>A message factory can be provided by passing it upon generation of
 * <code>StdGenerator</code>.
 * </ul>
 */
public class StdGenerator extends Generator {

   private static final String IATDist = "IATDist";
   private static final String LENGTH_DIST = "LengthDist";

   /** as required by {@link ReflectionConstructable} */
   public static StdGenerator createInstance(SimNode ownNode, Parameters pars) {
      ContinuousDistribution interArrivalTime = pars.get(ownNode, IATDist).asInstance(ContinuousDistribution.class, ownNode.createChildNode(IATDist), pars, Distribution.getDefaultRNG());
      DiscreteDistribution messageLengthDist = null;
      if (pars.hasParameter(ownNode, LENGTH_DIST)) {
         messageLengthDist = pars.get(ownNode, LENGTH_DIST).asInstance(DiscreteDistribution.class, ownNode.createChildNode(LENGTH_DIST), pars, Distribution.getDefaultRNG());
      }

      return new StdGenerator(interArrivalTime, messageLengthDist, ownNode, null, true);
   }

   protected final ContinuousDistribution interArrivalTime;
   protected final DiscreteDistribution messageLengthDist;

   /**
    * @param iat
    *           Distribution for the interarrival time
    * @param lengthdist
    *           Distribution for the message length. May be null, in that case
    *           the message length is used as set by the factory.
    * @param ownNode
    *           SimNode of the generator
    * @param factory
    *           Factory used to create new message objects. May be null, in that
    *           case a new {@link MessageFactory} is created.
    * @param activateOnStart
    *           If true, the generator automatically starts generating messages.
    *           Otherwise, {@link #setActivated(boolean)} has to be called
    *           manually.
    */
   public StdGenerator(ContinuousDistribution iat, DiscreteDistribution lengthdist, SimNode ownNode, Factory<Message> factory, boolean activateOnStart) {
      super(ownNode, factory, activateOnStart);

      interArrivalTime = iat;
      assert interArrivalTime != null;

      messageLengthDist = lengthdist;
   }

   @Override
   protected Message createMessage() {
      Message tempMes = super.createMessage();
      if (messageLengthDist != null)
         tempMes.setLength(messageLengthDist.next());
      return tempMes;
   }

   @Override
   protected PointInTime getNextEventTime(PointInTime currentTime) {
      return currentTime.plus(Duration.fromSeconds(interArrivalTime.next()));
   }
}
