/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.multiphase;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.entities.Entity;
import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.EventToken;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.output.SynchronousOutputPort;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.statistics.integral.IntegralStatistic;
import ikr.simlib.statistics.integral.StdIntegralStatistic;
import ikr.simlib.statistics.sample.SampleStatistic;
import ikr.simlib.statistics.sample.StdSampleStatistic;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * <code>MultiPhasePrioServer</code> provides a priority server component with
 * an arbitrary number of phases. A phase can be created by the method
 * <code>CreatePhase</code>. The overall preemption behaviour can be defined by
 * the preemption distance. A phase can be interrupted by another phase if the
 * difference of the priorities is greater than the preemption distance.
 * 
 * <p>
 * Each phase has its own name, priority, service time distribution, queueing
 * strategy and preemption strategy. A phase is realized by the class
 * <code>ServerPhase</code>.
 * </p>
 * 
 * <p>
 * All phases with the same priority share a common object of class
 * <code>ServerPriorityQueue</code> which is automatically created by
 * <code>MultiPhasePrioServer</code>.
 * </p>
 */
public class MultiPhasePrioServer extends Entity {

   protected static enum State {
      IDLE, BUSY
   }

   protected EventToken currentEventToken;

   protected final int preemptionDistance;
   protected final Map<String, ServerPhase> phaseMap;
   protected final List<ServerPriorityQueue> priorityQueueList;

   protected State state;
   protected ServerJob currentJob;
   protected PointInTime entryTime;
   protected int currentQueueLength;

   protected final IntegralStatistic occupancy;
   protected final IntegralStatistic queueLength;
   protected final SampleStatistic transferTime;
   protected final SampleStatistic waitingTime;

   protected void addPhase(ServerPhase phase) {
      phaseMap.put(phase.getNode().getName(), phase);
   }

   protected void removePhase(ServerPhase phase) {
      phaseMap.remove(phase.getNode().getName());
   }

   protected SynchronousOutputPort getOutputPort(ServerJob job) {
      return job.phase.outputPort;
   }

   protected ServerJob get() {
      ServerJob job = null;

      for (ServerPriorityQueue queue : priorityQueueList) {
         if (!queue.isEmpty()) {
            job = queue.popHead();
            updateStatisticsAtGet(job, Calendar.getInstance().getSystemTime());
            break;
         }
      }

      return job;
   }

   protected void run(ServerJob job) {
      PointInTime currentTime = Calendar.getInstance().getSystemTime();

      if ((job.nrOfInterrupts == 0) || (job.phase.getPreemptionStrategy() == PreemptionStrategy.REPEAT_WITH_RESAMPLING)) {
         job.scheduledServiceTime = job.phase.getServiceTime();
         job.residualServiceTime = job.scheduledServiceTime;
      }

      currentEventToken = Calendar.getInstance().postEvent(new CalendarCallback() {
         @Override
         public void execute() {
            MultiPhasePrioServer.this.handleProcessEvent();
         }
      }, null, currentTime.plus(job.residualServiceTime));
      state = State.BUSY;
      currentJob = job;
      entryTime = currentTime;
      updateStatisticsAtRun(job, currentTime);
   }

   protected void interrupt(ServerJob job) {
      PointInTime currentTime = Calendar.getInstance().getSystemTime();
      Duration serviceInterval = entryTime.until(currentTime);
      assert serviceInterval.isGreaterThanOrEqualTo(Duration.ZERO);

      job.nrOfInterrupts++;
      job.totalServiceTime = job.totalServiceTime.plus(serviceInterval);
      if (job.phase.getPreemptionStrategy() == PreemptionStrategy.RESUME) {
         job.residualServiceTime = job.residualServiceTime.plus(serviceInterval.negated());
      }

      job.phase.putBack(job);
      currentEventToken.cancel();
      updateStatisticsAtInterrupt(job, currentTime);
   }

   protected void updateStatisticsAtPut(ServerJob job, PointInTime currentTime) {
      queueLength.update(++currentQueueLength);
      job.phase.updateQueueLength(currentTime, 1);
   }

   protected void updateStatisticsAtGet(ServerJob job, PointInTime currentTime) {
      queueLength.update(--currentQueueLength);
      job.phase.updateQueueLength(currentTime, -1);
   }

   protected void updateStatisticsAtRun(ServerJob job, PointInTime currentTime) {
      occupancy.update(1);
      job.phase.updateOccupancy(currentTime, 1);
   }

   protected void updateStatisticsAtInterrupt(ServerJob job, PointInTime currentTime) {
      occupancy.update(0);
      queueLength.update(++currentQueueLength);
      job.phase.updateQueueLength(currentTime, 1);
      job.phase.updateOccupancy(currentTime, 0);
   }

   protected void updateStatisticsAtEndOfService(ServerJob job, PointInTime currentTime) {
      occupancy.update(0);
      transferTime.update(job.arrivalTime.until(currentTime).toSeconds());
      waitingTime.update(job.arrivalTime.until(currentTime).plus(job.totalServiceTime.negated()).toSeconds());
      job.phase.updateOccupancy(currentTime, 0);
      job.phase.updateTransferTime(job.arrivalTime.until(currentTime).toSeconds());
      job.phase.updateWaitingTime(job.arrivalTime.until(currentTime).plus(job.totalServiceTime.negated()).toSeconds());
   }

   public MultiPhasePrioServer(int preemptionDistance, SimNode ownNode) {
      super(ownNode);

      this.preemptionDistance = preemptionDistance;
      phaseMap = new HashMap<String, ServerPhase>();
      priorityQueueList = new ArrayList<ServerPriorityQueue>();

      state = State.IDLE;
      currentJob = null;
      entryTime = PointInTime.ZERO;
      currentQueueLength = 0;
      occupancy = new StdIntegralStatistic(ownNode.createChildNode("Occupancy"));
      queueLength = new StdIntegralStatistic(ownNode.createChildNode("QueueLength"));
      transferTime = new StdSampleStatistic(ownNode.createChildNode("TransferTime"));
      transferTime.addResultTag(ResultKeywords.UNIT, "s");
      waitingTime = new StdSampleStatistic(ownNode.createChildNode("WaitingTime"));
      waitingTime.addResultTag(ResultKeywords.UNIT, "s");
   }

   public void createPhase(String phaseName, ContinuousDistribution serviceTime, int priority, PreemptionStrategy preemptionStrategy, QueueingStrategy queueingStrategy) {
      ServerPhase phase = new ServerPhase(simNode.createChildNode(phaseName), this, serviceTime, priority, preemptionStrategy, queueingStrategy);

      addPhase(phase);
   }

   public final void createPhase(String phaseName, ContinuousDistribution serviceTime, int priority, PreemptionStrategy preemptionStrategy) {
      createPhase(phaseName, serviceTime, priority, preemptionStrategy, QueueingStrategy.FIFO);
   }

   public final void createPhase(String phaseName, ContinuousDistribution serviceTime, int priority) {
      createPhase(phaseName, serviceTime, priority, PreemptionStrategy.RESUME, QueueingStrategy.FIFO);

   }

   public final void createPhase(String phaseName, ContinuousDistribution serviceTime) {
      createPhase(phaseName, serviceTime, 0, PreemptionStrategy.RESUME, QueueingStrategy.FIFO);
   }

   public void deletePhase(String phaseName) {
      ServerPhase phase;

      if (!phaseMap.containsKey(phaseName)) {
         throw new SimulationException("Phase " + phaseName + " is not bound at server " + getName() + ".");
      }

      phase = phaseMap.get(phaseName);
      if (currentJob != null) {
         if (currentJob.phase == phase) {
            handleProcessEvent();
         }
      }

      removePhase(phase);
   }

   public int getNoOfPhases() {
      return phaseMap.size();
   }

   public void put(Message msg, ServerPhase phase) {
      ServerJob job = new ServerJob(msg, phase, Calendar.getInstance().getSystemTime());
      if (state == State.IDLE) {
         run(job);
      } else if (job.phase.getPriority() >= currentJob.phase.getPriority() + preemptionDistance) {
         interrupt(currentJob);
         run(job);
      } else {
         job.phase.append(job);
         updateStatisticsAtPut(job, Calendar.getInstance().getSystemTime());
      }
   }

   public ServerPriorityQueue getPriorityQueue(int priority) {
      ServerPriorityQueue queue = null;
      int index = 0;

      for (ServerPriorityQueue current : priorityQueueList) {
         if (current.getPriority() < priority) {
            break;
         } else if (current.getPriority() > priority) {
            index++;
         } else {
            queue = current;
            break;
         }
      }

      if (queue == null) {
         queue = new ServerPriorityQueue(priority);
         if (index == priorityQueueList.size()) {
            priorityQueueList.add(queue);
         } else {
            priorityQueueList.add(index, queue);
         }
      }

      return queue;
   }

   public void handleProcessEvent() {
      SynchronousOutputPort outputPort = getOutputPort(currentJob);
      Message message = currentJob.message;

      currentJob.totalServiceTime = currentJob.totalServiceTime.plus(currentJob.residualServiceTime);
      currentJob.residualServiceTime = Duration.ZERO;
      updateStatisticsAtEndOfService(currentJob, Calendar.getInstance().getSystemTime());

      outputPort.sendMessage(message);

      ServerJob nextJob = get();
      if (nextJob != null) {
         run(nextJob);
      } else {
         state = State.IDLE;
      }
   }
}
