/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.multiphase;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.entities.Entity;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.output.SynchronousOutputPort;
import ikr.simlib.statistics.integral.IntegralStatistic;
import ikr.simlib.statistics.integral.StdIntegralStatistic;
import ikr.simlib.statistics.sample.SampleStatistic;
import ikr.simlib.statistics.sample.StdSampleStatistic;

public class ServerPhase extends Entity {

   protected final MultiPhasePrioServer owner;
   protected final ServerInputPort inputPort;
   protected final SynchronousOutputPort outputPort;

   protected final ContinuousDistribution serviceTime;
   protected final PreemptionStrategy preemptionStrategy;
   protected final QueueingStrategy queueingStrategy;
   protected final ServerPriorityQueue priorityQueue;

   protected final IntegralStatistic occupancy;
   protected final IntegralStatistic queueLength;
   protected final SampleStatistic transferTime;
   protected final SampleStatistic waitingTime;
   protected int currentQueueLength;

   public ServerPhase(SimNode ownNode, MultiPhasePrioServer owner, ContinuousDistribution serviceTime, int priority, PreemptionStrategy preemptionStrategy, QueueingStrategy queueingStrategy) {
      super(ownNode);

      this.owner = owner;
      this.serviceTime = serviceTime;
      this.preemptionStrategy = preemptionStrategy;
      this.queueingStrategy = queueingStrategy;
      priorityQueue = owner.getPriorityQueue(priority);
      occupancy = new StdIntegralStatistic(ownNode.createChildNode("Occupancy"));
      queueLength = new StdIntegralStatistic(ownNode.createChildNode("QueueLength"));
      transferTime = new StdSampleStatistic(ownNode.createChildNode("TransferTime"));
      waitingTime = new StdSampleStatistic(ownNode.createChildNode("WaitingTime"));
      currentQueueLength = 0;

      inputPort = new ServerInputPort(owner, this, ownNode.getName() + "In");
      outputPort = new SynchronousOutputPort(owner, ownNode.getName() + "Out");
   }

   public int getPriority() {
      return priorityQueue.getPriority();
   }

   public PreemptionStrategy getPreemptionStrategy() {
      return preemptionStrategy;
   }

   public Duration getServiceTime() {
      return Duration.fromSeconds(serviceTime.next());
   }

   public void append(ServerJob job) {
      switch (queueingStrategy) {
      case FIFO:
         priorityQueue.append(job);
         break;
      case LIFO:
         priorityQueue.insert(job);
         break;
      }
   }

   public void putBack(ServerJob job) {
      priorityQueue.putBack(job);
   }

   public void updateOccupancy(PointInTime currentTime, double value) {
      occupancy.update(value);
   }

   public void updateQueueLength(PointInTime currentTime, int delta) {

      currentQueueLength += delta;
      queueLength.update(currentQueueLength);
   }

   public void updateTransferTime(double sample) {
      transferTime.update(sample);
   }

   public void updateWaitingTime(double sample) {
      waitingTime.update(sample);
   }
}
