/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.multiplexers;

import ikr.simlib.entities.Entity;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.Port;
import ikr.simlib.ports.output.OutputPort;

/**
 * The base class <code>Multiplexer</code> and its derived classes {@link ikr.simlib.entities.multiplexers.StdMultiplexer}, and
{@link ikr.simlib.entities.multiplexers.PriorityMultiplexer} represent components with multiple input ports
({@link ikr.simlib.ports.input.InputPort}) and one output port ({@link ikr.simlib.ports.output.OutputPort}). Except from {@link ikr.simlib.entities.multiplexers.PriorityMultiplexer},
all inputs are scanned in a round-robin scheduling algorithm. In {@link ikr.simlib.entities.multiplexers.StdMultiplexer}
the initial number of input ports is given on construction. Inputs are named: "input 1"
... "input N". Further ports may be added via {@link #addPort(Port)}.<br>
<br>
- Arriving <code>messageIndication()</code> signals are directly delivered to the output port.<br>
<br>
- If the recipient calls <code>isMessageAvailable()</code>, the request will be passed on to the
input port, which cyclically follows the input port from which the last message was
retrieved ("fairness").<br>
<br>
- {@link ikr.simlib.entities.multiplexers.StdMultiplexer} provides the methods <code>addPort()</code> and <code>removePort()</code> to create
and delete an input port named "input 1", "input 2", ... .<br>
<br>
- {@link ikr.simlib.entities.multiplexers.PriorityMultiplexer} is derived from {@link ikr.simlib.entities.multiplexers.StdMultiplexer} and overwrites the
method <code>isMessageAvailable()</code> in that way that it will always start its port scan with
the first input port. Although this prioritization of the input ports is only valid in "active
receiver" mode.
 */
public abstract class Multiplexer extends Entity {

   protected final OutputPort outputPort;

   protected Port currentPort;
   protected Port lastPort;

   public Multiplexer(SimNode ownNode) {
      super(ownNode);

      outputPort = new OutputPort(this) {
         @Override
         protected Message handleGetMessage() {
            lastPort = currentPort;
            return currentPort.getMessage();
         }

         @Override
         protected boolean handleIsMessageAvailable() {
            return Multiplexer.this.handleIsMessageAvailable(this);
         }
      };
   }

   protected abstract boolean handleIsMessageAvailable(Port port);

   public void handleMessageIndication(Port port, Message msg) {
      currentPort = port;
      outputPort.messageIndication(msg);
   }
}
