/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.multiplexers;

import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.Port;
import ikr.simlib.ports.input.InputPort;

import java.util.ArrayList;
import java.util.List;

/**
 * In StdMultiplexer the initial number of input ports is given on construction.
 * Inputs are named: "input 1" to "input N". Further ports may be added via
 * <code>AddPort</code> and removed via <code>RemovePort</code>.
 */
public class StdMultiplexer extends Multiplexer {

   protected final List<InputPort> inputPorts;

   public StdMultiplexer(int noOfPorts, SimNode ownNode) {
      super(ownNode);

      inputPorts = new ArrayList<InputPort>(noOfPorts);

      for (int i = 0; i < noOfPorts; i++) {
         addPort(InputPort.INPUT_PORT_NAME + " " + (i + 1));
      }
   }

   public StdMultiplexer(SimNode ownNode) {
      this(0, ownNode);
   }

   @Override
   public boolean handleIsMessageAvailable(Port port) {
      int i, j;

      for (i = 0; i < inputPorts.size(); i++) {
         if (inputPorts.get(i) == lastPort) {
            i++;
            break;
         }
      }

      if (i == inputPorts.size()) {
         i = 0;
      }

      j = i;
      do {
         if (inputPorts.get(j).isMessageAvailable()) {
            currentPort = inputPorts.get(j);
            return true;
         }
         j = (j + 1) % inputPorts.size();
      } while (j != i);

      return false;
   }

   public void addPort(String portName) {
      inputPorts.add(new InputPort(this, portName) {
         @Override
         protected void handleMessageIndication(Message msg) {
            StdMultiplexer.this.handleMessageIndication(this, msg);
         }
      });
   }

   /** Use "input <++N>" as port name */
   public String addPort() {
      String portName = InputPort.INPUT_PORT_NAME + " " + (inputPorts.size() + 1);
      addPort(portName);
      return portName;
   }

   public void removePort(String portName) {
      for (int i = 0; i < inputPorts.size(); i++) {
         InputPort p = inputPorts.get(i);

         if (p.getLocalName().equals(portName)) {
            if (p == lastPort) {
               lastPort = inputPorts.get((i + 1) % inputPorts.size());
            }
            inputPorts.remove(i);
            break;
         }
      }
   }
}
