/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.phases;

import ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager;
import ikr.simlib.control.simulationcontrol.callbacks.StartTransientPhaseCallbackI;
import ikr.simlib.entities.Entity;
import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.input.PolledInputPort;
import ikr.simlib.ports.output.SynchronousOutputPort;

/**
 * A clocked gate is a standard component that can be used to gate messages. The
 * gate will be opened in a cyclic manner and all available messages are
 * transferred at that time. The cycle time can be specified in the constructor.
 */
public class ClockedGate extends Entity {

   protected final PolledInputPort inputPort;
   protected final SynchronousOutputPort outputPort;

   protected final Duration cycleTime;

   protected final CalendarCallback calendarCallback = new CalendarCallback() {
      @Override
      public void execute() {
         while (inputPort.isMessageAvailable()) {
            Message message = inputPort.getMessage();
            outputPort.sendMessage(message);
         }

         Calendar.getInstance().postEvent(this, null, Calendar.getInstance().getSystemTime().plus(cycleTime));
      }
   };

   public ClockedGate(Duration cycleTime, SimNode ownNode) {
      super(ownNode);

      inputPort = new PolledInputPort(this);
      outputPort = new SynchronousOutputPort(this);
      this.cycleTime = cycleTime;

      SimulationControlManager.getInstance().registerStartTransientPhaseCallback(new StartTransientPhaseCallbackI() {
         @Override
         public void handleStartTransientPhase() {
            ClockedGate.this.handleStartTransientPhase();
         }
      });
   }

   public void handleStartTransientPhase() {
      Calendar.getInstance().postEvent(calendarCallback, null, Calendar.getInstance().getSystemTime().plus(cycleTime));
   }
}
