/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.events.calendar;

import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.EventToken;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.exceptions.SimulationException;

import java.util.TreeSet;

/**
 * Calendar which relies on a {@link TreeSet} to keep the events in order.
 * <p>
 * Each time an event is posted, an associated Entry object is created. Entry
 * objects are comarable by time, priority and ID. The ID is an artificial
 * attribute which discriminates two entries with the same time and priority.
 * This
 * <ul>
 * <li>prohibits that these entries overwrite each other in the set</li>
 * <li>ensures that events with same time and priority are processed in
 * insertion order</li>
 * </ul>
 * <p>
 * The data structure used in this class requires only two objects to be created
 * for each event that is placed in the calendar ({@code TreeSetCalendar.Entry}
 * and {@code TreeMap.Entry}). {@link SimpleOrderedCalendar}, for comparison,
 * requires two additional objects ({@code LinkedList} and
 * {@code LinkedList.Entry}).
 * <p>
 * The complexity of the
 * {@link #postEvent(CalendarCallback, CalendarCallback, PointInTime, int)} and
 * {@link EventToken#cancel()} methods is log(n), where n is the number of
 * events in the calendar.
 */
public class OrderedTreeSetCalendar extends Calendar {
   private static final long INITIAL_ID = Long.MIN_VALUE + 1;
   private static long nextID = INITIAL_ID;

   class Entry implements EventToken, Comparable<Entry> {
      public final PointInTime time;
      public final int priority;
      public final long id;

      private final CalendarCallback process;
      private final CalendarCallback cancel;

      private boolean executed;
      private boolean cancelled;

      private Entry(CalendarCallback process, CalendarCallback cancel, PointInTime scheduledTime, int priority) {
         this.time = scheduledTime;
         this.priority = priority;
         this.id = nextID++;
         this.process = process;
         this.cancel = cancel;
         this.executed = false;
         this.cancelled = false;

         if (id < INITIAL_ID)
            throw new SimulationException("ID overflow");
      }

      void callProcess(PointInTime t) {
         assert t.equals(time);
         assert isPending();
         executed = true;
         if (process != null)
            process.execute();
         assert !isPending();
      }

      @Override
      public void cancel() {
         removeEntryFromTimeLine(this);
         callCancel();
      }

      void callCancel() {
         assert isPending();
         cancelled = true;
         if (cancel != null)
            cancel.execute();
         assert !isPending();
      }

      @Override
      public boolean isPending() {
         return !(hasBeenExecuted() || hasBeenCancelled());
      }

      @Override
      public boolean hasBeenExecuted() {
         return executed;
      }

      @Override
      public boolean hasBeenCancelled() {
         return cancelled;
      }

      @Override
      public int compareTo(Entry o) {
         if (time.compareTo(o.time) != 0)
            return time.compareTo(o.time);
         else if (priority < o.priority)
            return 1;
         else if (priority > o.priority)
            return -1;// elements with higher priority go first
         else if (id < o.id)
            return -1;// elements with higher priority go first
         else if (id > o.id)
            return 1;
         else
            return 0;
      }

      @Override
      public int hashCode() {
         final int prime = 31;
         int result = 1;
         result = prime * result + (int) (id ^ (id >>> 32));
         result = prime * result + priority;
         result = prime * result + time.hashCode();
         return result;
      }

      @Override
      public boolean equals(Object obj) {
         if (this == obj)
            return true;
         if (obj == null)
            return false;
         if (!(obj instanceof Entry))
            return false;
         Entry other = (Entry) obj;
         if (id != other.id)
            return false;
         if (priority != other.priority)
            return false;
         if (!time.equals(other.time))
            return false;
         return true;
      }
   }

   private final TreeSet<Entry> timeline = new TreeSet<Entry>();

   @Override
   public EventToken postEvent(CalendarCallback process, CalendarCallback cancel, PointInTime eventTime, int eventPriority) {
      final Entry e = new Entry(process, cancel, eventTime, eventPriority);
      timeline.add(e);
      return e;
   }

   private void removeEntryFromTimeLine(Entry entry) {
      final boolean removed = timeline.remove(entry);
      if (!removed)
         throw new SimulationException("entry not found in timeline");
   }

   @Override
   public void cancelAllEvents() {
      for (Entry e : timeline)
         e.callCancel();
      timeline.clear();
   }

   @Override
   public boolean isEventAvailable() {
      return !timeline.isEmpty();
   }

   @Override
   public PointInTime peekNextEventTime() {
      if (timeline.isEmpty())
         throw new SimulationException("there is no event in the calendar");

      final Entry entry = timeline.first();
      assert entry.time.isAfterOrEqual(systemTime);
      assert entry.isPending();

      return entry.time;
   }

   @Override
   public void processNextEvent() {
      if (timeline.isEmpty())
         throw new SimulationException("there is no event in the calendar");

      final Entry entry = timeline.pollFirst();
      assert entry.time.isAfterOrEqual(systemTime);
      assert entry.isPending();

      // update number of events with same time, in case we advance in time
      // value is updated in the statistic
      if (eventsWithSameTimeStat != null && entry.time.isAfter(systemTime)) {
         eventsWithSameTimeStat.update((double) concurrentEvents);
         concurrentEvents = 1;
      } else if (entry.time.equals(systemTime)) {
         concurrentEvents++;
      }
      systemTime = entry.time;
      entry.callProcess(systemTime);
   }

}
