/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.events.calendar;

import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.EventToken;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.exceptions.SimulationException;

import java.util.LinkedList;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * A calendar implementation that uses java collection for the sake of
 * simplicity.
 */
public abstract class SimpleCalendar extends Calendar {

   /**
    * the time line assigns lists of events to their time instances
    */
   protected final SortedMap<PointInTime, LinkedList<Entry>> timeline;

   public SimpleCalendar() {
      super();

      timeline = new TreeMap<PointInTime, LinkedList<Entry>>();
   }

   /**
    * returns the first event of the earliest event list
    */
   private Entry popNextEvent() {

      if (!isEventAvailable())
         throw new SimulationException("there is no event in the calendar");

      // get first element on the time line
      final PointInTime earliestTime = timeline.firstKey();

      // lookup event list
      final LinkedList<Entry> entries = timeline.get(earliestTime);
      assert entries.size() > 0;

      // get event
      final Entry entry = entries.removeFirst();

      // remove time line entry if empty
      if (entries.size() == 0)
         timeline.remove(earliestTime);

      assert earliestTime.isAfterOrEqual(systemTime);
      // update number of events with same time, in case we advance in time
      // value is updated in the statistic
      if (eventsWithSameTimeStat != null && earliestTime.isAfter(systemTime)) {
         eventsWithSameTimeStat.update((double) concurrentEvents);
         concurrentEvents = 1;
      } else if (earliestTime.equals(systemTime)) {
         concurrentEvents++;
      }
      systemTime = earliestTime;

      return entry;
   }

   /**
    * cancels all events
    */
   @Override
   public void cancelAllEvents() {
      for (java.util.Map.Entry<PointInTime, LinkedList<Entry>> mapEntry : timeline.entrySet())
         for (Entry entry : mapEntry.getValue())
            entry.callCancel(mapEntry.getKey());
      timeline.clear();
   }

   void removeEntryFromTimeLine(Entry entry, PointInTime scheduledTime) {
      final LinkedList<Entry> entries = timeline.get(scheduledTime);

      if (entries == null)
         throw new SimulationException("no entries in timeline at t=" + scheduledTime);

      final boolean removed = entries.remove(entry);

      if (!removed)
         throw new SimulationException("entry not found at t=" + scheduledTime);

      if (entries.isEmpty())
         timeline.remove(scheduledTime);
   }

   /**
    * processes the next event
    */
   @Override
   public void processNextEvent() {
      final Entry entry = popNextEvent();
      entry.callProcess(systemTime);
   }

   @Override
   public boolean isEventAvailable() {
      return timeline.size() > 0;
   }

   @Override
   public PointInTime peekNextEventTime() {
      PointInTime earliestTime;

      if (isEventAvailable()) {
         return timeline.firstKey();
      }

      return PointInTime.ZERO;
   }

   @Override
   public PointInTime getSystemTime() {
      return systemTime;
   }

   protected class Entry implements Comparable<Entry>, EventToken {
      private final CalendarCallback process;
      private final CalendarCallback cancel;
      private final int priority;
      private final PointInTime scheduledTime;

      private boolean executed;
      private boolean cancelled;

      protected Entry(CalendarCallback process, CalendarCallback cancel, PointInTime scheduledTime, int priority) {
         this.process = process;
         this.cancel = cancel;
         this.scheduledTime = scheduledTime;
         this.priority = priority;
         this.executed = false;
         this.cancelled = false;
      }

      @Override
      public int compareTo(Entry o) {
         // elements with higher priority go first
         return Integer.compare(o.priority, priority);
      }

      void callProcess(PointInTime t) {
         executed = true;
         assert t.equals(scheduledTime);
         if (process != null)
            process.execute();
      }

      @Override
      public void cancel() {
         removeEntryFromTimeLine(this, scheduledTime);
         callCancel(scheduledTime);
      }

      void callCancel(PointInTime t) {
         cancelled = true;
         assert t.equals(scheduledTime);
         if (cancel != null)
            cancel.execute();
      }

      @Override
      public boolean isPending() {
         return !(hasBeenExecuted() || hasBeenCancelled());
      }

      @Override
      public boolean hasBeenExecuted() {
         return executed;
      }

      @Override
      public boolean hasBeenCancelled() {
         return cancelled;
      }
   }
}
