/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.events.calendar;

import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.EventToken;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.random.RandomNumberGenerator;
import ikr.simlib.random.StdRandomNumberGenerator;

import java.util.Collections;
import java.util.LinkedList;
import java.util.Random;

/**
 * A calendar implementation that uses java collection for the sake of
 * simplicity. Events with same time time stamp and priority are executed in
 * random (but reproducible) order.
 */
public class SimpleRandomCalendar extends SimpleCalendar {

   private final RandomNumberGenerator shufflingRNG;

   public SimpleRandomCalendar() {
      super();

      this.shufflingRNG = new StdRandomNumberGenerator();
   }

   @Override
   public EventToken postEvent(CalendarCallback process, CalendarCallback cancel, PointInTime eventTime, int eventPriority) {
      if (systemTime.isAfter(eventTime))
         throw new SimulationException("Event Time < System Time");

      LinkedList<Entry> entries;

      // find list to append the event
      entries = timeline.get(eventTime);
      if (entries == null) {
         // add new list
         entries = new LinkedList<Entry>();
         timeline.put(eventTime, entries);
      }

      // append event, descending priorities
      final Entry e = new Entry(process, cancel, eventTime, eventPriority);
      entries.add(e);
      if (entries.size() > 1) {
         // shuffle to ensure random ordering
         Collections.shuffle(entries, new Random(Double.doubleToLongBits(shufflingRNG.next())));
         // sort to ensure descending priorities
         Collections.sort(entries);
      }

      return e;
   }
}
