/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.events.time;

import ikr.simlib.exceptions.SimulationException;

/**
 * Predefined units to use in combination with {@link TimeBase}.
 * 
 * The enum members can be used to easily set the smallest time unit that can be
 * expressed by the classes {@link PointInTime} and {@link Duration}.
 * 
 * @author sscholz
 *
 */
public enum TimeUnits {
   ps {
      @Override
      public double getTimeBase() {
         return 1000000000000.;
      }

      @Override
      public String getName() {
         return "PicoSeconds";
      }

      @Override
      public String getShortName() {
         return "ps";
      }
   },
   ns {
      @Override
      public double getTimeBase() {
         return 1000000000.;
      }

      @Override
      public String getName() {
         return "NanoSeconds";
      }

      @Override
      public String getShortName() {
         return "ns";
      }
   },
   us {
      @Override
      public double getTimeBase() {
         return 1000000.;
      }

      @Override
      public String getName() {
         return "MicroSeconds";
      }

      @Override
      public String getShortName() {
         return "us";
      }
   },
   ms {
      @Override
      public double getTimeBase() {
         return 1000.;
      }

      @Override
      public String getName() {
         return "MilliSeconds";
      }

      @Override
      public String getShortName() {
         return "ms";
      }
   },
   s {
      @Override
      public double getTimeBase() {
         return 1.;
      }

      @Override
      public String getName() {
         return "Seconds";
      }

      @Override
      public String getShortName() {
         return "s";
      }
   },
   m {
      @Override
      public double getTimeBase() {
         return 1. / 60;
      }

      @Override
      public String getName() {
         return "Minute";
      }

      @Override
      public String getShortName() {
         return "m";
      }
   },
   h {
      @Override
      public double getTimeBase() {
         return 1. / 3600;
      }

      @Override
      public String getName() {
         return "Hour";
      }

      @Override
      public String getShortName() {
         return "h";
      }
   },
   d {
      @Override
      public double getTimeBase() {
         return 1. / 86400;
      }

      @Override
      public String getName() {
         return "Day";
      }

      @Override
      public String getShortName() {
         return "d";
      }
   };

   public double getTimeBase() {
      throw new AbstractMethodError();
   }

   public String getName() {
      throw new AbstractMethodError();
   }

   public String getShortName() {
      throw new AbstractMethodError();
   }

   public static TimeUnits getTimeUnitByShortName(String timeUnit) {
      if (timeUnit.equals(ps.getShortName())) {
         return ps;
      } else if (timeUnit.equals(ns.getShortName())) {
         return ns;
      } else if (timeUnit.equals(us.getShortName())) {
         return us;
      } else if (timeUnit.equals(ms.getShortName())) {
         return ms;
      } else if (timeUnit.equals(s.getShortName())) {
         return s;
      } else if (timeUnit.equals(m.getShortName())) {
         return m;
      } else if (timeUnit.equals(h.getShortName())) {
         return h;
      } else if (timeUnit.equals(d.getShortName())) {
         return d;
      } else {
         throw new SimulationException("Time unit not found");
      }
   }

}
