/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.messages;

/**
 * The center of the simulation is the exchange of messages (that represent data
 * packets) between individual model components.<br>
 * All messages are derived from the class <code>Message</code> and therefore
 * inherit the following characteristics:<br>
 * <br>
 * - Messages have a length which is set to one by default. This length can be
 * set by the method {@link #setLength(long)} and retrieved by
 * {@link #getLength()}.<br>
 * <br>
 * Creating messages usually takes place in a generator (
 * {@link ikr.simlib.entities.generators.Generator}). Generally a generator
 * contains an element of the class
 * {@link ikr.simlib.messages.factories.MessageFactory}, which creates new
 * message objects. By using the port concept, a message is passed from one
 * model component to the next, whereby a reference to each message is
 * additionally passed on during the process of the handshake protocol between
 * two components.<br>
 * <br>
 * <h2>Special Message Types</h2> Within an individual model, it might be
 * necessary to have a message with special members.
 * {@link ikr.simlib.messages.LabelMessage} and
 * {@link ikr.simlib.messages.IPPacket} are two implemented examples.
 * {@link ikr.simlib.messages.LabelMessage} has one additional label whereas
 * {@link ikr.simlib.messages.IPPacket} includes the most important fields of
 * the IP header.
 */
public class Message {

   public static final int DEFAULT_MESSAGE_LENGTH = 1;

   private long length;

   public Message(long length) {
      this.length = length;
   }

   public Message() {
      this(DEFAULT_MESSAGE_LENGTH);
   }

   public long getLength() {
      return length;
   }

   public void setLength(long length) {
      this.length = length;
   }

   @Override
   public String toString() {
      return "Message length = " + getLength();
   }
}
