/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.meters;

import ikr.simlib.entities.Entity;
import ikr.simlib.messages.Message;
import ikr.simlib.ports.Port;
import ikr.simlib.ports.PortMonitor;

/**
 * Base class for all meters which are connected to two ports (e.g. transfer
 * time measurement).
 */
public abstract class TwoPointMeter extends Meter {

   protected final PortMonitor toPortHandler;
   protected final PortMonitor fromPortHandler;

   protected TwoPointMeter() {
      toPortHandler = new PortMonitor() {

         @Override
         public void monitorGetMessage(Message message, Port port) {
            evaluateMessageToPort(message, port);
         }

         @Override
         public void monitorIsMessageAvailable(boolean messageAvailable, Port port) {
         }

         @Override
         public void monitorMessageIndication(Message message, Port port) {
         }
      };

      fromPortHandler = new PortMonitor() {
         @Override
         public void monitorGetMessage(Message message, Port port) {
            evaluateMessageFromPort(message, port);
         }

         @Override
         public void monitorIsMessageAvailable(boolean messageAvailable, Port port) {
         }

         @Override
         public void monitorMessageIndication(Message message, Port port) {
         }
      };
   }

   /**
    * Attaches this meter to the specified port. All messages traveling accross
    * that port are provided to the meter for it's measurements.
    * <p>
    * The port is assumed to be a from-port. See the derived meter class for a
    * definition of from-port and to-port.
    * 
    * @param entity
    *           entity which owns the port
    * @param localPortName
    *           port name
    */
   public void attachFromPort(Entity entity, String localPortName) {
      Port port = entity.getPortByName(localPortName);
      attachFromPort(port);
   }

   /**
    * Attaches this meter to the specified port. All messages traveling accross
    * that port are provided to the meter for it's measurements.
    * <p>
    * The port is assumed to be a from-port. See the derived meter class for a
    * definition of from-port and to-port.
    * 
    * @param port
    *           the port to connect to
    */
   public void attachFromPort(Port port) {
      port.addPortMonitor(fromPortHandler);
   }

   /**
    * Attaches this meter to the specified port. All messages traveling accross
    * that port are provided to the meter for it's measurements.
    * <p>
    * The port is assumed to be a to-port. See the derived meter class for a
    * definition of from-port and to-port.
    * 
    * @param entity
    *           entity which owns the port
    * @param localPortName
    *           port name
    */
   public void attachToPort(Entity entity, String localPortName) {
      Port port = entity.getPortByName(localPortName);
      attachToPort(port);
   }

   /**
    * Attaches this meter to the specified port. All messages traveling accross
    * that port are provided to the meter for it's measurements.
    * <p>
    * The port is assumed to be a to-port. See the derived meter class for a
    * definition of from-port and to-port.
    * 
    * @param port
    *           the port to connect to
    */
   public void attachToPort(Port port) {
      port.addPortMonitor(toPortHandler);
   }

   protected abstract void evaluateMessageFromPort(Message msg, Port port);

   protected abstract void evaluateMessageToPort(Message msg, Port port);
}
