/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.meters.time;

import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.messages.Message;
import ikr.simlib.meters.TwoPointMeter;
import ikr.simlib.ports.Port;
import ikr.simlib.statistics.UnweightedSampleCollector;

import java.util.Map;
import java.util.WeakHashMap;

/**
 * A time meter can be attached to to two or more ports to measure message
 * transfer times. The {@link #useMessage(Message)} method can be overridden to
 * use only those messages for measurement that satisfy certain criteria, e.g.,
 * based on message content, or message type.
 * <p>
 * The figure below depicts the principle of a transfer time meter.
 * <p>
 * <center>
 * 
 * <pre>
 * <img src="{@docRoot}/ikr/simlib/doc-files/transferTimeMeterP84.svg" width="600" alt="Principle of
 * a transfer time meter"><br>
 * Principle of a transfer time meter
 * </pre>
 * 
 * </center>
 */
public class TimeMeter extends TwoPointMeter {

   /**
    * creates a new {@link TimeMeter} with a statistic that accepts Duration
    * objects
    * 
    * @param statistic
    *           statistic to fill with the values
    * @return the new {@link TimeMeter}
    */
   public static TimeMeter createWithDurationStatistic(UnweightedSampleCollector<Duration> statistic) {
      return new TimeMeter(statistic);
   }

   /**
    * creates a new {@link TimeMeter} with a statistic that accepts Double
    * values. The measured durations are converted to seconds and then filled
    * into the statistic.
    * 
    * @param statistic
    *           statistic to fill with the values
    * @return the new {@link TimeMeter}
    */
   public static TimeMeter createWithDoubleStatistic(final UnweightedSampleCollector<? super Double> statistic) {
      return new TimeMeter(new UnweightedSampleCollector<Duration>() {
         @Override
         public void update(Duration value) {
            statistic.update(value.toSeconds());
         }
      });
   }

   /** this map logs the time when the message passes the from port */
   private final Map<Message, PointInTime> timeStamps;
   /** statistic to update with the durations */
   private final UnweightedSampleCollector<Duration> statistic;

   protected TimeMeter(UnweightedSampleCollector<Duration> statistic) {
      this.timeStamps = new WeakHashMap<Message, PointInTime>();
      this.statistic = statistic;
   }

   /** put from-port time stamp into the map */
   protected void addTimeStamp(Message message, PointInTime timeStamp) {
      timeStamps.put(message, timeStamp);
   }

   /** remove from-port time stamp from the map */
   protected PointInTime getAndRemoveTimeStamp(Message message) {
      return timeStamps.remove(message);
   }

   /**
    * called when a message was seen at one of the from-ports and at one of the
    * to-ports
    * 
    * @param msg
    *           the message
    * @param fromPortTime
    *           the time the message was seen at the from-port
    * @param toPortTime
    *           the time the message was seen at the to-port
    */
   protected void evaluateSample(Message msg, PointInTime fromPortTime, PointInTime toPortTime) {
      statistic.update(fromPortTime.until(toPortTime));
   }

   /**
    * filters messages that should be used for transfer time measurement
    * 
    * @param msg
    *           the observed message
    */
   protected boolean useMessage(Message msg) {
      return true;
   }

   @Override
   protected void evaluateMessageFromPort(Message msg, Port port) {
      if (useMessage(msg)) {
         addTimeStamp(msg, Calendar.getInstance().getSystemTime());
      }
   }

   @Override
   protected void evaluateMessageToPort(Message msg, Port port) {
      if (useMessage(msg)) {
         PointInTime fromPortTime = getAndRemoveTimeStamp(msg);

         if (fromPortTime != null)
            evaluateSample(msg, fromPortTime, Calendar.getInstance().getSystemTime());
      }
   }

}
