/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.model;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;

public class SimNodeManager {

   private static SimNodeManager INSTANCE = null;

   public static SimNodeManager getInstance() {
      if (INSTANCE == null)
         throw new SimulationException("SimNodeManager has not been set yet");

      return INSTANCE;
   }

   /**
    * The root of the model hierarchy
    */
   private final SimNode rootNode;

   public SimNodeManager(SimNode rootNode) {
      INSTANCE = this;
      this.rootNode = rootNode;
   }

   public SimNode getRootNode() {
      if (rootNode == null)
         throw new ModelException("The root of the model hierarchy has not been set yet.");

      return rootNode;
   }

   /**
    * Writes the result items.
    * 
    * @param type
    *           the result type
    * @param writer
    *           result writer
    * @throws PrintException
    */
   public void doPrintResults(ResultType type, PrintResultWriter writer) throws PrintException {
      printResults(rootNode, type, writer);
   }

   private void printResults(SimNode simNode, ResultType type, PrintResultWriter writer) throws PrintException {
      if (simNode.subtreeHasPrintHandler()) {
         if (simNode.getPrintHandler() != null)
            writer.writeStartNode(simNode.getPrintHandler().getClass().getSimpleName(), simNode.getName());
         else
            writer.writeStartNode(null, simNode.getName());

         // go through the model node tree
         for (SimNode child : simNode.getChildren())
            printResults(child, type, writer);

         // finally go through the own children to print
         if (simNode.getPrintHandler() != null)
            simNode.getPrintHandler().printResults(type, writer);

         writer.writeEndNode();
      }
   }

   /**
    * Triggers the exporting of the state.
    * 
    * @throws PrintException
    */
   public void doExportBatch(ResultWriter writer) throws PrintException {
      exportBatch(rootNode, writer);
   }

   private void exportBatch(SimNode simNode, ResultWriter writer) throws PrintException {
      if (simNode.subtreeHasExportHandler()) {
         if (simNode.getExportHandler() != null)
            writer.writeStartNode(simNode.getExportHandler().getClass().getSimpleName(), simNode.getName());
         else
            writer.writeStartNode(null, simNode.getName());

         // go through the model node tree
         for (SimNode child : simNode.getChildren())
            exportBatch(child, writer);

         // finally go through the children of the ModelNode
         if (simNode.getExportHandler() != null)
            simNode.getExportHandler().exportBatch(writer);

         writer.writeEndNode();
      }
   }

   /**
    * Triggers the importing of the state.
    * 
    * @throws ImportException
    */
   public void doImportBatch(ResultReader reader) throws ImportException {
      importBatch(rootNode, reader);
   }

   private void importBatch(SimNode simNode, ResultReader reader) throws ImportException {
      if (simNode.subtreeHasExportHandler()) {
         reader.readStartNode(null, simNode.getName());

         // go through the model node tree
         for (SimNode child : simNode.getChildren())
            importBatch(child, reader);

         // finally go through the children of the ModelNode
         if (simNode.getExportHandler() != null)
            simNode.getExportHandler().importBatch(reader);

         reader.readEndNode();
      }
   }
}
