/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.parameters.fileparser;

import ikr.simlib.parameters.ParameterException;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.fileparser.input.Line;
import ikr.simlib.parameters.fileparser.input.LineReader;
import ikr.simlib.parameters.fileparser.tokens.EndArrayToken;
import ikr.simlib.parameters.fileparser.tokens.ScalarToken;
import ikr.simlib.parameters.fileparser.tokens.StartArrayToken;
import ikr.simlib.parameters.fileparser.tokens.Token;
import ikr.simlib.parameters.fileparser.tokens.TokenArrayIterator;
import ikr.simlib.parameters.fileparser.tokens.TokenExpander;
import ikr.simlib.parameters.values.Array;
import ikr.simlib.parameters.values.Scalar;
import ikr.simlib.parameters.values.Value;

import java.io.InputStream;
import java.io.PrintWriter;
import java.util.LinkedList;
import java.util.List;

public class FileParser {

   public static final char ASSIGNMENT_CHARACTER = '=';
   public static final char OPENING_ARRAY_BRACKET = '[';
   public static final char CLOSING_ARRAY_BRACKET = ']';
   public static final char COMMENT = '#';

   public static void parseInputStream(InputStream input, Parameters parameters, PrintWriter parserLogStream) {
      // instantiate line reader
      final LineReader reader = new LineReader(input);

      // read all lines
      while (reader.hasNext()) {
         parseLine(reader.next(), parameters);
      }
   }

   /**
    * Checks the structure of a single line and adds the parameter to the
    * database.
    */
   private static void parseLine(Line line, Parameters parameters) {
      // remove white spaces
      String text = line.text.trim();

      // skip empty lines and comments
      if (!text.isEmpty() && !(text.charAt(0) == COMMENT)) {

         // split up line into tokens
         List<Token> tokens = TokenExpander.getParts(text, OPENING_ARRAY_BRACKET, CLOSING_ARRAY_BRACKET, COMMENT);

         // there must be at least 3 tokens
         if (tokens.size() < 3) {
            error(line, "line doesn't contain enough tokens");
         }

         // first token must not be an array
         String key = null;
         Token token1 = tokens.get(0);
         if (!(token1 instanceof ScalarToken)) {
            error(line, "malformed key");
         } else {
            key = ((ScalarToken) token1).value;
         }

         // second token must be a '=' character
         Token token2 = tokens.get(1);
         if (!(token2 instanceof ScalarToken) && !(((ScalarToken) token2)).value.equals(ASSIGNMENT_CHARACTER)) {
            error(line, "cannot find = character");
         }

         // split tokens into key, assignment character, and value(s)
         List<Token> valueTokens = new LinkedList<Token>();
         for (int i = 2; i < tokens.size(); i++) {
            valueTokens.add(tokens.get(i));
         }

         final TokenArrayIterator iter = new TokenArrayIterator(valueTokens);
         parameters.add(key, expand(line, iter));
         if (iter.hasNext()) {
            error(line, "line contains more tokens than required (a single token with blanks should be quoted)");
         }
      }
   }

   /**
    * Converts a sequence of tokens into a hierarchical structure of value
    * objects.
    * 
    * @param it
    *           the iterator giving access to the tokens
    * @return the structure of value objects
    */
   public static Value expand(Line line, TokenArrayIterator it) {
      if ((it.peek() instanceof ScalarToken)) {

         // read scalar
         Value value = new Scalar(((ScalarToken) it.get()).value);

         return value;

      } else {

         // read array
         List<Value> arrayValues = new LinkedList<Value>();

         // read [
         if (!(it.get() instanceof StartArrayToken))
            error(line, "array doesn't start with a [");

         // read values
         while (!(it.peek() instanceof EndArrayToken)) {
            arrayValues.add(expand(line, it));
         }

         // read ]
         if (!(it.get() instanceof EndArrayToken))
            error(line, "array doesn't end with a ]");

         // convert list to an Array object
         Value value = new Array(arrayValues);

         return value;
      }

   }

   private static void error(String error) {
      throw new ParameterException(error);
   }

   private static void error(Line line, String error) {
      if (line != null)
         error("error reading line number " + line.number + ": " + error);
      else
         error(error);
   }
}
