/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.parameters.reflection;

import ikr.simlib.parameters.ParameterException;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

public class ReflectionLoader {
   /**
    * This method uses reflection to create a new instance of the class defined
    * as String. It checks that the configured class matches the type which is
    * requested. It looks for a static method which is callable with the
    * specified arguments and uses that to create an instance.
    * <ul>
    * <li>
    * Reflection is rather costly and therefore this method should not be used
    * very often.</li>
    * <li>
    * Since types and therefore type safety can only be checked at compile time,
    * this method may fail at runtime for any mistake in configuration.</li>
    * <li>
    * Primitive types as arguments of the static method are currently not
    * supported</li>
    * </ul>
    * 
    * @param className
    *           the fully qualified name of the class to load
    * @param superClass
    *           the type which the created object should have. The returned
    *           instance is a subclass of this type or implements this
    *           interface, respectively.
    * @param parameters
    *           arguments given to the static create method. It is the users
    *           responsibility to check that there is a constructor with these
    *           arguments in all classes which could be configured to be loaded.
    *           Instantciation fails if there is no constructor which excactly
    *           matches these parameters.
    * @return an instance of the class with the given name, which is a subclass
    *         of the class specified as first parameter.
    */
   public static <T> T newInstance(String className, Class<T> superClass, Object... parameters) {
      final Class<? extends T> cl = getClass(className, superClass);
      final Method createInstanceMethod = getCreateInstanceMethod(cl, superClass, parameters);
      if (createInstanceMethod == null)
         throw new ParameterException("Could not find suitable " + ReflectionConstructable.CREATE_INSTANCE_METHOD_NAME + "() method in class " + cl.getName() + " providing these arguments: " + getTypeString(parameters) + "\nNote that using variable arguments is not supported yet.");
      else
         return createInstance(createInstanceMethod, parameters, superClass, className);
   }

   /**
    * step 1: get the class
    * 
    * @param className
    *           name of the class to load
    * @param superClass
    *           super class or interface which has to be extended / implemented
    * @return the class
    */
   private static <T> Class<? extends T> getClass(String className, Class<T> superClass) {
      try {
         return Class.forName(className).asSubclass(superClass);
      } catch (ClassNotFoundException ex) {
         throw new ParameterException("Couldn't get class for " + className, ex);
      } catch (ClassCastException ex) {
         throw new ParameterException("The configured class " + className + " is not a subclass of the requested class " + superClass.getName(), ex);
      }
   }

   /**
    * step 2: get a createInstance() method with the correct arguments
    * 
    * @param cl
    *           class to load
    * @param superClass
    *           return type of the method has to be a subtype of this class
    * @param parameters
    *           parameters of the static method
    * @return the constructor
    */
   private static <T> Method getCreateInstanceMethod(Class<T> cl, Class<? super T> superClass, Object[] parameters) {
      Method method = null;
      for (Method mIter : cl.getDeclaredMethods()) {
         if (Modifier.isStatic(mIter.getModifiers()) && mIter.getName().equals(ReflectionConstructable.CREATE_INSTANCE_METHOD_NAME)) {
            if (superClass.isAssignableFrom(mIter.getReturnType()) && parametersMatch(parameters, mIter) && mIter.getReturnType() == cl) {
               if (method != null)
                  throw new ParameterException("Multiple constructors of class " + cl.getName() + " are assignable from the arguments " + getTypeString(parameters));
               method = mIter;
            }
         }
      }
      return method;
   }

   /**
    * step 3: create an instance of the class
    * 
    * @param createInstanceMethod
    *           static method to use
    * @param parameters
    *           the parameters to give to the constructor
    * @param superClass
    *           type to which the return value has to be casted
    * @param className
    *           the class name (for error messages)
    * @return the new instance
    */
   private static <T> T createInstance(Method createInstanceMethod, Object[] parameters, Class<T> superClass, String className) {
      try {
         return superClass.cast(createInstanceMethod.invoke(null, parameters));
      } catch (IllegalArgumentException e) {
         throw new ParameterException("Signature of the " + ReflectionConstructable.CREATE_INSTANCE_METHOD_NAME + "() method does not match. This should never happen and is a bug. Please contact SimLib professional for assistance.", e);
      } catch (IllegalAccessException e) {
         throw new ParameterException("Java access control prohibited calling the " + ReflectionConstructable.CREATE_INSTANCE_METHOD_NAME + "() method in " + className + ".", e);
      } catch (InvocationTargetException e) {
         throw new ParameterException("Couldn't instantiate " + className + " due to an InvocationTargetException in the called  " + ReflectionConstructable.CREATE_INSTANCE_METHOD_NAME + "() method", e);
      }
   }

   /**
    * creates a string with the class names of all arguments, for error messages
    */
   private static String getTypeString(Object[] args) {
      StringBuilder sb = new StringBuilder();
      sb.append("[");
      for (Object arg : args)
         sb.append(arg.getClass().getName()).append(", ");
      sb.append("]");
      return sb.toString();
   }

   /**
    * 
    * @param vars
    *           argument values
    * @param cons
    *           a candidate constructor
    * @return true if the constructor is no varArgs constructor and all its
    *         parameters are assignable from the given values.
    */
   private static boolean parametersMatch(Object[] vars, Method m) {
      if (m.isVarArgs()) {
         return false;
      } else if (m.getParameterTypes().length == vars.length) {
         for (int i = 0; i < vars.length; i++)
            if (!m.getParameterTypes()[i].isAssignableFrom(vars[i].getClass()))
               return false;
         // all parameters match
         return true;
      } else {
         return false;
      }
   }

}
