/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.ports.output;

import ikr.simlib.entities.Entity;
import ikr.simlib.messages.Message;
import ikr.simlib.ports.PortException;

/**
 * NonBlockingOutputPort is an output port that sends a message when SendMessage
 * is called. If the previous message is still around when SendMessage is called
 * again, an exception is thrown.
 * 
 * <p>
 * <em>Caution:</em><br>
 * NonBlockingOutputPorts stores the message until the next entity calls
 * HandleGetMessage. When an entity with a NonBlockingOutputPort gathers
 * statistical data, this storing capability has to be considered.
 * <p>
 * Message must be fetched by receiving port <strong>before</strong>
 * SendMessage() is called <strong>again</strong>.
 * </p>
 */
public class NonBlockingOutputPort extends OutputPort {

   protected Message message;

   public NonBlockingOutputPort(Entity owner, String portName) {
      super(owner, portName);
   }

   public NonBlockingOutputPort(Entity owner) {
      this(owner, OUTPUT_PORT_NAME);
   }

   public void sendMessage(Message m) {
      if (message != null) {
         throw new PortException("Previous Message has not been sent");
      }

      message = m; // save it

      messageIndication(m); // tell next port
   }

   @Override
   protected Message handleGetMessage() {
      if (message == null) {
         throw new PortException("No message available");
      }

      Message m = message;
      message = null;

      return m;
   }

   @Override
   protected boolean handleIsMessageAvailable() {
      return message != null;
   }
}
