/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.random;

import ikr.simlib.support.cmdline.ArgumentException;

import java.util.HashSet;
import java.util.Set;

public class IndependentRNGFactory {
   private static int globalSeedIndex = -1;
   private static RandomNumberGeneratorFactory factory = null;
   private static boolean initialized = false;
   private static Set<String> usedSeedStrings = new HashSet<String>();

   /**
    * Use this method to mark the SystemRNG of the RandomNumberGenerator as
    * invalid. After calling this method, all newly created Distributions throw
    * a runtime exception, when not explicitly configured to use another RNG.
    * This method does not influence the distributions that where instantiated
    * before this method was called. Those will still use the SystemRNG.
    */
   public static void disableSystemRNG() {
      RandomNumberGenerator.setSystemRNG(null);
   }

   /**
    * Use this method to mark the SystemRNG of the RandomNumberGenerator as
    * deprecated. After calling this method, all usage of the systemRNG results
    * in warnings on stderr. This method does not influence the distributions
    * that where instantiated before this method was called. Those will still
    * use the SystemRNG.
    */
   public static void wrapSystemRNGWithWarnings() {
      final RandomNumberGenerator oldRNG = RandomNumberGenerator.getSystemRNG();
      RandomNumberGenerator.setSystemRNG(new WarningRNG(oldRNG));
   }

   /**
    * Use this method to set the seed index, which should be given on the
    * command line.
    * 
    * @param seedIndex
    *           seed index - must be >= 0
    */
   public static void initialize(int seedIndex, RandomNumberGeneratorFactory rngFactory) {
      if (initialized)
         throw new RuntimeException("The RNG factory has already been initialized.");
      if (seedIndex < 0)
         throw new ArgumentException("Negative seed indices are not supported.");
      globalSeedIndex = seedIndex;
      factory = rngFactory;
      initialized = true;
   }

   /**
    * Instantiates a new RandomNumberGenerator. Its seed is a combination of the
    * globalSeedIndex and the parameter seedString. Calling this method twice
    * with the same parameter and the same globalSeedIndex configured will
    * return two RNGs which produce the same sequence of pseudo random numbers.
    * 
    * @param seedString
    *           use this to create a seed depending on the module or class which
    *           uses this RNG
    * @return a new instance of StdRandomNumberGenerator
    */
   public static RandomNumberGenerator getNewRNG(String seedString) {
      if (!initialized)
         throw new FactoryNotInitializedException();

      if (usedSeedStrings.contains(seedString))
         throw new RuntimeException("Seed string " + seedString + " used multiple times");
      usedSeedStrings.add(seedString);

      final int seedIndex = seedString.hashCode() + globalSeedIndex;
      return factory.createRNG(seedIndex);
   }

   public static class FactoryNotInitializedException extends RuntimeException {
      private static final long serialVersionUID = 1L;

      public FactoryNotInitializedException() {
         super("The factory has to be initialized with the seed index, " + "which should be given on the command line. ");
      }
   }

   public static class UseIndependentRNGFactoryException extends RuntimeException {
      private static final long serialVersionUID = 1L;

      public UseIndependentRNGFactoryException() {
         super("The usage of the global random number generator is " + "discouraged in this simulation. Use " + "IndependentRNGFactory.getNewRNG(String) to create a new, " + "seeded RNG that is independent of the other RNGs in the system.");
      }
   }

   /**
    * Wraps the SystemRNG, prints warnings
    */
   public static class WarningRNG extends RandomNumberGenerator {
      private final RandomNumberGenerator wrappedRNG;
      private boolean warningPrinted = false;

      public WarningRNG(RandomNumberGenerator wrappedRNG) {
         this.wrappedRNG = wrappedRNG;
      }

      @Override
      public double next() {
         if (!warningPrinted) {
            System.err.println("WARNING: Old random number generator is still in use!");
            warningPrinted = true;
         }
         return wrappedRNG.next();
      }
   }

}
