/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.result;

public class ResultKeyword {

   /**
    * The actual print keyword of this node
    */
   private final String keyword;

   /**
    * Creates a new keyword
    */
   public ResultKeyword(String keyword) {
      this.keyword = keyword.toLowerCase();
   }

   /**
    * A print keyword is equal to a string if they both contain the same text
    * ignoring case.
    * 
    * <p>
    * <em>This brakes the {@link #equals(Object) equals} and {@link #hashCode() hashCode} contract with {@link String}. You can't mix print keywords and strings in maps for example.</em>
    * </p>
    */
   @Override
   public boolean equals(Object obj) {
      if (obj instanceof String) {
         String keywordName = (String) obj;

         return keyword.equalsIgnoreCase(keywordName);
      }

      return super.equals(obj);
   }

   @Override
   public String toString() {
      return keyword;
   }

   /**
    * Returns a version of the given keyword which is suitable as a list
    * keyword.
    */
   public ResultKeyword asList() {
      return new ResultKeyword(asList(keyword.toString()));
   }

   public static String asList(String keyword) {
      return new String(keyword + "list");
   }

   /**
    * Returns a version of the given keyword which is suitable as a export
    * keyword. Note that the keyword itself requires some unique component to
    * prevent deriving classes of accidentally using the same export keyword as
    * the super class.
    * 
    * <p>
    * The keyword will be prefixed by
    * <q>ex_</q>, although this might change.
    * </p>
    */
   public ResultKeyword asExport() {
      return new ResultKeyword(asExport(keyword.toString()));
   }

   public static String asExport(String keyword) {
      return new String("ex_" + keyword);
   }

   /**
    * Returns a version of the given keyword as batch keyword. (These are
    * prefixed with <q>b</q>).
    */
   public ResultKeyword asBatch() {
      return new ResultKeyword("b" + keyword.toString());
   }

   /**
    * Returns a version of the given keyword as confidence interval keyword.
    * (These are prefixed with <q>cint</q>).
    */
   public ResultKeyword asConfidenceInterval() {
      return new ResultKeyword("cint" + keyword.toString());
   }

   /**
    * Returns a version of the given keyword as mean keyword. (These are
    * suffixed with <q>mean</q>).
    */
   public ResultKeyword asMean() {
      return new ResultKeyword(keyword.toString() + "mean");
   }

   /**
    * Returns a version of the given keyword as minimum keyword. (These are
    * prefixed with <q>min</q>).
    */
   public ResultKeyword asMin() {
      return new ResultKeyword("min" + keyword.toString());
   }

   /**
    * Returns a version of the given keyword as maximum keyword. (These are
    * prefixed with <q>max</q>).
    */
   public ResultKeyword asMax() {
      return new ResultKeyword("max" + keyword.toString());
   }

}
