/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.result.compacting;

import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.PrintException;

import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * This class is a wrapper to the ResultWriter interface. It removes method
 * calls that would cause empty XML nodes.
 */
public class CompactingResultWriter implements ResultWriter {

   /**
    * the actual writer
    */
   private final ResultWriter writer;

   /**
    * the queue, that represents the method calls held back
    */
   private final Deque<Node> drag;

   public CompactingResultWriter(ResultWriter writer) {
      this.writer = writer;
      drag = new LinkedList<Node>();
   }

   @Override
   public void close() throws PrintException {
      writer.close();
   }

   @Override
   public void writeDouble(String name, double value) throws PrintException {
      flush();
      writer.writeDouble(name, value);
   }

   @Override
   public void writeLong(String name, long value) throws PrintException {
      flush();
      writer.writeLong(name, value);
   }

   @Override
   public void writeString(String name, String value) throws PrintException {
      flush();
      writer.writeString(name, value);
   }

   @Override
   public void writeEndResults() throws PrintException {
      ascent(NodeType.RESULTS);
   }

   @Override
   public void writeEndNode() throws PrintException {
      ascent(NodeType.NODE);
   }

   @Override
   public void writeEndArray() throws PrintException {
      ascent(NodeType.ARRAY);
   }

   @Override
   public void writeEndArrayItem() throws PrintException {
      ascent(NodeType.ARRAYITEM);
   }

   @Override
   public void writeStartResults() throws PrintException {
      descent(NodeType.RESULTS, null, null, 0);
   }

   @Override
   public void writeStartNode(String type, String name) throws PrintException {
      descent(NodeType.NODE, type, name, 0);
   }

   @Override
   public void writeStartArray(String name) throws PrintException {
      descent(NodeType.ARRAY, null, name, 0);
   }

   @Override
   public void writeStartArrayItem(long index) throws PrintException {
      descent(NodeType.ARRAYITEM, null, null, index);
   }

   /**
    * This method is being called upon descending in the result tree
    */
   private void descent(NodeType nodeType, String type, String name, long index) {
      drag.addLast(new Node(nodeType, type, name, index));
   }

   /**
    * This method is being called upon ascending in the result tree
    * 
    * @throws PrintException
    */
   private void ascent(NodeType node) throws PrintException {

      if (drag.isEmpty()) {

         // we didn't queue method calls, so do the actual method call;
         switch (node) {
         case RESULTS:
            writer.writeEndResults();
            break;
         case NODE:
            writer.writeEndNode();
            break;
         case ARRAY:
            writer.writeEndArray();
            break;
         case ARRAYITEM:
            writer.writeEndArrayItem();
            break;
         }

      } else {

         // we did queue a method call and there was no flush before, so remove
         // the last entry since it's useless
         drag.removeLast();
      }
   }

   /**
    * This method executed all method calls queued so far
    * 
    * @throws PrintException
    */
   private void flush() throws PrintException {

      // do the method calls, that were queued
      Iterator<Node> it = drag.iterator();
      while (it.hasNext()) {
         Node node = it.next();
         switch (node.nodeType) {
         case RESULTS:
            writer.writeStartResults();
            break;
         case NODE:
            writer.writeStartNode(node.type, node.name);
            break;
         case ARRAY:
            writer.writeStartArray(node.name);
            break;
         case ARRAYITEM:
            writer.writeStartArrayItem(node.index);
            break;
         }
      }
      drag.clear();
   }
}
