/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.result.xml;

import static ikr.simlib.result.xml.XMLConstants.DOUBLE_TYPE;
import static ikr.simlib.result.xml.XMLConstants.LONG_TYPE;
import static ikr.simlib.result.xml.XMLConstants.STRING_TYPE;
import static ikr.simlib.result.xml.XMLConstants.VERSION;
import ikr.simlib.result.PathHelper;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.exceptions.ImportException;

import java.io.InputStream;
import java.util.ArrayDeque;
import java.util.Arrays;
import java.util.Deque;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;
import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

public class XMLResultReader implements ResultReader {

   enum State {
      /**
       * Initial state where DTD, XML header and SimulationResults havn't been
       * read yet.
       */
      DOCUMENT,

      /**
       * Completed the tree. Only {@link XMLResultReader#hasNext() hasNext} may
       * be called.
       */
      FINISHED,

      /**
       * Waiting for SimulationResults tag.
       */
      START_RESULTS,

      /**
       * Waiting for SimulationResults end tag.
       */
      END_RESULTS,

      /**
       * Last event was a result item an the current type is long.
       */
      RESULT_LONG,

      /**
       * Last event was a result item an the current type is double.
       */
      RESULT_DOUBLE,

      /**
       * Last event was a result item an the current type is string.
       */
      RESULT_STRING,

      /**
       * Element started. This allows {@link XMLResultReader#getIndex()
       * getIndex} to be called.
       */
      START_ARRAY_ITEM,

      /**
       * Node started and we can call {@link XMLResultReader#getName() getName}
       * and {@link XMLResultReader#getTypeName() getTypeName}.
       */
      START_NODE,

      /**
       * In a node (or element).
       */
      NODE,

      /**
       * Starting an array.
       */
      START_ARRAY,

      /**
       * After the first element we end up here.
       */
      ARRAY
   }

   // QNames of elements and attributes
   private static final QName SIMULATION_RESULTS_QNAME = QName.valueOf(XMLConstants.SIMULATION_RESULTS);
   private static final QName RESULT_NODE_QNAME = QName.valueOf(XMLConstants.NODE);
   private static final QName RESULT_QNAME = QName.valueOf(XMLConstants.RESULT);
   private static final QName ARRAY_QNAME = QName.valueOf(XMLConstants.ARRAY);
   private static final QName ARRAY_ITEM_QNAME = QName.valueOf(XMLConstants.ARRAY_ITEM);

   private static final QName VERSION_ATTRIBUTE_QNAME = QName.valueOf(XMLConstants.VERSION_ATTRIBUTE);
   private static final QName TYPE_ATTRIBUTE_QNAME = QName.valueOf(XMLConstants.TYPE_ATTRIBUTE);
   private static final QName NAME_ATTRIBUTE_QNAME = QName.valueOf(XMLConstants.NAME_ATTRIBUTE);
   private static final QName VALUE_ATTRIBUTE_QNAME = QName.valueOf(XMLConstants.VALUE_ATTRIBUTE);
   private static final QName INDEX_ATTRIBUTE_QNAME = QName.valueOf(XMLConstants.INDEX_ATTRIBUTE);

   /**
    * Maps type attributes to the corresponding result states (int, long,
    * double, string and array).
    */
   private static final Map<String, State> TYPE_STATES;

   static {
      TYPE_STATES = new HashMap<String, State>();
      TYPE_STATES.put(LONG_TYPE, State.RESULT_LONG);
      TYPE_STATES.put(DOUBLE_TYPE, State.RESULT_DOUBLE);
      TYPE_STATES.put(STRING_TYPE, State.RESULT_STRING);
   }

   private final XMLEventReader reader;

   private State state;
   private final Deque<State> stateStack;

   private long index;
   private final Deque<Long> indexStack;
   private final String fileName;
   private final PathHelper pathHelper;

   /**
    * Type of the current result or {@code null}.
    */
   private String type;

   /**
    * Value of the current result item or {@code null}.
    */
   private String value;

   /**
    * Name of the current result or {@code null}.
    */
   private String name;

   public XMLResultReader(InputStream input, String fileNameForErrorMessages) throws ImportException {
      try {
         reader = XMLInputFactory.newInstance().createXMLEventReader(input);
      } catch (XMLStreamException xse) {
         throw new ImportException(fileNameForErrorMessages, "", xse);
      } catch (FactoryConfigurationError fce) {
         throw new ImportException(fileNameForErrorMessages, "", fce);
      }

      this.fileName = fileNameForErrorMessages;
      this.pathHelper = new PathHelper();

      state = State.DOCUMENT;
      stateStack = new ArrayDeque<State>();

      index = -1;
      indexStack = new ArrayDeque<Long>();
   }

   @Override
   public void close() throws ImportException {
      state = State.FINISHED;
   }

   /**
    * Handles the simulation results tag.
    * 
    * @throws ImportException
    *            in case the world comes to an end
    */
   private void handleSimulationResults(StartElement se) throws ImportException {
      assert state == State.DOCUMENT;

      if (!SIMULATION_RESULTS_QNAME.equals(se.getName())) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Wrong start tag for root simulation results. Element: " + se);
      }

      Attribute version = se.getAttributeByName(VERSION_ATTRIBUTE_QNAME);
      if (version == null || !VERSION.equals(version.getValue())) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Wrong result file version. Expected: " + VERSION + ", read: " + (version != null ? version.getValue() : "unknown"));
      }

      state = State.START_RESULTS;
   }

   /**
    * Handles a result node.
    * 
    * <p>
    * The state is set to {@link State#START_NODE}. The stack is left unchanged
    * and has to be setup before this method is called.
    * </p>
    */
   private void handleNode(StartElement se) throws ImportException {

      // check for node tag
      if (!RESULT_NODE_QNAME.equals(se.getName())) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Wrong start tag for result node: " + se);
      }

      // get type attribute, which is optional
      Attribute attributeType = se.getAttributeByName(TYPE_ATTRIBUTE_QNAME);
      if (attributeType != null) {
         type = attributeType.getValue();
      } else {
         type = null;
      }

      // get name attribute, which is mandatory
      Attribute attributeName = se.getAttributeByName(NAME_ATTRIBUTE_QNAME);
      if (attributeName != null) {
         name = attributeName.getValue();
      } else {
         throw new ImportException(fileName, pathHelper.getPathName(), "Namee attribute missing for result node");
      }

      value = null;

      state = State.START_NODE;
   }

   private void handleArray(StartElement se) throws ImportException {

      // check for array tag
      if (!ARRAY_QNAME.equals(se.getName())) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Wrong start tag for node: " + se);
      }

      type = null;

      // get name attribute, which is optional
      Attribute attributeName = se.getAttributeByName(NAME_ATTRIBUTE_QNAME);
      if (attributeName != null) {
         name = attributeName.getValue();
      } else {
         name = null;
      }

      value = null;
      state = State.START_ARRAY;
   }

   private void handleArrayItem(StartElement se) throws ImportException {

      // check for array item tag
      if (!ARRAY_ITEM_QNAME.equals(se.getName())) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Wrong start tag for node: " + se);
      }

      type = null;

      name = null;

      // Although we don't have to give a damn we still check for sanity
      // reasons
      Attribute indexAttr = se.getAttributeByName(INDEX_ATTRIBUTE_QNAME);
      if (indexAttr == null) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Index attribute is missing on array element");
      }

      if (!Long.toString(index).equals(indexAttr.getValue())) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Wrong index of array element. Expected: " + index + ", read: " + indexAttr.getValue());
      }

      value = null;

      state = State.START_ARRAY_ITEM;
   }

   /**
    * Handles a result item or array.
    * 
    * <p>
    * The state is set to {@link State#START_ARRAY} or one of {@code RESULT_...}
    * . The stack is left unchanged and has to be setup before this method is
    * called. The index stack however is modified if the result is an array.
    * </p>
    */
   private void handleResult(StartElement se) throws ImportException {

      // check for result tag
      if (!RESULT_QNAME.equals(se.getName())) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Wrong start tag for result item: " + se);
      }

      // get type attribute, which is mandatory
      Attribute attributeType = se.getAttributeByName(TYPE_ATTRIBUTE_QNAME);
      if (attributeType == null) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Type attribute missing for result item");
      }
      type = attributeType.getValue();

      // Lookup new state
      State newState = TYPE_STATES.get(attributeType.getValue());
      if (newState == null) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Unknown result type: " + attributeType.getValue());
      }

      // get name attribute, which is optional
      Attribute attributeName = se.getAttributeByName(NAME_ATTRIBUTE_QNAME);
      if (attributeName != null) {
         name = attributeName.getValue();
      } else {
         name = null;
      }

      // get value attribute, which is mandatory
      Attribute attributeValue = se.getAttributeByName(VALUE_ATTRIBUTE_QNAME);
      if (attributeValue == null) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Value attribute missing for result");
      }
      // We can't do sanity checks here because the value is really a
      // string and only the caller knows what to expect.
      value = attributeValue.getValue();

      state = newState;
   }

   private void assertName(String expectedName) throws ImportException {
      if (expectedName != null && !expectedName.equals(getName())) {
         throw new ImportException(fileName, pathHelper.getPathName(), "invalid name - expected name \"" + expectedName + "\", but name is \"" + getName() + "\"");
      }
   }

   private void assertState(State... expected) throws ImportException {
      boolean isExpected = false;
      for (State e : expected) {
         if (state == e) {
            isExpected = true;
         }
      }
      if (!isExpected) {
         throw new ImportException(fileName, pathHelper.getPathName(), "invalid state - expected states " + Arrays.toString(expected) + ", but state is " + state);
      }
   }

   private void assertEvent(Event evt, Event... expected) throws ImportException {
      boolean isExpected = false;
      for (Event e : expected) {
         if (evt == e) {
            isExpected = true;
         }
      }
      if (!isExpected) {
         throw new ImportException(fileName, pathHelper.getPathName(), "invalid event - expected events " + Arrays.toString(expected) + ", but event is " + evt);
      }
   }

   @Override
   public double getDouble() throws ImportException {
      assertState(State.RESULT_DOUBLE);

      String _value = getValue();

      try {
         return Double.parseDouble(_value);
      } catch (NumberFormatException nfe) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Invalid double value " + _value, nfe);
      }
   }

   @Override
   public long getIndex() throws ImportException {
      assertState(State.START_ARRAY_ITEM);
      assert index >= 0;
      return index;
   }

   @Override
   public long getLong() throws ImportException {
      assertState(State.RESULT_LONG);

      String _value = getValue();

      try {
         return Long.parseLong(_value);
      } catch (NumberFormatException nfe) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Invalid long value " + _value, nfe);
      }
   }

   @Override
   public String getName() throws ImportException {
      assertState(State.START_NODE, State.START_ARRAY, State.RESULT_LONG, State.RESULT_DOUBLE, State.RESULT_STRING);
      assert name != null;
      return name;
   }

   @Override
   public String getString() throws ImportException {
      assertState(State.RESULT_STRING);
      return getValue();
   }

   @Override
   public String getTypeName() throws ImportException {
      assertState(State.START_NODE, State.RESULT_LONG, State.RESULT_DOUBLE, State.RESULT_STRING);
      return type;
   }

   @Override
   public String getValue() throws ImportException {
      assertState(State.RESULT_LONG, State.RESULT_DOUBLE, State.RESULT_STRING);
      assert value != null;
      return value;
   }

   @Override
   public boolean hasNext() {
      return state != State.FINISHED;
   }

   @Override
   public Event next() throws ImportException {
      if (!hasNext()) {
         throw new ImportException(fileName, pathHelper.getPathName(), "result file has no next node");
      }

      Event evt = null;

      // A lot of magic happens here. The XML events don't map 1:1 to our
      // result events. We have to skip whitespace, comments, processing
      // instructions and the XML/DTD declaration at the beginning.
      eventLoop: while (reader.hasNext()) {
         XMLEvent xevt;

         try {
            xevt = reader.nextEvent();
         } catch (XMLStreamException xse) {
            throw new ImportException(fileName, pathHelper.getPathName(), "Failed to read next XML event", xse);
         }

         switch (xevt.getEventType()) {
         // Whitespace, comments and processing instructions are
         // allowed anywhere.
         case XMLStreamConstants.COMMENT:
         case XMLStreamConstants.PROCESSING_INSTRUCTION:
         case XMLStreamConstants.SPACE:
         case XMLStreamConstants.DTD:
            continue eventLoop;

         case XMLStreamConstants.CHARACTERS:
            if (xevt.asCharacters().isWhiteSpace()) {
               continue eventLoop;
            }

            throw new ImportException(fileName, pathHelper.getPathName(), "Non-whitespace characters are not allowed in the result XML");

            // These are handled by our state machine
         case XMLStreamConstants.START_DOCUMENT:
            assert state == State.DOCUMENT;

            // Wait for simulation results now
            break;

         case XMLStreamConstants.END_DOCUMENT:
            assert state == State.END_RESULTS;

            state = State.FINISHED;

            // We've completed our mission
            try {
               reader.close();
            } catch (XMLStreamException xse) {
               throw new ImportException(fileName, pathHelper.getPathName(), xse);
            }

            evt = Event.END_RESULTS;
            break eventLoop;

         case XMLStreamConstants.START_ELEMENT:
            final StartElement se = xevt.asStartElement();

            switch (state) {
            case DOCUMENT:
               stateStack.push(State.FINISHED);
               handleSimulationResults(se);

               evt = Event.START_RESULTS;
               break;

            case START_RESULTS:
               // Read the root result node

               stateStack.push(State.END_RESULTS);

               handleNode(se);

               evt = Event.START_NODE;
               break;

            case START_ARRAY:
            case ARRAY:
               // Read the next element
               if (state == State.START_ARRAY) {
                  indexStack.push(index);
                  index = 0;
               } else {
                  index += 1;
               }

               stateStack.push(State.ARRAY);

               handleArrayItem(se);

               evt = Event.START_ELEMENT;
               break;

            case START_NODE:
            case START_ARRAY_ITEM:
            case NODE:
               // We read a new result node, an array or an result item
               // now

               stateStack.push(State.NODE);

               if (RESULT_NODE_QNAME.equals(se.getName())) {
                  // A new node begins
                  handleNode(se);

                  evt = Event.START_NODE;

               } else if (ARRAY_QNAME.equals(se.getName())) {

                  handleArray(se);
                  evt = Event.START_ARRAY;

               } else if (RESULT_QNAME.equals(se.getName())) {
                  // Either array or result item
                  handleResult(se);

                  switch (state) {
                  case RESULT_LONG:
                  case RESULT_DOUBLE:
                  case RESULT_STRING:
                     evt = Event.RESULT;
                     break;

                  default:
                     throw new AssertionError();
                  }
               } else {
                  throw new ImportException(fileName, pathHelper.getPathName(), "Unexpected start tag in result node: " + se);
               }

               break;

            default:
               throw new AssertionError("Unexpected state while handling start tag. Element: " + se);
            }

            // evt has to be set by the preceding switch block
            assert evt != null;

            break eventLoop;

         case XMLStreamConstants.END_ELEMENT:
            final EndElement ee = xevt.asEndElement();

            switch (state) {
            case START_ARRAY_ITEM:
            case START_NODE:
            case NODE:
               assert (RESULT_NODE_QNAME.equals(ee.getName()) || ARRAY_ITEM_QNAME.equals(ee.getName()));

               type = null;
               name = null;

               state = stateStack.pop();

               if (state == State.ARRAY) {
                  evt = Event.END_ELEMENT;
               } else {
                  evt = Event.END_NODE;
               }

               break;

            case START_RESULTS:
            case END_RESULTS:
               assert SIMULATION_RESULTS_QNAME.equals(ee.getName());

               state = stateStack.pop();

               evt = Event.END_RESULTS;
               break;

            case START_ARRAY:
               assert ARRAY_QNAME.equals(ee.getName());

               type = null;
               name = null;

               state = stateStack.pop();

               evt = Event.END_ARRAY;
               break;

            case ARRAY:
               assert ARRAY_QNAME.equals(ee.getName());

               type = null;
               name = null;

               index = indexStack.pop();
               state = stateStack.pop();

               evt = Event.END_ARRAY;
               break;

            case RESULT_LONG:
            case RESULT_DOUBLE:
            case RESULT_STRING:
               assert RESULT_QNAME.equals(ee.getName());

               type = null;
               name = null;
               value = null;

               state = stateStack.pop();

               // Continue the event loop, since we're not interested in a
               // end of the current result item
               continue eventLoop;

            default:
               throw new AssertionError("Unexpected state while handling end tag. Element: " + ee);
            }

            assert evt != null;

            break eventLoop;

         default:
            throw new AssertionError("Unexpected XML event");
         }
      }

      // Make sure to die here instead through a NPE some time later
      assert evt != null;

      return evt;
   }

   @Override
   public void readStartResults() throws ImportException {
      Event evt = next();
      assertEvent(evt, Event.START_RESULTS);
   }

   @Override
   public void readEndResults() throws ImportException {
      Event evt = next();
      assertEvent(evt, Event.END_RESULTS);
   }

   @Override
   public void readEndArrayItem() throws ImportException {
      Event evt = next();
      assertEvent(evt, Event.END_ELEMENT);
      pathHelper.leaveSubTree();
   }

   @Override
   public void readEndNode() throws ImportException {
      Event evt = next();
      assertEvent(evt, Event.END_NODE);
      pathHelper.leaveSubTree();
   }

   @Override
   public void readStartNode(String expectedType, String expectedName) throws ImportException {
      pathHelper.enterSubTree(expectedName);

      Event evt = next();
      assertEvent(evt, Event.START_NODE);

      String typeName = getTypeName();
      if (expectedType != null && !expectedType.equals(typeName)) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Read result type \"" + type + "\" doesn't match expected \"" + expectedType + "\"");
      }

      assertName(expectedName);
   }

   @Override
   public void readEndArray() throws ImportException {
      Event evt = next();
      assertEvent(evt, Event.END_ARRAY);
      pathHelper.leaveSubTree();
   }

   @Override
   public void readStartArray(String expectedName) throws ImportException {
      pathHelper.enterSubTree(expectedName);
      Event evt = next();
      assertEvent(evt, Event.START_ARRAY);
      assertName(expectedName);
   }

   @Override
   public Event nextResult() throws ImportException {
      Event evt = next();
      assertEvent(evt, Event.START_NODE, Event.START_ARRAY, Event.RESULT);
      return evt;
   }

   @Override
   public long readStartArrayItem(long expectedIndex) throws ImportException {
      pathHelper.enterSubTree(Long.toString(expectedIndex));

      Event evt = next();
      assertEvent(evt, Event.START_ELEMENT);

      final long indexFromFile = getIndex();
      if (indexFromFile != expectedIndex) {
         throw new ImportException(fileName, pathHelper.getPathName(), "Read index \"" + indexFromFile + "\" doesn't match expected \"" + expectedIndex + "\"");
      }

      return indexFromFile;
   }

   @Override
   public double readDouble(String expectedName) throws ImportException {
      pathHelper.enterSubTree(expectedName);
      nextResult();
      assertName(expectedName);
      final double val = getDouble();
      pathHelper.leaveSubTree();
      return val;
   }

   @Override
   public long readLong(String expectedName) throws ImportException {
      pathHelper.enterSubTree(expectedName);
      nextResult();
      assertName(expectedName);
      final long val = getLong();
      pathHelper.leaveSubTree();
      return val;
   }

   @Override
   public String readString(String expectedName) throws ImportException {
      pathHelper.enterSubTree(expectedName);
      nextResult();
      assertName(expectedName);
      final String val = getString();
      pathHelper.leaveSubTree();
      return val;
   }
}
