/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics;

import ikr.simlib.statistics.conditional.ConditionalUnweightedSampleCollector;
import ikr.simlib.statistics.conditional.ConditionalWeightedSampleCollector;

import java.util.Arrays;
import java.util.Collection;

public class Statistics {

   /**
    * wraps a {@link UnweightedSampleCollector} which accepts no values, so that it
    * accepts any values but ignores those
    */
   public static <Value> UnweightedSampleCollector<Value> ignoreValue(final UnweightedSampleCollector<Void> stat) {
      return new UnweightedSampleCollector<Value>() {
         @Override
         public void update(Value value) {
            stat.update(null);
         }
      };
   }

   /**
    * wraps an {@link UnweightedSampleCollector}, so that it accepts a weight but
    * ignores it
    */
   public static <Value, Weight> WeightedSampleCollector<Value, Weight> asWeightedStatistic(final UnweightedSampleCollector<Value> stat) {
      return new WeightedSampleCollector<Value, Weight>() {
         @Override
         public void update(Value value, Weight weight) {
            stat.update(value);
         }

         @Override
         public void update(Value value) {
            stat.update(value);
         }
      };
   }

   /**
    * wraps an {@link UnweightedSampleCollector}, so that it accepts a weight but
    * ignores it
    */
   public static <Key, Value, Weight> ConditionalWeightedSampleCollector<Key, Value, Weight> asWeightedStatistic(final ConditionalUnweightedSampleCollector<Key, Value> stat) {
      return new ConditionalWeightedSampleCollector<Key, Value, Weight>() {
         @Override
         public void update(Key key, Value value, Weight weight) {
            stat.update(key, value);
         }
      };
   }

   /**
    * wraps an {@link UnweightedSampleCollector} so that it implements the interface
    * {@link ConditionalUnweightedSampleCollector}. Each time the method
    * <code>update()</code> is called, the key is ignored and the method
    * <code>update()</code> of the {@link UnweightedSampleCollector} is called with
    * the value.
    */
   public static <Key, Value> ConditionalUnweightedSampleCollector<Key, Value> asConditionalStatistic(final UnweightedSampleCollector<Value> stat) {
      return new ConditionalUnweightedSampleCollector<Key, Value>() {
         @Override
         public void update(Key key, Value value) {
            stat.update(value);
         }
      };
   }

   /**
    * wraps a {@link WeightedSampleCollector} so that it implements the interface
    * {@link ConditionalWeightedSampleCollector}. Each time the method
    * <code>update()</code> is called, the key is ignored and the method
    * <code>update()</code> of the {@link WeightedSampleCollector} is called with the
    * value and the weight.
    */
   public static <Key, Value, Weight> ConditionalWeightedSampleCollector<Key, Value, Weight> asConditionalStatistic(final WeightedSampleCollector<Value, Weight> stat) {
      return new ConditionalWeightedSampleCollector<Key, Value, Weight>() {
         @Override
         public void update(Key key, Value value, Weight weight) {
            stat.update(value, weight);
         }
      };
   }

   /**
    * wraps a {@link WeightedSampleCollector} with weight of type long, so that it
    * implements the interface {@link UnweightedSampleCollector}. Each time the method
    * <code>update()</code> is called, the method <code>update()</code> of the
    * wrapped statistic is called with a weight set to one.
    */
   public static <Value> UnweightedSampleCollector<Value> weightOneLong(final WeightedSampleCollector<Value, Long> stat) {
      return map(stat, new Function<Value, Value>() {
         @Override
         public Value map(Value input) {
            return input;
         }
      }, new Function<Value, Long>() {
         @Override
         public Long map(Value input) {
            return 1l;
         }
      });
   }

   /**
    * wraps a {@link WeightedSampleCollector} with weight of type double, so that it
    * implements the interface {@link UnweightedSampleCollector}. Each time the method
    * <code>update()</code> is called, the method <code>update()</code> of the
    * wrapped statistic is called with a weight set to one.
    */
   public static <Value> UnweightedSampleCollector<Value> weightOneDouble(final WeightedSampleCollector<Value, Double> stat) {
      return map(stat, new Function<Value, Value>() {
         @Override
         public Value map(Value input) {
            return input;
         }
      }, new Function<Value, Double>() {
         @Override
         public Double map(Value input) {
            return 1d;
         }
      });
   }

   /**
    * wraps an {@link UnweightedSampleCollector}, so that it accepts a weight. The
    * weight is converted to multiple calls of the <code>update()</code> method
    * of the base statistic.
    */
   public static <Value> WeightedSampleCollector<Value, Long> weightToMultipleCalls(final UnweightedSampleCollector<Value> stat) {
      return new WeightedSampleCollector<Value, Long>() {
         @Override
         public void update(Value value, Long weight) {
            for (long i = 0; i < weight; i++)
               stat.update(value);
         }

         @Override
         public void update(Value value) {
            stat.update(value);
         }
      };
   }

   /**
    * combines multiple {@link UnweightedSampleCollector} instances. Each time the
    * method <code>update()</code> of the wrapping statistic is called, the
    * <code>update()</code> methods of each of the wrapped statistics is called
    * with the same value.
    */
   public static <Value> UnweightedSampleCollector<Value> combineUnweighted(final Collection<UnweightedSampleCollector<? super Value>> stats) {
      return new UnweightedSampleCollector<Value>() {
         @Override
         public void update(Value value) {
            for (UnweightedSampleCollector<? super Value> stat : stats)
               stat.update(value);
         }
      };
   }

   /**
    * combines multiple {@link UnweightedSampleCollector} instances. Each time the
    * method <code>update()</code> of the wrapping statistic is called, the
    * <code>update()</code> methods of each of the wrapped statistics is called
    * with the same value.
    */
   @SafeVarargs
   public static <Value> UnweightedSampleCollector<Value> combine(UnweightedSampleCollector<? super Value>... stats) {
      return combineUnweighted(Arrays.asList(stats));
   }

   /**
    * combines multiple {@link WeightedSampleCollector} instances. Each time the
    * method <code>update()</code> of the wrapping statistic is called, the
    * <code>update()</code> methods of each of the wrapped statistics is called
    * with the same value and weight.
    */
   public static <Value, Weight> WeightedSampleCollector<Value, Weight> combineWeighted(final Collection<WeightedSampleCollector<? super Value, ? super Weight>> stats) {
      return new WeightedSampleCollector<Value, Weight>() {
         @Override
         public void update(Value value, Weight weight) {
            for (WeightedSampleCollector<? super Value, ? super Weight> stat : stats)
               stat.update(value, weight);
         }

         @Override
         public void update(Value value) {
            for (WeightedSampleCollector<? super Value, ? super Weight> stat : stats)
               stat.update(value);
         }
      };
   }

   /**
    * combines multiple {@link WeightedSampleCollector} instances. Each time the
    * method <code>update()</code> of the wrapping statistic is called, the
    * <code>update()</code> methods of each of the wrapped statistics is called
    * with the same value and weight.
    */
   @SafeVarargs
   public static <Value, Weight> WeightedSampleCollector<Value, Weight> combine(WeightedSampleCollector<? super Value, ? super Weight>... stats) {
      return combineWeighted(Arrays.asList(stats));
   }

   /**
    * combines multiple {@link ConditionalUnweightedSampleCollector} instances. Each
    * time the method <code>update()</code> of the wrapping statistic is called,
    * the <code>update()</code> methods of each of the wrapped statistics is
    * called with the same key and value.
    */
   public static <Key, Value> ConditionalUnweightedSampleCollector<Key, Value> combineCondUnweighted(final Collection<ConditionalUnweightedSampleCollector<? super Key, ? super Value>> stats) {
      return new ConditionalUnweightedSampleCollector<Key, Value>() {
         @Override
         public void update(Key key, Value value) {
            for (ConditionalUnweightedSampleCollector<? super Key, ? super Value> stat : stats)
               stat.update(key, value);
         }
      };
   }

   /**
    * combines multiple {@link ConditionalUnweightedSampleCollector} instances. Each
    * time the method <code>update()</code> of the wrapping statistic is called,
    * the <code>update()</code> methods of each of the wrapped statistics is
    * called with the same key and value.
    */
   @SafeVarargs
   public static <Key, Value> ConditionalUnweightedSampleCollector<Key, Value> combine(ConditionalUnweightedSampleCollector<? super Key, ? super Value>... stats) {
      return combineCondUnweighted(Arrays.asList(stats));
   }

   /**
    * combines multiple {@link ConditionalWeightedSampleCollector} instances. Each
    * time the method <code>update()</code> of the wrapping statistic is called,
    * the <code>update()</code> methods of each of the wrapped statistics is
    * called with the same key, value, and weight.
    */
   public static <Key, Value, Weight> ConditionalWeightedSampleCollector<Key, Value, Weight> combineCondWeighted(final Collection<ConditionalWeightedSampleCollector<? super Key, ? super Value, ? super Weight>> stats) {
      return new ConditionalWeightedSampleCollector<Key, Value, Weight>() {
         @Override
         public void update(Key key, Value value, Weight weight) {
            for (ConditionalWeightedSampleCollector<? super Key, ? super Value, ? super Weight> stat : stats)
               stat.update(key, value, weight);
         }
      };
   }

   /**
    * combines multiple {@link ConditionalWeightedSampleCollector} instances. Each
    * time the method <code>update()</code> of the wrapping statistic is called,
    * the <code>update()</code> methods of each of the wrapped statistics is
    * called with the same key, value, and weight.
    */
   @SafeVarargs
   public static <Key, Value, Weight> ConditionalWeightedSampleCollector<Key, Value, Weight> combine(ConditionalWeightedSampleCollector<? super Key, ? super Value, ? super Weight>... stats) {
      return combineCondWeighted(Arrays.asList(stats));
   }

   /**
    * generic function with a single input and a single output parameter
    *
    * @param <Input>
    *           type of the input parameter
    * @param <Output>
    *           type of the output parameter
    */
   public interface Function<Input, Output> {
      Output map(Input input);
   }

   /**
    * Wraps an {@link UnweightedSampleCollector} which accepts values of type
    * <code>O</code>, so that the result accepts values of type <code>I</code>.
    * The specified mapping function has the input type <code>I</code> and the
    * output type <code>O</code>. It is used to convert each value.
    */
   public static <Input, Output> UnweightedSampleCollector<Input> map(final UnweightedSampleCollector<Output> stat, final Function<Input, Output> mapping) {
      return new UnweightedSampleCollector<Input>() {
         @Override
         public void update(Input value) {
            stat.update(mapping.map(value));
         }
      };
   }

   /**
    * Wraps a {@link WeightedSampleCollector} which accepts values of type
    * <code>V</code> and weights of type <code>W</code>, so that the result is
    * an unweighted statistic which accepts values of type <code>I</code>. The
    * first mapping function has the input type <code>I</code> and the output
    * type <code>V</code>. It is used to convert each value. The second mapping
    * function (types <code>I</code>, <code>W</code>) is used to map the input
    * to the weight.
    */
   public static <Input, Value, Weight> UnweightedSampleCollector<Input> map(final WeightedSampleCollector<Value, Weight> stat, final Function<Input, Value> valueMapping, final Function<Input, Weight> weightMapping) {
      return new UnweightedSampleCollector<Input>() {
         @Override
         public void update(Input value) {
            stat.update(valueMapping.map(value), weightMapping.map(value));
         }
      };
   }

   /**
    * Wraps a {@link WeightedSampleCollector} which accepts values of type
    * <code>O</code>, so that the result accepts values of type <code>I</code>.
    * The specified mapping function has the input type <code>I</code> and the
    * output type <code>O</code>. It is used to convert each value. The weight
    * is not mapped.
    */
   public static <Input, Output, Weight> WeightedSampleCollector<Input, Weight> map(final WeightedSampleCollector<Output, Weight> stat, final Function<Input, Output> mapping) {
      return new WeightedSampleCollector<Input, Weight>() {
         @Override
         public void update(Input value, Weight weight) {
            stat.update(mapping.map(value), weight);
         }

         @Override
         public void update(Input value) {
            stat.update(mapping.map(value));
         }
      };
   }

   /**
    * Wraps a {@link WeightedSampleCollector} which accepts weights of type
    * <code>O</code>, so that the result accepts weights of type <code>I</code>.
    * The specified mapping function has the input type <code>I</code> and the
    * output type <code>O</code>. It is used to convert each weight. The value
    * is not mapped.
    * <p>
    * It is not allowed to use the update() method without weight any more. In
    * case it is used, a runtime exception is thrown.
    */
   public static <Input, Output, Value> WeightedSampleCollector<Value, Input> mapWeight(final WeightedSampleCollector<Value, Output> stat, final Function<Input, Output> mapping) {
      return new WeightedSampleCollector<Value, Input>() {
         @Override
         public void update(Value value, Input weight) {
            stat.update(value, mapping.map(weight));
         }

         @Override
         public void update(Value value) {
            throw new UnsupportedOperationException();
         }
      };
   }

   /**
    * Wraps a {@link ConditionalUnweightedSampleCollector} which accepts values of
    * type <code>O</code>, so that the result accepts values of type
    * <code>I</code>. The specified mapping function has the input type
    * <code>I</code> and the output type <code>O</code>. It is used to convert
    * each value. The key is not mapped.
    */
   public static <Key, Input, Output> ConditionalUnweightedSampleCollector<Key, Input> map(final ConditionalUnweightedSampleCollector<Key, Output> stat, final Function<Input, Output> mapping) {
      return new ConditionalUnweightedSampleCollector<Key, Input>() {
         @Override
         public void update(Key key, Input value) {
            stat.update(key, mapping.map(value));
         }
      };
   }

   /**
    * Wraps a {@link ConditionalUnweightedSampleCollector} which accepts keys of type
    * <code>O</code>, so that the result accepts keys of type <code>I</code>.
    * The specified mapping function has the input type <code>I</code> and the
    * output type <code>O</code>. It is used to convert each key. The value is
    * not mapped.
    */
   public static <Input, Value, Output> ConditionalUnweightedSampleCollector<Input, Value> mapKey(final ConditionalUnweightedSampleCollector<Output, Value> stat, final Function<Input, Output> mapping) {
      return new ConditionalUnweightedSampleCollector<Input, Value>() {
         @Override
         public void update(Input key, Value value) {
            stat.update(mapping.map(key), value);
         }
      };
   }

   /**
    * Wraps a {@link ConditionalWeightedSampleCollector} which accepts values of type
    * <code>O</code>, so that the result accepts values of type <code>I</code>.
    * The specified mapping function has the input type <code>I</code> and the
    * output type <code>O</code>. It is used to convert each value. The key and
    * the weight are not mapped.
    */
   public static <Key, Input, Weight, Output> ConditionalWeightedSampleCollector<Key, Input, Weight> map(final ConditionalWeightedSampleCollector<Key, Output, Weight> stat, final Function<Input, Output> mapping) {
      return new ConditionalWeightedSampleCollector<Key, Input, Weight>() {
         @Override
         public void update(Key key, Input value, Weight weight) {
            stat.update(key, mapping.map(value), weight);
         }
      };
   }

   /**
    * Wraps a {@link ConditionalWeightedSampleCollector} which accepts keys of type
    * <code>O</code>, so that the result accepts keys of type <code>I</code>.
    * The specified mapping function has the input type <code>I</code> and the
    * output type <code>O</code>. It is used to convert each key. The value and
    * the weight are not mapped.
    */
   public static <Input, Value, Weight, Output> ConditionalWeightedSampleCollector<Input, Value, Weight> mapKey(final ConditionalWeightedSampleCollector<Output, Value, Weight> stat, final Function<Input, Output> mapping) {
      return new ConditionalWeightedSampleCollector<Input, Value, Weight>() {
         @Override
         public void update(Input key, Value value, Weight weight) {
            stat.update(mapping.map(key), value, weight);
         }
      };
   }

   /**
    * Wraps a {@link ConditionalWeightedSampleCollector} which accepts weights of type
    * <code>O</code>, so that the result accepts weights of type <code>I</code>.
    * The specified mapping function has the input type <code>I</code> and the
    * output type <code>O</code>. It is used to convert each weight. The key and
    * the value are not mapped.
    */
   public static <Key, Value, Input, Output> ConditionalWeightedSampleCollector<Key, Value, Input> mapWeight(final ConditionalWeightedSampleCollector<Key, Value, Output> stat, final Function<Input, Output> mapping) {
      return new ConditionalWeightedSampleCollector<Key, Value, Input>() {
         @Override
         public void update(Key key, Value value, Input weight) {
            stat.update(key, value, mapping.map(weight));
         }
      };
   }

   /**
    * Wraps a {@link ConditionalWeightedSampleCollector}, so that the result is an
    * {@link UnweightedSampleCollector}. Key, value, and weight of the wrapped
    * statistics are generated from the single input value of the
    * {@link UnweightedSampleCollector} by three different functions.
    */
   public static <Input, Key, Value, Weight> UnweightedSampleCollector<Input> map(final ConditionalWeightedSampleCollector<Key, Value, Weight> stat, final Function<Input, Key> keyMapping, final Function<Input, Value> valueMapping, final Function<Input, Weight> weightMapping) {
      return new UnweightedSampleCollector<Input>() {
         @Override
         public void update(Input input) {
            stat.update(keyMapping.map(input), valueMapping.map(input), weightMapping.map(input));
         }
      };
   }

   /**
    * generic function that maps a single value to a boolean
    * 
    * @param <Input>
    *           type of the input value
    */
   public interface Predicate<Input> {
      boolean test(Input input);
   }

   /**
    * Wraps an {@link UnweightedSampleCollector}, so that calls to the update() method
    * are only executed when the filterFunction returns true. The output of the
    * filterFunction is a function of the value.
    */
   public static <Value> UnweightedSampleCollector<Value> filter(final UnweightedSampleCollector<Value> stat, final Predicate<Value> filterFunction) {
      return new UnweightedSampleCollector<Value>() {
         @Override
         public void update(Value value) {
            if (filterFunction.test(value))
               stat.update(value);
         }
      };
   }

   /**
    * Wraps a {@link WeightedSampleCollector}, so that calls to the update() method
    * are only executed when the filterFunction returns true. The output of the
    * filterFunction is a function of the value.
    */
   public static <Value, Weight> WeightedSampleCollector<Value, Weight> filter(final WeightedSampleCollector<Value, Weight> stat, final Predicate<Value> filterFunction) {
      return new WeightedSampleCollector<Value, Weight>() {
         @Override
         public void update(Value value, Weight weight) {
            if (filterFunction.test(value))
               stat.update(value, weight);
         }

         @Override
         public void update(Value value) {
            if (filterFunction.test(value))
               stat.update(value);
         }
      };
   }

   /**
    * Wraps a {@link WeightedSampleCollector}, so that calls to the update() method
    * are only executed when the filterFunction returns true. The output of the
    * filterFunction is a function of the weight.
    * <p>
    * It is not allowed to use the update() method without weight any more. In
    * case it is used, a runtime exception is thrown.
    */
   public static <Value, Weight> WeightedSampleCollector<Value, Weight> filterByWeight(final WeightedSampleCollector<Value, Weight> stat, final Predicate<Weight> filterFunction) {
      return new WeightedSampleCollector<Value, Weight>() {
         @Override
         public void update(Value value, Weight weight) {
            if (filterFunction.test(weight))
               stat.update(value, weight);
         }

         @Override
         public void update(Value value) {
            throw new UnsupportedOperationException();
         }
      };
   }

   /**
    * Wraps a {@link ConditionalUnweightedSampleCollector}, so that calls to the
    * update() method are only executed when the filterFunction returns true.
    * The output of the filterFunction is a function of the value.
    */
   public static <Key, Value> ConditionalUnweightedSampleCollector<Key, Value> filter(final ConditionalUnweightedSampleCollector<Key, Value> stat, final Predicate<Value> filterFunction) {
      return new ConditionalUnweightedSampleCollector<Key, Value>() {
         @Override
         public void update(Key key, Value value) {
            if (filterFunction.test(value))
               stat.update(key, value);
         }
      };
   }

   /**
    * Wraps a {@link ConditionalUnweightedSampleCollector}, so that calls to the
    * update() method are only executed when the filterFunction returns true.
    * The output of the filterFunction is a function of the key.
    */
   public static <Key, Value> ConditionalUnweightedSampleCollector<Key, Value> filterByKey(final ConditionalUnweightedSampleCollector<Key, Value> stat, final Predicate<Key> filterFunction) {
      return new ConditionalUnweightedSampleCollector<Key, Value>() {
         @Override
         public void update(Key key, Value value) {
            if (filterFunction.test(key))
               stat.update(key, value);
         }
      };
   }

   /**
    * Wraps a {@link ConditionalWeightedSampleCollector}, so that calls to the
    * update() method are only executed when the filterFunction returns true.
    * The output of the filterFunction is a function of the value.
    */
   public static <Key, Value, Weight> ConditionalWeightedSampleCollector<Key, Value, Weight> filter(final ConditionalWeightedSampleCollector<Key, Value, Weight> stat, final Predicate<Value> filterFunction) {
      return new ConditionalWeightedSampleCollector<Key, Value, Weight>() {
         @Override
         public void update(Key key, Value value, Weight weight) {
            if (filterFunction.test(value))
               stat.update(key, value, weight);
         }
      };
   }

   /**
    * Wraps a {@link ConditionalWeightedSampleCollector}, so that calls to the
    * update() method are only executed when the filterFunction returns true.
    * The output of the filterFunction is a function of the weight.
    */
   public static <Key, Value, Weight> ConditionalWeightedSampleCollector<Key, Value, Weight> filterByWeight(final ConditionalWeightedSampleCollector<Key, Value, Weight> stat, final Predicate<Weight> filterFunction) {
      return new ConditionalWeightedSampleCollector<Key, Value, Weight>() {
         @Override
         public void update(Key key, Value value, Weight weight) {
            if (filterFunction.test(weight))
               stat.update(key, value, weight);
         }
      };
   }

   /**
    * Wraps a {@link ConditionalWeightedSampleCollector}, so that calls to the
    * update() method are only executed when the filterFunction returns true.
    * The output of the filterFunction is a function of the key.
    */
   public static <Key, Value, Weight> ConditionalWeightedSampleCollector<Key, Value, Weight> filterByKey(final ConditionalWeightedSampleCollector<Key, Value, Weight> stat, final Predicate<Key> filterFunction) {
      return new ConditionalWeightedSampleCollector<Key, Value, Weight>() {
         @Override
         public void update(Key key, Value value, Weight weight) {
            if (filterFunction.test(key))
               stat.update(key, value, weight);
         }
      };
   }
}
