/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics;

import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.confidenceinterval.ConfidenceIntervalEstimatorManager;

import java.io.Serializable;

/**
 * A summation of samples.
 * 
 */
public final class Summation implements Serializable, Cloneable {

   private static final long serialVersionUID = -8170204854173921074L;

   // -- Result file export --

   private static final String SUMMATION_TYPE = "Summation";
   private static final String EXPORT_SUM = ResultKeyword.asExport(ResultKeywords.SUM.toString());
   private static final String EXPORT_NUMBER_OF_SAMPLES = ResultKeyword.asExport(ResultKeywords.NUMBER_OF_SAMPLES.toString());
   private static final String EXPORT_SUM_OF_SQUARES = ResultKeyword.asExport("sumofsquares");

   /**
    * Sum of the samples.
    */
   private double sum;

   /**
    * Sum of the squared samples.
    */
   private double sumOfSquares;

   /**
    * Number of samples.
    */
   private long numberOfSamples;

   /**
    * Constructs a new summation.
    */
   public Summation() {
   }

   @Override
   public boolean equals(Object obj) {
      if (obj == null) {
         return false;
      }

      if (obj.getClass() != getClass()) {
         return false;
      }

      Summation summation = (Summation) obj;

      return summation.sum == sum && summation.sumOfSquares == sumOfSquares && summation.numberOfSamples == numberOfSamples;
   }

   @Override
   public int hashCode() {
      return (int) (numberOfSamples ^ (numberOfSamples >>> 32) ^ Double.valueOf(sum).hashCode());
   }

   @Override
   public String toString() {
      StringBuilder str = new StringBuilder();

      str.append("[sum=");
      str.append(getSum());

      str.append(", mean=");
      double mean = getMean();
      str.append(mean);

      str.append(", variance=");
      double var = getVariance();
      str.append(var);

      str.append(", deviation=");
      double dev = getStandardDeviation();
      str.append(dev);

      str.append(']');

      return str.toString();
   }

   /**
    * Exports a summation as a result node.
    */
   public void exportSum(String name, ResultWriter writer) throws PrintException {

      writer.writeStartNode(SUMMATION_TYPE, name);

      writer.writeDouble(EXPORT_SUM, sum);
      writer.writeDouble(EXPORT_SUM_OF_SQUARES, sumOfSquares);
      writer.writeLong(EXPORT_NUMBER_OF_SAMPLES, numberOfSamples);

      writer.writeEndNode();
   }

   /**
    * Imports a summation from a result node.
    */
   public void importSum(String name, ResultReader reader) throws ImportException {
      reader.readStartNode(SUMMATION_TYPE, name);

      sum = reader.readDouble(EXPORT_SUM);
      sumOfSquares = reader.readDouble(EXPORT_SUM_OF_SQUARES);
      numberOfSamples = reader.readLong(EXPORT_NUMBER_OF_SAMPLES);

      reader.readEndNode();
   }

   /**
    * Adds a sample to this sum.
    * 
    * @param sample
    *           value to be added
    */
   public void update(double sample) {
      sum += sample;
      sumOfSquares += (sample * sample);
      numberOfSamples += 1;
   }

   /**
    * Resets this summation.
    */
   public void reset() {
      sum = 0;
      sumOfSquares = 0;
      numberOfSamples = 0;
   }

   /**
    * Returns the sum of all samples.
    * 
    * @return sum of all samples
    */
   public double getSum() {
      if (numberOfSamples > 0) {
         return sum;
      }

      return Double.NaN;
   }

   /**
    * Returns the mean of all samples added to this sum.
    * 
    * @return mean of the added samples or NaN
    */
   public double getMean() {
      double value;
      if (numberOfSamples > 0) {
         value = sum / numberOfSamples;
      } else {
         value = Double.NaN;
      }

      return value;
   }

   /**
    * Returns the variation of this sum.
    * 
    * @return variation of this sum or NaN
    */
   public double getVariance() {
      double value;
      if (numberOfSamples > 1) {
         value = (sumOfSquares - ((sum * sum) / numberOfSamples)) / (numberOfSamples - 1);

         if (value < 0) {
            value = 0;// get rid of possible rounding errors
         }
      } else {
         value = Double.NaN;
      }

      return value;
   }

   /**
    * Returns the coefficient of variation of this sum.
    * 
    * @return coefficient of variation of this sum or NaN
    */
   public double getCoefficientOfVariation() {
      double value;
      double mean;
      double standardDeviation;

      if (numberOfSamples > 1) {
         mean = getMean();
         standardDeviation = getStandardDeviation();

         if (mean != 0) {
            value = Math.abs(standardDeviation / mean);
         } else {
            if (standardDeviation == 0) {
               value = 0;
            } else {
               value = Double.NaN;
            }
         }
      } else {
         value = Double.NaN;
      }

      return value;
   }

   /**
    * Returns the standard deviation of this sum.
    * 
    * @return standard deviation of this sum or NaN
    */
   public double getStandardDeviation() {
      double value;
      if (numberOfSamples > 1) {
         value = Math.sqrt(getVariance());
      } else {
         value = Double.NaN;
      }

      return value;
   }

   /**
    * Returns the number of samples added to this sum.
    * 
    * @return number of samples added
    */
   public long getNumberOfSamples() {
      return numberOfSamples;
   }

   /**
    * Returns the confidence interval of this sum.
    * 
    * @return confidence interval of this sum
    */
   public double getConfidenceInterval() {
      return ConfidenceIntervalEstimatorManager.getInstance().getConfidenceInterval(numberOfSamples, getStandardDeviation());
   }
}
