/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.aggregating;

import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * {@link CountAggregatingStatistic} aggregates a fixed number n of values with
 * the help of an {@link Aggregator}. Calls of the method
 * {@link #update(Object)} are counted. Upon each n'th call of
 * {@link #update(Object)}, the aggregated value is written to the backing
 * statistic and the aggregator is reset.
 * <p>
 * Implements the interface {@link UnweightedSampleCollector}, so that it can be used
 * as a replacement for a normal statistic in any meter.
 * <p>
 * The type of input and output can be different, as aggregation may convert the
 * type. Example: the average of long values can be returned as double.
 * 
 * @param <T>
 *           type of the input values
 * @param <V>
 *           type of the output values
 */
public class CountAggregatingStatistic<T, V> implements UnweightedSampleCollector<T> {
   private final UnweightedSampleCollector<V> baseStat;
   private final int numberOfValuesToAggregate;
   private final Aggregator<? super T, ? extends V> aggregator;
   private int count;

   public CountAggregatingStatistic(UnweightedSampleCollector<V> baseStat, int numberOfValuesToAggregate, Aggregator<? super T, ? extends V> aggregator) {
      this.baseStat = baseStat;
      this.numberOfValuesToAggregate = numberOfValuesToAggregate;
      this.aggregator = aggregator;
   }

   @Override
   public void update(T value) {
      aggregator.putValue(value);
      count++;
      if (count == numberOfValuesToAggregate)
         putAggregatedValueToStatistic();
   }

   private void putAggregatedValueToStatistic() {
      final V aggregate = aggregator.getAndReset();
      baseStat.update(aggregate);
      count = 0;
   }
}
