/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.conditional;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultTaggable;
import ikr.simlib.statistics.support.BucketUtility;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

/**
 * Base class for conditional statistics. The conditioner / key can be any type
 * and is specified by the generic parameter <code>K</code>. Also the type of
 * the statistic is generic and specified by the generic type <code>S</code>.
 * <p>
 * Three constructors are provided for different use cases.
 * <ul>
 * <li>
 * {@link #ConditionalSampleCollectorBase(SimNode, Class, StatisticCreator, Iterable, StatisticTagger)}
 * is intended to be used with generic discrete keys. All possible keys have to
 * be provided at initialization time, so that the statistics can be created
 * before the simulation starts. Make sure that the {@link Iterable} always
 * returns the keys in the same order (i.e. don't iterate over a {@link HashSet}
 * ), otherwise the order of the statistics is undefined and import simulations
 * will fail.</li>
 * <li>
 * {@link #ConditionalSampleCollectorBase(SimNode, Class, StatisticCreator, BucketUtility)}
 * is intended to be used with continuous double keys, which are mapped to
 * discrete buckets by a {@link BucketUtility}.</li>
 * <li> {@link #ConditionalSampleCollectorBase(SimNode, Class, StatisticCreator, int)}
 * is a convenience constructor to be used with integer keys. Keys start at
 * zero. For other use cases with integer keys use the generic constructor.</li>
 * </ul>
 * <p>
 * All constructors require the class of the key as parameter. This is used for
 * type-checking at runtime, which is not possible with java generics alone.
 * <p>
 * In addition, the constructors require an instance of a
 * {@link StatisticCreator}. This is a function which creates new instances of
 * the intended statistic class.
 * 
 * @param <K>
 *           type of the key, a.k.a conditioner
 * @param <S>
 *           type of the statistic
 */
public abstract class ConditionalSampleCollectorBase<K, S extends ResultTaggable> {

   public interface ToIndexMapper<K> {
      int mapToIndex(K key);
   }

   public interface StatisticCreator<S extends ResultTaggable> {
      S createNewStatistic(SimNode ownNode);
   }

   public interface StatisticTagger<K, S extends ResultTaggable> {
      void tagStatistic(S stat, K key);
   }

   public static StatisticTagger<Object, ResultTaggable> TO_STRING_TAGGER = new StatisticTagger<Object, ResultTaggable>() {
      @Override
      public void tagStatistic(ResultTaggable stat, Object key) {
         stat.addResultTag("key", key.toString());
      }
   };

   private final List<S> statistics;
   private final ToIndexMapper<K> mapper;

   private final SimNode ownNode;
   private final Class<K> keyClass;
   private final StatisticCreator<? extends S> statCreator;
   private final StatisticTagger<? super K, ? super S> tagger;

   private final Map<K, Integer> keyToIndex;

   /** for enumerable keys */
   protected ConditionalSampleCollectorBase(SimNode ownNode, Class<K> keyClass, StatisticCreator<? extends S> statCreator, Iterable<? extends K> keys, StatisticTagger<? super K, ? super S> tagger) {
      this.ownNode = ownNode;
      this.keyClass = keyClass;
      this.statCreator = statCreator;
      this.tagger = tagger;

      statistics = new ArrayList<S>();
      keyToIndex = new HashMap<K, Integer>();
      for (K key : keys) {
         final int idx = statistics.size();
         S newStat = statCreator.createNewStatistic(ownNode.createChildNode("BucketStat_" + idx));
         statistics.add(newStat);
         assert statistics.get(idx) == newStat;
         tagger.tagStatistic(newStat, key);
         keyToIndex.put(key, idx);
      }

      mapper = new ToIndexMapper<K>() {
         @Override
         public int mapToIndex(K key) {
            final Integer index = keyToIndex.get(key);
            if (index == null)
               throw new SimulationException("key " + key.toString() + " is unknown - check initialization");
            return index;
         }
      };
   }

   @SuppressWarnings("unchecked")
   private static <K> ToIndexMapper<K> castToGeneric(ToIndexMapper<Double> input, Class<K> keyClass) {
      if (!keyClass.equals(Double.class))
         throw new ClassCastException();
      return (ToIndexMapper<K>) input;
   }

   /** for doubles (lin or log spaced mapping) */
   protected ConditionalSampleCollectorBase(SimNode ownNode, Class<K> keyClass, StatisticCreator<? extends S> statCreator, final BucketUtility bucketUtility) {
      this.ownNode = ownNode;
      this.keyClass = keyClass;
      this.statCreator = statCreator;
      this.tagger = null;
      this.keyToIndex = null;

      if (!keyClass.equals(Double.class))
         throw new SimulationException("this configuration of the conditional statistic supports only Double keys");

      statistics = new ArrayList<S>(bucketUtility.getArraySize());
      for (int i = 0; i < bucketUtility.getArraySize(); i++) {
         S newStat = statCreator.createNewStatistic(ownNode.createChildNode("BucketStat_" + i));
         statistics.add(newStat);
         newStat.addResultTag(ResultKeywords.LBOUND, Double.toString(bucketUtility.getLowerBound(i)));
         newStat.addResultTag(ResultKeywords.UBOUND, Double.toString(bucketUtility.getUpperBound(i)));
      }

      mapper = castToGeneric(bucketUtility, keyClass);
   }

   private static <K> Iterable<K> integerRange(int size, Class<K> keyClass) {
      if (!keyClass.equals(Integer.class))
         throw new SimulationException("this configuration of the conditional statistic supports only Integer keys");

      final ArrayList<K> retval = new ArrayList<K>(size);
      for (int i = 0; i < size; i++)
         retval.add(keyClass.cast(Integer.valueOf(i)));
      return retval;
   }

   /** for integers (0..numBuckets-1) */
   protected ConditionalSampleCollectorBase(SimNode ownNode, Class<K> keyClass, StatisticCreator<? extends S> statCreator, int numBuckets) {
      this(ownNode, keyClass, statCreator, integerRange(numBuckets, keyClass), TO_STRING_TAGGER);
   }

   protected S getStatistic(K key) {
      if (!keyClass.isAssignableFrom(key.getClass()))
         throw new ClassCastException("this instance of the conditional statistic supports only keys of type " + keyClass.toString());
      return statistics.get(mapper.mapToIndex(key));
   }

   /**
    * register an additional key. A new statistic is created, tagged, and added
    * to the internal lookup map.
    * <p>
    * <strong>Attention</strong> if keys are added at simulation runtime, it is
    * not possible to import batch results, because the structure of the export
    * file does not match the objects instantiated at import time.
    * 
    * @param key
    *           the new key
    */
   public void registerKey(K key) {
      if (keyToIndex == null)
         throw new RuntimeException("the method registerKey() can only be called in case the generic constructor has been used to instanciate this object");
      final int idx = statistics.size();
      S newStat = statCreator.createNewStatistic(ownNode.createChildNode("BucketStat_" + idx));
      statistics.add(newStat);
      assert statistics.get(idx) == newStat;
      tagger.tagStatistic(newStat, key);
      keyToIndex.put(key, idx);
   }

}
