/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.distribution.trace.data;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * {@link DataSource} backed by a file.
 */
public class TraceDataSource implements DataSource {
   private final File file;

   public TraceDataSource(File file) {
      this.file = file;
   }

   public File getFile() {
      return file;
   }

   /**
    * reopen the file and start from the first entry.
    */
   @Override
   public Iterator<Sample> iterator() {
      try {
         return new TraceReader(file).getIterator();
      } catch (TraceFilesInvalidException e) {
         throw new RuntimeException(e);
      }
   }

   /**
    * Merge multiple sorted files to a temporary sorted file. The file is
    * deleted when the program exits.
    * 
    * @return a {@link DataSource} backed by the merged file.
    */
   @Override
   public TraceDataSource combineSorted(List<DataSource> input) {
      final List<File> files = new ArrayList<File>();
      for (DataSource source : input)
         files.add(((TraceDataSource) source).file);
      final File dir = files.get(0).getAbsoluteFile().getParentFile();
      final File tmpFile;
      try {
         tmpFile = File.createTempFile("tempTrace", ".dat", dir);
      } catch (IOException e) {
         throw new RuntimeException("couldn't create tmp file in " + dir, e);
      }
      tmpFile.deleteOnExit();
      try {
         TraceSorter.mergeSort(files, tmpFile);
      } catch (TraceFilesInvalidException e) {
         throw new RuntimeException(e);
      }
      return new TraceDataSource(tmpFile);
   }

}
