/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.distribution.trace.data;

import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;

public class TraceSorter {
   private static final Logger logger = Logger.getLogger(TraceSorter.class.getName());
   private static final int MAX_SORT_IN_MEM = 50000000 / Sample.SIZE_BYTES; // 50MB

   /**
    * sorts file in place
    * 
    * @param file
    *           the data file
    * @throws TraceFilesInvalidException
    *            if the file cannot be read
    */
   public static void sort(File file) throws TraceFilesInvalidException {
      final File tmpFile = new File(file.getAbsolutePath() + ".sorted");
      if (tmpFile.isFile()) {
         System.err.println("WARNING: tmp file " + tmpFile.getAbsolutePath() + " exists - overwriting!");
         tmpFile.delete();
      }
      sort(file, tmpFile);
      file.delete();
      tmpFile.renameTo(file);
   }

   /**
    * sorts file src and writes sorted values to dst
    * 
    * @param src
    *           source file
    * @param dst
    *           destination file
    * @throws TraceFilesInvalidException
    *            if the file cannot be read
    */
   public static void sort(File src, File dst) throws TraceFilesInvalidException {
      final long startMillis = System.currentTimeMillis();
      final TraceReader reader = new TraceReader(src);
      final Iterator<Sample> iter = reader.getIterator();
      final ArrayList<File> segments = new ArrayList<File>();
      long totalSamples = 0;
      while (iter.hasNext()) {
         final File segmentDst = new File(dst.getAbsolutePath() + ".segment" + segments.size());
         segmentDst.deleteOnExit();
         totalSamples += readSortAndWriteSegment(iter, segmentDst);
         segments.add(segmentDst);
      }
      if (segments.size() == 1) {
         segments.get(0).renameTo(dst);
      } else {
         // overwrite, as samples can have been merged
         totalSamples = mergeSort(segments, dst);
         for (File f : segments)
            f.delete();
      }
      reader.close();
      final long stopMillis = System.currentTimeMillis();
      logger.fine("sorting trace file " + src.getName() + " with " + totalSamples + " samples took " + (stopMillis - startMillis) + "ms");
   }

   /**
    * Merges all files given in the map and writes the result to dst. Assumes
    * that the segment files are <strong>not sorted</strong>.
    * 
    * @param segments
    *           map with files and the number of samples in each file
    * @param dst
    *           destination file
    * @return the total number of samples
    * @throws TraceFilesInvalidException
    *            if the file cannot be read
    */
   public static long sortAndMergeSort(List<File> segments, File dst) throws TraceFilesInvalidException {
      for (File file : segments)
         TraceSorter.sort(file);
      return mergeSort(segments, dst);
   }

   /**
    * Merges all files given in the map and writes the result to dst. Assumes
    * that the segment files are <strong>sorted</strong>.
    * 
    * @param segments
    *           map with files and the number of samples in each file
    * @param dst
    *           destination file
    * @return the total number of samples
    * @throws TraceFilesInvalidException
    *            if the file cannot be read
    */
   public static long mergeSort(List<File> segments, File dst) throws TraceFilesInvalidException {
      final long startMillis = System.currentTimeMillis();
      // open files
      final TraceReader[] readers = new TraceReader[segments.size()];
      int i = 0;
      for (File seg : segments)
         readers[i++] = new TraceReader(seg);
      final TraceWriter output = new TraceWriter(dst, true);
      // sort
      Sample[] nextValues = new Sample[readers.length];
      boolean[] available = new boolean[readers.length];
      for (i = 0; i < readers.length; i++) {
         try {
            nextValues[i] = readers[i].readNext();
            available[i] = true;
         } catch (EOFException e) {
            available[i] = false;
         }
      }
      while (true) {
         int minIndex = -1;
         Sample minValue = null;
         boolean anyAvailable = false;
         for (i = 0; i < readers.length; i++) {
            anyAvailable |= available[i];
            // use Double.compare() to avoid problems with NaN etc.
            if (available[i] && (minValue == null || Double.compare(nextValues[i].value, minValue.value) <= 0)) {
               minValue = nextValues[i];
               minIndex = i;
            }
         }
         if (anyAvailable) {
            assert minIndex >= 0;
            assert minValue != null;
            output.write(minValue);
            try {
               nextValues[minIndex] = readers[minIndex].readNext();
            } catch (EOFException e) {
               available[minIndex] = false;
            }
         } else {
            break;
         }
      }

      // cleanup
      for (TraceReader r : readers)
         r.close();
      output.close();

      final long stopMillis = System.currentTimeMillis();
      logger.fine("merging " + segments.size() + " trace files to " + dst.getName() + " with " + output.getSamplesInFile() + " samples took " + (stopMillis - startMillis) + "ms");

      return output.getSamplesInFile();
   }

   private static long readSortAndWriteSegment(Iterator<Sample> iter, File dst) {
      // use an array because Collections.sort() internally copies the list to
      // an array, and memory could be expensive here
      final Sample[] data = new Sample[MAX_SORT_IN_MEM];
      int idx = 0;
      while (iter.hasNext() && idx < MAX_SORT_IN_MEM)
         data[idx++] = iter.next();
      final int numElem = idx;
      Arrays.sort(data, 0, numElem);
      TraceWriter writer = new TraceWriter(dst, true);
      for (int i = 0; i < numElem; i++)
         writer.write(data[i]);
      writer.close();
      return writer.getSamplesInFile();
   }

   /** for testing only */
   @SuppressWarnings("restriction")
   public static void main(String[] args) throws TraceFilesInvalidException, IOException {
      File src = File.createTempFile("testtrace", ".dat");
      src.deleteOnExit();
      TraceWriter w = new TraceWriter(src, true);
      final java.util.Random r = new java.util.Random();
      for (int i = 0; i < 100; i++)
         w.write(r.nextInt(10), 1);
      w.close();
      sort(src);
      TraceDataSource re = new TraceDataSource(src);
      int line = 0;
      for (Sample val : re)
         System.out.println("" + (line++) + " " + val);
   }

}
