/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.distribution.trace.dist;

import ikr.simlib.statistics.distribution.trace.data.DataSource;
import ikr.simlib.statistics.distribution.trace.data.Sample;
import ikr.simlib.statistics.distribution.trace.quantile.QuantileDescription;

import java.util.Arrays;
import java.util.Iterator;

/**************************************************************************
 * quantile calculation
 ***************************************************************************/

abstract class DistributionData {
   protected final QuantileDescription quantiles;
   protected final double[] quantileValues;
   protected double smallestValue;

   public DistributionData(QuantileDescription quantiles) {
      this.quantiles = quantiles;
      quantileValues = new double[quantiles.getNumberOfQuantiles()];
   }

   public int getNumberOfQuantiles() {
      return quantileValues.length;
   }

   /**
    * assumes a sorted data source
    */
   protected void calculateQuantiles(DataSource srcFile) {
      final double sumOfWeights = sumOfWeights(srcFile);

      final Iterator<Sample> iter = srcFile.iterator();

      if (sumOfWeights == 0) {
         Arrays.fill(quantileValues, Double.NaN);
         smallestValue = Double.NaN;
      } else {
         final Sample firstSample = iter.next();
         smallestValue = firstSample.value;
         Sample prevSample = null;
         Sample currSample = firstSample;
         /** includes the weight of the current sample */
         double aggregatedWeight = currSample.weight;

         for (int q = 0; q < quantileValues.length; q++) {
            final double relativeQuantilePosition = quantiles.getRelativeQuantilePosition(q);

            while (aggregatedWeight / sumOfWeights < relativeQuantilePosition) {
               prevSample = currSample;
               currSample = iter.next();
               assert Double.compare(currSample.value, prevSample.value) >= 0;
               aggregatedWeight += currSample.weight;
            }
            quantileValues[q] = currSample.value;
         }
      }
   }

   private static double sumOfWeights(DataSource srcFile) {
      double sum = 0;
      for (Sample s : srcFile)
         sum += s.weight;
      return sum;
   }

   public double getSmallestValue() {
      return smallestValue;
   }

   /**
    * @param index
    *           0 means first quantile, not smallest value
    * @return P(x<=thisValue) = getQuantileProbability(index)
    */
   public double getQuantileValue(int index) {
      return quantileValues[index];
   }

   /**
    * 
    * @param index
    *           0 means first quantile, not smallest value
    * @return P(x<=getQuantileValue(index)) = thisValue
    */
   public double getQuantileProbability(int index) {
      return quantiles.getRelativeQuantilePosition(index);
   }
}