/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.median;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * The classes <code>
   MedianStatistic
</code> or
 * {@link ikr.simlib.statistics.median.StdMedianStatistic} serve to register
 * single measured values (e.g., run times) of the type <code>
   double
</code> with the help of
 * the method {@link #update(double)}. <br>
 * <br>
 * The following functions are available for evaluation:
 * <ul>
 * <li>
 * Median value of all returned measured values ({@link #getMedian()}) including
 * a confidence interval ({@link #getMedianConfidenceInterval()})
 * <li>
 * Value range ({@link #getMinimum()} and {@link #getMaximum()}) in reference to
 * all measured values
 * <li>
 * Number of measured values from the batch ({@link #getBatchEvents()})
 * </ul>
 * If the parameter <i>maxBufferSize</i> is set to zero, all samples will be
 * stored internally and the median will be calculated at the end of the
 * simulation. Since this is very memory consuming, <i>maxBufferSize</i> can be
 * set to a number > 0, which determines the maximum number of samples the
 * statistic should hold in memory. The median is then calculated at the end of
 * the simulation as an approximation that is based on all samples.
 * <p>
 * For printing the results it defines the following keywords:
 * </p>
 * 
 * <dl>
 * <dt>
 * median</dt>
 * <dd>
 * Median value of all measured values in reference to all batches</dd>
 * <dt>
 * cintm</dt>
 * <dd>
 * confidence interval to median value</dd>
 * <dt>
 * max</dt>
 * <dd>
 * maximum sample of all batches</dd>
 * <dt>
 * min</dt>
 * <dd>
 * minimum sample of all batches</dd>
 * <dt>
 * bmedian</dt>
 * <dd>
 * number of samples in actual Batch</dd>
 * </dl>
 * The calculation of the median usually needs to buffer all measured values. In
 * order to save memory an approximation algorithm is served by the median
 * statistic class:
 * <ul>
 * <li>
 * all values are saved in a list of value pairs of one double (value) and one
 * integer value (weight)
 * <li>
 * the weight is set to 1 for every new value added to the list
 * <li>
 * when the list size exceeds the maximum buffer size, all entries are sorted by
 * the value and grouped to pairs by adding their weight and calculating like
 * this: <br>
 * value = value1  weight1 + value2  weight2 <br>
 * <li>
 * When the method {@link #getMedian()} is called the list is sorted again by
 * value. Then the list is iterated from beginning until the sum of the
 * weight-values is larger than the half of the total sum of weights. After this
 * the found item and the previous item are selected and an linear approximation
 * between the two value pairs is done.
 * </ul>
 */
public abstract class MedianStatistic extends Statistic implements Printable, UnweightedSampleCollector<Double> {

   public MedianStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      switch (type) {
      case FINAL:
         printComments(writer);
         writer.writeDouble(ResultKeywords.MEDIAN, getMedian());
         writer.writeDouble(ResultKeywords.MIN, getMinimum());
         writer.writeDouble(ResultKeywords.MAX, getMaximum());
         break;

      case BATCH:
         writer.writeDouble(ResultKeywords.BATCH_MEDIAN, getBatchMedian());
         break;
      }

   }

   public abstract void update(double sample);

   @Override
   public void update(Double sample) {
      update(sample.doubleValue());
   }

   public abstract double getMedian();

   public abstract double getMedianConfidenceInterval();

   public abstract long getEvents();

   public abstract double getMaximum();

   public abstract double getMinimum();

   public abstract double getBatchMedian();

   public abstract long getBatchEvents();
}
