/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.probability;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.WeightedSampleCollector;

/**
 * The class <code>ProbabilityStatistic</code> serves to record the results of
 * Bernoulli trials, which are random experiments with only two possible results
 * (e.g., failure probability in a queue with limited capacity).
 * {@link #update(long, boolean)} records the number of trials that have the
 * same result (<code>true</code> or <code>false</code> ). <br>
 * <br>
 * The following functions are available for evaluation:
 * <ul>
 * <li>
 * Batch probability as a quotient of the number of favorable events and the
 * total number of events ({@link #getBatchProbability()})
 * <li>
 * Mean value of the batch probabilities ({@link #getMean()}) including a
 * confidence interval ({@link #getConfidenceInterval()})
 * </ul>
 * The above mentioned functions, that have only been declared in a virtual
 * manner in <code>ProbabilityStatistic</code>, have been implemented in the
 * derived class
 * {@link ikr.simlib.statistics.probability.StdProbabilityStatistic}. An object
 * of this class can be created with the
 * <code>createProbabilityStatistic()</code> method from
 * <code>StatisticManager</code> and deleted with
 * <code>deleteProbabilityStatistic()</code>.
 * <p>
 * For printing the results it defines the following keywords:
 * </p>
 * 
 * <dl>
 * <dt>
 * mean</dt>
 * <dd>
 * Mean value of the probabilities in all batches</dd>
 * <dt>
 * cintmean</dt>
 * <dd>
 * Confidence interval for the mean value in +/- notation</dd>
 * <dt>
 * bprob</dt>
 * <dd>
 * Probability in reference to the current batches</dd>
 * </dl>
 */
public abstract class ProbabilityStatistic extends Statistic implements Printable, WeightedSampleCollector<Boolean, Long> {

   public ProbabilityStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      switch (type) {
      case FINAL:
         printComments(writer);
         writer.writeDouble(ResultKeywords.MEAN, getMean());
         writer.writeDouble(ResultKeywords.CINT, getConfidenceInterval());
         break;

      case BATCH:
         writer.writeDouble(ResultKeywords.BATCH_PROBABILITY, getBatchProbability());
         break;
      }

   }

   // -- Results --

   public abstract void update(long number, boolean special);

   @Override
   public void update(Boolean special) {
      update(1, special.booleanValue());
   }

   @Override
   public void update(Boolean sample, Long weight) {
      update(weight.longValue(), sample.booleanValue());
   }

   public abstract double getMean();

   public abstract double getBatchProbability();

   public abstract double getConfidenceInterval();
}
