/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.rate;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * The class <code>RateStatistic</code> serves to record rates, i.e. the amount
 * of information per time as well as the number of samples per time. The
 * {@link #update(double)} function takes a double value indicating the amount
 * of information. Besides <i>name</i> and <i>owner</i>, an entity is required
 * for constructor call to have access to the method
 * <code>getSystemTime()</code>. <br>
 * <br>
 * The following functions are available for evaluation: <br>
 * <br>
 * <ul>
 * <li>
 * Mean value in [samples/time] of all returned measured rates (
 * {@link #getSampleRate()}) including a confidence interval (
 * {@link #getRateConfInterval()}).
 * </ul>
 * <p>
 * For printing the results it defines the following keywords:
 * </p>
 * 
 * <dl>
 * <dt>
 * rate</dt>
 * <dd>
 * Mean of all rates [units/time] in all batches</dd>
 * <dt>
 * cintrate</dt>
 * <dd>
 * Confidence interval to mean rates</dd>
 * <dt>
 * samplerate</dt>
 * <dd>
 * Mean of all rates [samples/time] in all batches</dd>
 * <dt>
 * cintsamplerate</dt>
 * <dd>
 * Confidence interval to mean rates</dd>
 * </dl>
 */
public abstract class RateStatistic extends Statistic implements Printable, UnweightedSampleCollector<Double> {

   public RateStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      switch (type) {
      case FINAL:
         printComments(writer);
         writer.writeDouble(ResultKeywords.RATE, getRate());
         writer.writeDouble(ResultKeywords.RATE_CINT, getRateConfInterval());
         writer.writeDouble(ResultKeywords.SAMPLE_RATE, getSampleRate());
         writer.writeDouble(ResultKeywords.SAMPLE_RATE_CINT, getSampleRateConfInterval());
         break;

      case BATCH:
         writer.writeDouble(ResultKeywords.BATCH_RATE, getBatchRate());
         writer.writeDouble(ResultKeywords.BATCH_SAMPLE_RATE, getBatchSampleRate());
         break;
      }
   }

   // -- Results --

   public abstract void update(double amount);

   @Override
   public void update(Double amount) {
      update(amount.doubleValue());
   }

   public abstract double getRate();

   public abstract double getRateConfInterval();

   public abstract double getBatchRate();

   public abstract double getSampleRate();

   public abstract double getSampleRateConfInterval();

   public abstract double getBatchSampleRate();
}
