/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.sample;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * Similar to {@link SampleStatistic}, but ignores batches. It therefore does
 * not calculate confidence intervals. The calculated mean is the true mean over
 * all samples, independent of the number of values per batch. For simplicity,
 * only the mean and the number of events is written to the result file.
 */
public class MeanStatistic extends Statistic implements Printable, BatchExportable, UnweightedSampleCollector<Double> {
   private double sum;
   private long count;

   public MeanStatistic(SimNode ownNode) {
      super(ownNode);
      simNode.setPrintHandler(this);
      simNode.setExportHandler(this);
   }

   @Override
   public void update(Double value) {
      sum += value;
      count++;
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      if (type == ResultType.FINAL) {
         printComments(writer);
         writer.writeDouble(ResultKeywords.MEAN, getMean());
         writer.writeLong(ResultKeywords.NUMBER_OF_EVENTS, getEvents());
      }
   }

   @Override
   public void resetStatistic() {
      sum = 0;
      count = 0;
   }

   @Override
   public void resetBatchStatistic() {
      // noop
   }

   @Override
   public void computeMeasures(int batchNumber) {
      // noop
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      writer.writeDouble("Sum", sum);
      writer.writeLong("Count", count);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      sum += reader.readDouble("Sum");
      count += reader.readLong("Count");
   }

   public double getMean() {
      return sum / count;
   }

   public long getEvents() {
      return count;
   }

}
