/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.weighted;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.Summation;
import ikr.simlib.statistics.WeightedSampleCollector;

/**
 * <code>WeightedDistStatistic</code> provides an implementation for a weighted
 * distribution statistic.
 * 
 * <p>
 * For printing the results it defines the following keywords:
 * </p>
 * 
 * <dl>
 * <dt>list</dt>
 * <dd>prints a list of results; what exactly is printed on each line has to be
 * specified using the following keywords</dd>
 * <dt>ubound</dt>
 * <dd>upper bound of current bucket</dd>
 * <dt>lbound</dt>
 * <dd>lower bound of current bucket</dd>
 * <dt>muprob</dt>
 * <dd>underflow probability (mean of batch results)</dd>
 * <dt>cintuprob</dt>
 * <dd>underflow probability confidence interval</dd>
 * <dt>moprob</dt>
 * <dd>overflow probability (mean of batch results)</dd>
 * <dt>cintoprob</dt>
 * <dd>overflow probability confidence interval</dd>
 * <dt>dist</dt>
 * <dd>(cumulative) distribution (cdf)</dd>
 * <dt>prob</dt>
 * <dd>probability (density) (pdf)</dd>
 * <dt>compdist</dt>
 * <dd>complementary (cumulative) distribution (ccdf)</dd>
 * <dt>cintdist</dt>
 * <dd>confidence interval for cdf/ccdf</dd>
 * <dt>cintprob</dt>
 * <dd>confidence interval for pdf</dd>
 * </dl>
 */
public class WeightedDistStatistic extends Statistic implements BatchExportable, Printable, WeightedSampleCollector<Double, Double> {

   private static final String SUM_OF_WEIGHTS = ResultKeyword.asExport("sumofweights");
   private static final String WEIGHT = ResultKeyword.asExport("weight");
   private static final String WEIGHT_LIST = ResultKeyword.asExport("weightlist");
   private static final String UNDERFLOWS = ResultKeyword.asExport("underflows");
   private static final String OVERFLOWS = ResultKeyword.asExport("overflows");

   private final int arraySize;
   private final double classWidth;
   private final double lowerLimit;
   private final double upperLimit;

   private double underflows;
   private double overflows;
   private double[] weights;
   private double sumOfWeights;

   private final Summation underflowProb;
   private final Summation overflowProb;
   private final Summation[] marginalDist;
   private final Summation[] cumulativeDist;

   public WeightedDistStatistic(int arraySize, double lowerLimit, double upperLimit, SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
      simNode.setExportHandler(this);

      this.arraySize = arraySize;
      this.lowerLimit = lowerLimit;
      this.upperLimit = upperLimit;
      weights = new double[arraySize];
      underflowProb = new Summation();
      overflowProb = new Summation();
      marginalDist = new Summation[arraySize];
      cumulativeDist = new Summation[arraySize];

      classWidth = (upperLimit - lowerLimit) / arraySize;

      for (int i = 0; i < arraySize; i++) {
         marginalDist[i] = new Summation();
         cumulativeDist[i] = new Summation();
      }

      resetStatistic();
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      switch (type) {
      case FINAL:
         printComments(writer);
         writer.writeStartArray(ResultKeywords.LIST.toString());
         for (int i = 0; i < arraySize; i++) {
            writer.writeStartArrayItem(i);
            writer.writeDouble(ResultKeywords.UBOUND, getUpperBound(i));

            if (!Double.isNaN(getDistribution(i))) {
               writer.writeDouble(ResultKeywords.CDIST, 1.0 - getDistribution(i));
            } else {
               writer.writeDouble(ResultKeywords.CDIST, Double.NaN);
            }

            writer.writeDouble(ResultKeywords.DISTRIBUTION_CINT, getDistributionConfInterval(i));
            writer.writeEndArrayItem();
         }
         writer.writeEndArray();
         break;

      case BATCH:
         writer.writeStartArray(ResultKeywords.LIST.toString());
         for (int i = 0; i < arraySize; i++) {
            writer.writeStartArrayItem(i);
            writer.writeDouble(ResultKeywords.UBOUND, getUpperBound(i));

            if (!Double.isNaN(getDistribution(i))) {
               writer.writeDouble(ResultKeywords.CDIST, 1.0 - getDistribution(i));
            } else {
               writer.writeDouble(ResultKeywords.CDIST, Double.NaN);
            }

            writer.writeEndArrayItem();
         }
         writer.writeEndArray();
         break;
      }
   }

   // -- State --

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      writer.writeDouble(UNDERFLOWS, underflows);
      writer.writeDouble(OVERFLOWS, overflows);
      writer.writeDouble(SUM_OF_WEIGHTS, sumOfWeights);

      writer.writeStartArray(WEIGHT_LIST);
      for (int i = 0; i < weights.length; i++) {
         writer.writeStartArrayItem(i);
         writer.writeDouble(WEIGHT, weights[i]);
         writer.writeEndArrayItem();
      }
      writer.writeEndArray();
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      underflows = reader.readDouble(UNDERFLOWS);
      overflows = reader.readDouble(OVERFLOWS);
      sumOfWeights = reader.readDouble(SUM_OF_WEIGHTS);

      reader.readStartArray(WEIGHT_LIST);
      for (int i = 0; i < weights.length; i++) {
         reader.readStartArrayItem(i);
         weights[i] = reader.readDouble(WEIGHT);
         reader.readEndArrayItem();
      }

      computeMeasures(0);
      reader.readEndArray();
   }

   // -- Statistic --

   public void update(double sample, double weight) {
      if (weight < 0) {
         throw new SimulationException("weight must be >= 0!");
      }

      if (sample < lowerLimit) {
         underflows += weight;
      } else if (sample >= upperLimit) {
         overflows += weight;
      } else {
         int index = (int) ((sample - lowerLimit) / classWidth);
         weights[index] += weight;
      }

      sumOfWeights += weight;
   }

   @Override
   public void update(Double sample, Double weight) {
      update(sample.doubleValue(), weight.doubleValue());
   }

   public final void update(double sample) {
      update(sample, 1);
   }

   @Override
   public void update(Double sample) {
      update(sample.doubleValue());
   }

   @Override
   public void computeMeasures(int batchNumber) {
      if (sumOfWeights > 0) {
         underflowProb.update(underflows / sumOfWeights);
         overflowProb.update(overflows / sumOfWeights);

         double probabilitySum = underflows / sumOfWeights;
         for (int i = 0; i < arraySize; i++) {
            double probability = weights[i] / sumOfWeights;
            marginalDist[i].update(probability);
            probabilitySum += probability;
            cumulativeDist[i].update(probabilitySum);
         }
      }
   }

   @Override
   public void resetBatchStatistic() {
      underflows = 0;
      overflows = 0;
      sumOfWeights = 0;
      for (int i = 0; i < arraySize; i++) {
         weights[i] = 0;
      }
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();
      underflowProb.reset();
      overflowProb.reset();
      for (int i = 0; i < arraySize; i++) {
         marginalDist[i].reset();
         cumulativeDist[i].reset();
      }
   }

   public int getArraySize() {
      return arraySize;
   }

   public double getClassWidth() {
      return classWidth;
   }

   public double getLowerLimit() {
      return lowerLimit;
   }

   public double getUpperLimit() {
      return upperLimit;
   }

   public double getLowerBound(int bucketIndex) {
      return lowerLimit + bucketIndex * classWidth;
   }

   public double getUpperBound(int bucketIndex) {
      return lowerLimit + (bucketIndex + 1) * classWidth;
   }

   public double getUnderflowProbability() {
      return underflowProb.getMean();
   }

   public double getUnderflowConfInterval() {
      return underflowProb.getConfidenceInterval();
   }

   public double getOverflowProbability() {
      return overflowProb.getMean();
   }

   public double getOverflowConfInterval() {
      return overflowProb.getConfidenceInterval();
   }

   public double getProbability(int bucketIndex) {
      return marginalDist[bucketIndex].getMean();
   }

   public double getProbabilityConfInterval(int bucketIndex) {
      return marginalDist[bucketIndex].getConfidenceInterval();
   }

   public double getDistribution(int bucketIndex) {
      return cumulativeDist[bucketIndex].getMean();
   }

   public double getDistributionConfInterval(int bucketIndex) {
      return cumulativeDist[bucketIndex].getConfidenceInterval();
   }
}
