/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.support;

/**
 * This class contains support functions for working with matrices.
 * 
 */
public final class Matrices {
   /**
    * Subtracts the second from the first matrix and returns a new matrix.
    */
   public static double[][] sub(double[][] matA, double[][] matB) {
      final int N = matA.length;
      if (N != matB.length) {
         throw new IllegalArgumentException("Matrices are not of equal size");
      }
      if (N == 0) {
         return new double[0][0];
      }

      final int M = matA[0].length;
      if (M != matB[0].length) {
         throw new IllegalArgumentException("Matrices are not of equal size");
      }
      if (M == 0) {
         return new double[0][0];
      }

      final double[][] result = new double[N][M];

      for (int n = 0; n < N; n++) {
         for (int m = 0; m < M; m++) {
            result[n][m] = matA[n][m] - matB[n][m];
         }
      }

      return result;
   }

   /**
    * Creates a identity matrix.
    * 
    * @param size
    *           number of rows an columns
    */
   public static double[][] identity(int size) {
      final int N = size;
      final double[][] result = new double[N][N];

      for (int n = 0; n < N; n++) {
         result[n][n] = 1;
      }

      return result;
   }

   /**
    * Creates a diagonal matrix from a vector.
    */
   public static double[][] diagonal(double[] vec) {
      final int N = vec.length;

      final double[][] result = new double[N][N];

      for (int n = 0; n < N; n++) {
         result[n][n] = vec[n];
      }

      return result;
   }

   /**
    * Creates a transposed matrix.
    */
   public static double[][] transpose(double[][] mat) {
      final int N = mat.length;
      if (N == 0) {
         return new double[0][0];
      }

      // We don't check all lengths to be equal. The VM will recognize it
      // anyway
      final int M = mat[0].length;

      final double[][] result = new double[M][N];

      for (int n = 0; n < N; n++) {
         for (int m = 0; m < M; m++) {
            result[m][n] = mat[n][m];
         }
      }

      return result;
   }

   /**
    * Creates a submatrix excluding the specified row and column.
    */
   public static double[][] submatrix(double[][] mat, int row, int column) {
      final int N = mat.length;
      if (N == 0) {
         throw new IllegalArgumentException("Matrix has no elements");
      }

      final int M = mat[0].length;
      if (row < 0 || row >= N || column < 0 || column >= M) {
         throw new IllegalArgumentException("Row or column out of range");
      }

      final double[][] result = new double[N - 1][M - 1];

      for (int n = 0, i = 0; i < N - 1; n++, i++) {
         if (n == row) {
            n++;
         }

         for (int m = 0, j = 0; j < M - 1; m++, j++) {
            if (m == column) {
               m++;
            }

            result[i][j] = mat[n][m];
         }
      }

      return result;
   }

   /**
    * Calculates the determinant of a matrix.
    */
   public static double determinant(double[][] mat) {
      final int N = mat.length;
      if (N == 0) {
         return 0;
      }

      // Again we take the length of all rows as equal
      if (mat[0].length != N) {
         throw new IllegalArgumentException("Matrix not square");
      }

      switch (N) {
      case 1:
         return mat[0][0];

      case 2:
         return mat[0][0] * mat[1][1] - mat[0][1] * mat[1][0];

      case 3:
         return mat[0][0] * mat[1][1] * mat[2][2] + mat[0][1] * mat[1][2] * mat[2][0] + mat[0][2] * mat[1][0] * mat[2][1] - mat[0][2] * mat[1][1] * mat[2][0] - mat[0][0] * mat[1][2] * mat[2][1] - mat[0][1] * mat[1][0] * mat[2][2];

      }

      double det = 0;

      double factor = 1;
      for (int n = 1; n < N; n++) {
         if (mat[0][n - 1] != 0) {
            det += factor * mat[0][n - 1] * determinant(submatrix(mat, 0, n));
         }

         factor *= -1;
      }

      return det;
   }
}
