/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.statistics.variancetime;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.Summation;

import java.io.FileNotFoundException;
import java.io.PrintWriter;

/**
 * <code>VTStatistic</code> is a base class for a variance-time statistic in
 * order to estimate the Hurst parameter.
 */
public class VTStatistic extends Statistic implements BatchExportable, Printable, ReflectionConstructable2<SimNode, Parameters> {

   private static final String LD_MIN_BLOCK_SIZE = "LdMinBlockSize";
   private static final String LD_MAX_BLOCK_SIZE = "LdMaxBlockSize";
   private static final String TRACE_FILE = "TraceFile";

   /** as required by {@link ReflectionConstructable} */
   public static VTStatistic createInstance(SimNode ownNode, Parameters pars) {
      final int ldMinBlockSize = pars.get(ownNode, LD_MIN_BLOCK_SIZE).asInteger();
      final int ldMaxBlockSize = pars.get(ownNode, LD_MAX_BLOCK_SIZE).asInteger();
      final String traceFileName = pars.get(ownNode, TRACE_FILE).asString();
      return new VTStatistic(ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode);
   }

   public static final String EXPORT_BLOCK_MEAN = ResultKeyword.asExport("blockmean");
   public static final String EXPORT_BLOCK_MEAN_LIST = ResultKeyword.asList(EXPORT_BLOCK_MEAN);

   // FIXME: Use some proper logging facility here
   protected PrintWriter traceFile;
   protected String traceFileName;

   protected final int ldMinBlockSize;
   protected final int ldMaxBlockSize;

   protected final int minBlockSize;

   protected final Summation[] measuresSummations;
   protected final Summation[] blockMeansSummations;

   public VTStatistic(int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, SimNode ownNode) {
      super(ownNode);

      simNode.setExportHandler(this);
      simNode.setPrintHandler(this);

      if (ldMaxBlockSize > 32) {
         throw new SimulationException("maximum block size must be < 2^32");
      }

      if (ldMinBlockSize > ldMaxBlockSize) {
         throw new SimulationException("maximum block size must be >= minimum block size");
      }

      this.ldMinBlockSize = ldMinBlockSize;
      this.ldMaxBlockSize = ldMaxBlockSize;
      this.traceFileName = traceFileName;
      minBlockSize = 1 << ldMinBlockSize;
      traceFile = null;

      blockMeansSummations = new Summation[ldMaxBlockSize - ldMinBlockSize + 1];
      measuresSummations = new Summation[ldMaxBlockSize - ldMinBlockSize + 1];

      for (int i = 0; i <= ldMaxBlockSize - ldMinBlockSize; i++) {
         blockMeansSummations[i] = new Summation();
         measuresSummations[i] = new Summation();
      }

      resetStatistic();
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      final int maxElements = ldMaxBlockSize - ldMinBlockSize + 1;
      int elements = 0;

      switch (type) {
      case BATCH:
         elements = 0;
         while (elements < maxElements && !Double.isNaN(getBatchVariance(elements))) {
            elements++;
         }
         writer.writeStartArray(ResultKeywords.BATCH_VARIANCE_LIST.toString());
         for (int i = 0; i < elements; i++) {
            writer.writeStartArrayItem(i);
            writer.writeDouble(ResultKeywords.BATCH_BLOCKSIZE, (int) Math.pow(minBlockSize, i));
            writer.writeDouble(ResultKeywords.BATCH_VARIANCE, getBatchVariance(i));
            writer.writeEndArrayItem();
         }
         writer.writeEndArray();
         break;

      case FINAL:
         elements = 0;
         while (elements < maxElements && !Double.isNaN(getVariance(elements))) {
            elements++;
         }
         writer.writeStartArray(ResultKeywords.VARIANCE_LIST.toString());
         for (int i = 0; i < elements; i++) {
            writer.writeStartArrayItem(i);
            writer.writeDouble(ResultKeywords.BLOCKSIZE, (int) Math.pow(minBlockSize, i));
            writer.writeDouble(ResultKeywords.VARIANCE, getVariance(i));
            writer.writeDouble(ResultKeywords.CINT, getConfidenceInterval(i));
            writer.writeEndArrayItem();
         }
         writer.writeEndArray();

         writer.writeStartArray(ResultKeywords.NORM_VAR_LIST.toString());
         for (int i = 0; i < elements; i++) {
            writer.writeStartArrayItem(i);
            writer.writeDouble(ResultKeywords.BLOCKSIZE, (int) Math.pow(minBlockSize, i));
            writer.writeDouble(ResultKeywords.VARIANCE, getVariance(i) / getVariance(0));
            writer.writeDouble(ResultKeywords.CINT, getConfidenceInterval(i) / getVariance(0));
            writer.writeEndArrayItem();
         }
         writer.writeEndArray();

         break;
      }
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      writer.writeStartArray(EXPORT_BLOCK_MEAN_LIST);
      for (int i = 0; i < blockMeansSummations.length; i++) {
         writer.writeStartArrayItem(i);

         blockMeansSummations[i].exportSum(EXPORT_BLOCK_MEAN, writer);

         writer.writeEndArrayItem();
      }
      writer.writeEndArray();
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      reader.readStartArray(EXPORT_BLOCK_MEAN_LIST);
      for (int i = 0; i < blockMeansSummations.length; i++) {
         reader.readStartArrayItem(i);

         blockMeansSummations[i].importSum(EXPORT_BLOCK_MEAN, reader);

         reader.readEndArrayItem();
      }
      reader.readEndArray();
   }

   @Override
   public void handleStopTransientPhase() {
      super.handleStopTransientPhase();

      if (traceFile == null && !traceFileName.equals("")) {
         if (traceFileName.equals("__cout")) {
            traceFile = new PrintWriter(System.out, true);
         } else {
            try {
               traceFile = new PrintWriter(traceFileName);
            } catch (FileNotFoundException fnfe) {
               throw new SimulationException("Trace file \"" + traceFileName + "\"not found");
            }
         }
      }
   }

   @Override
   public void computeMeasures(int batchNumber) {
      for (int i = 0; i <= ldMaxBlockSize - ldMinBlockSize && !Double.isNaN(blockMeansSummations[i].getVariance()); i++) {
         measuresSummations[i].update(blockMeansSummations[i].getVariance());
      }
      resetBatchStatistic();
   }

   @Override
   public void resetBatchStatistic() {
      for (Summation s : blockMeansSummations) {
         s.reset();
      }
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();

      for (Summation s : measuresSummations) {
         s.reset();
      }
   }

   public double getVariance(int ldBlockSize) {
      return measuresSummations[ldBlockSize].getMean();
   }

   public double getConfidenceInterval(int ldBlockSize) {
      return measuresSummations[ldBlockSize].getConfidenceInterval();
   }

   public double getBatchVariance(int ldBlockSize) {
      return blockMeansSummations[ldBlockSize].getMean();
   }
}
