/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.trace;

import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * TraceFile is supportive class which is responsible for handling the trace
 * files. The trace file must have the following structure:
 * 
 * <p>
 * <code>hh:mm:ss.sss length</code>
 * </p>
 * 
 * <p>
 * The time information can either be absolute or relative to the preeceding
 * event (=interarrival time), this can be set by <code>absoluteTime</code>. For
 * absolute time stamps, all time stamps are shifted such that the first event
 * is at time zero.
 * </p>
 * 
 * <p>
 * Further the behavior at the end of the trace file can be determined. If
 * <code>wrapFile</code> is set, the trace file is reread from the beginning.
 * Otherwise the simulation interrupts with an error when the end of the trace
 * file is reached. If time information is absolute, the duration between the
 * last item of run <code>n</code> and the first item of run <code>n+1</code> is
 * set to twice the duration between the last two items of run <code>n</code>.
 * All following time stamps are shifted such that the remaining interarrival
 * times are the same as the differences of the time values in the trace file.
 * </p>
 * 
 * <p>
 * Comments are initiated by a leading '#'.
 * </p>
 * 
 * <p>
 * <code>TraceFile</code> is a parseable class. It accepts the following
 * keywords in the parameter file:
 * </p>
 * 
 * <dl>
 * <dt>FileName</dt>
 * <dd>The name of the trace file</dd>
 * <dt>AbsoluteTime (boolean, optional)</dt>
 * <dd>Set to <code>TRUE</code> if the times in the trace file represent
 * absolute times. Set to <code>FALSE</code> if they represent relative times
 * (i.e., time differences). Default is <code>TRUE</code></dd>
 * <dt>WrapFile (boolean, optional)</dt>
 * <dd>Set to <code>TRUE</code> if you want the trace to be reread from the
 * beginning once the end is reached. If set to FALSE, an exception will be
 * thrown when the end of the trace file is reached. Default is
 * <code>FALSE</code>.</dd>
 * </dl>
 */
public class TraceFile implements ReflectionConstructable2<SimNode, Parameters> {
   private static final Pattern TRACE = Pattern.compile("^([0-9]+):([0-9]+):([\\.0-9]+)\\s+([0-9]+)");

   /** as required by {@link ReflectionConstructable} */
   public static TraceFile createInstance(SimNode ownNode, Parameters pars) {
      final String fileName = pars.get(ownNode, "FileName").asString();
      final boolean absoluteTime = pars.get(ownNode, "AbsoluteTime").asBoolean();
      final boolean wrapFile = pars.get(ownNode, "WrapFile").asBoolean();
      return new TraceFile(fileName, absoluteTime, wrapFile);
   }

   protected final String fileName;
   protected final RandomAccessFile file;
   /** Contains the traceFile absolute or relative times */
   protected final boolean absoluteTime;
   /** Behavior, if the end of the file is reached */
   protected final boolean wrapFile;
   /** Determined from the first value */
   protected final Duration timeDiffForFirstItem;
   protected TraceItem currentItem;
   protected TraceItem previousItem;
   protected Duration iat;
   protected boolean endOfFile;
   protected boolean firstRunFinished;

   public TraceFile(String fileName, boolean absoluteTime, boolean wrapFile) {
      this.fileName = fileName;
      this.absoluteTime = absoluteTime;
      this.wrapFile = wrapFile;

      try {
         file = new RandomAccessFile(fileName, "r");
      } catch (IOException ioe) {
         throw new SimulationException("Couldn't open trace file " + fileName + ": " + ioe.getMessage());
      }

      currentItem = readNextItem();
      timeDiffForFirstItem = currentItem.getTimeAsDuration().negated();
   }

   public TraceFile(String filename) {
      this(filename, true, false);
   }

   public PointInTime getNextEventTime(PointInTime currentTime) {
      final PointInTime nextEventTime;
      if (absoluteTime) {
         if (!firstRunFinished) { // first run
            if (previousItem == null) { // first item always at time zero
               nextEventTime = currentItem.getTimeAsPointInTime().plus(timeDiffForFirstItem);
               assert nextEventTime.equals(PointInTime.ZERO);
            } else { // convert absolute times to IAT
               iat = previousItem.getTimeAsPointInTime().until(currentItem.getTimeAsPointInTime());
               assert iat.isGreaterThanOrEqualTo(Duration.ZERO);
               nextEventTime = currentTime.plus(iat);
            }
         } else { // wrapped
            assert wrapFile;
            if (!endOfFile) {
               iat = currentItem.getTimeAsPointInTime().until(previousItem.getTimeAsPointInTime());
               assert iat.isGreaterThanOrEqualTo(Duration.ZERO);
               nextEventTime = currentTime.plus(iat);
            } else { // just switched from the last to the first line
               assert previousItem.getTimeAsPointInTime().isAfter(currentItem.getTimeAsPointInTime());
               // in order to make the jump from the end of the trace file
               // to the beginning smooth, the last IAT is considered
               // twice. Therefore a higher priority is given to it.
               nextEventTime = currentTime.plus(iat.multiplyBy(2));
               endOfFile = false;
            }
         }
      } else { // relative time
         assert !firstRunFinished || wrapFile;
         nextEventTime = currentTime.plus(currentItem.getTimeAsDuration());
      }
      return nextEventTime;
   }

   /**
    * returns the length of the current message and moves to the next message
    * 
    * @return length of the next message
    */
   public int getNextLength() {
      int length = currentItem.getLength();
      previousItem = currentItem;
      currentItem = readNextItem();
      return length;
   }

   /**
    * reads the next line from the trace file and creates a new TraceItem from
    * that
    * 
    * @return the new TraceItem
    */
   private TraceItem readNextItem() {
      TraceItem nextone = null;

      try {
         do {
            String line = file.readLine();

            // End of file?
            if (line == null) {
               if (wrapFile) {
                  file.seek(0);
                  endOfFile = true;
                  firstRunFinished = true;
                  continue;
               }
               throw new SimulationException("At end of trace file!");
            }

            line = line.trim();

            // Skip blank lines and comments
            if (line.equals("") || line.charAt(0) == '#') {
               continue;
            }

            Matcher match = TRACE.matcher(line);
            if (!match.matches()) {
               throw new SimulationException("Invalid trace \"" + line + "\"");
            }

            try {
               int hour = Integer.parseInt(match.group(1));
               int minutes = Integer.parseInt(match.group(2));
               double seconds = Double.parseDouble(match.group(3));
               int length = Integer.parseInt(match.group(4));

               nextone = new TraceItem(hour, minutes, seconds, length);
            } catch (NumberFormatException nfe) {
               throw new SimulationException("Invalid trace \"" + line + "\"");
            }
         } while (nextone == null);
      } catch (IOException ioe) {
         throw new SimulationException("I/O error during trace file parsing: " + ioe.getMessage());
      }

      return nextone;
   }

   public String getFileName() {
      return fileName;
   }
}
